/*
 ******************************************************************
 * HISTORY
 * 15-Oct-94  Jeff Shufelt (js), Carnegie Mellon University
 *      Prepared for 15-681, Fall 1994.
 *
 ******************************************************************
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "image.h"
#include "util.h"

#define STR_SIZE 2048

/* 
 *
 * function definitions 
 *
 *
 */


img_t *img_new ()
{
  img_t *new;

  new = (img_t *) malloc (sizeof (img_t));
  if (new == NULL) {
    printf("img_new: Couldn't allocate image structure\n");
    return (NULL);
  }
  new->chan_cnt = 1;
  new->height = 0;
  new->width = 0;
  new->stride = 0;
  new->name = NULL;
  new->data = NULL;
  new->ftype = IMG_FTYPE_NONE;
  return (new);
}

/* sets the image buffer to all zeroes */

img_t *img_create (char *name, img_dim_t wd, img_dim_t ht, int chan_cnt)
{
  img_t *new;

  new = img_new();
  if (new == NULL)
    return (NULL);
  new->data = (pix_val_t *) calloc (1, ht * wd * chan_cnt 
				    * sizeof(pix_val_t));
  if (new->data == NULL) {
    img_free (new);
    return (NULL);
  }
  if (name != NULL) {
    new->name = strdup (img_basename(name));
    if (new->name == NULL) {
      img_free (new);
      return (NULL);
    }
  }
  new->height = ht;
  new->width = wd;
  new->chan_cnt = chan_cnt;
  new->stride = wd * chan_cnt;
  return (new);
}


/* create a new image, but copies the meta data from an existing 
 * image */

img_t *img_create_basis (img_t *src, img_dim_t wd, img_dim_t ht, int chan_cnt)
{
  img_t *new = img_create (src->name, wd, ht, chan_cnt);

  /* replicate the meta data */
  new->ftype = src->ftype;

  return (new);
}


void img_free (img_t *img)
{
  if (img) {
    if (img->data) free ((char *) img->data);
    if (img->name) free ((char *) img->name);
    free ((char *) img);
  }
}


void img_setpixel_rc (img_t *img, img_dim_t y, img_dim_t x, pix_val_t val)
{
  img_dim_t wd;

  wd = img->width;
  img->data[(y * wd) + x] = val;
}


pix_val_t img_getpixel_rc (img_t *img, img_dim_t y, img_dim_t x)
{
  int wd;

  wd = img->width;
  return (img->data[(y * wd) + x]);
}


img_list_t *imgl_new()
{
  img_list_t *new;

  new = (img_list_t *) malloc (sizeof (img_list_t));
  if (new == NULL) {
    printf("IMGL_NEW: Couldn't allocate image list\n");
    return(NULL);
  }

  new->cnt = 0;
  new->image = NULL;
  new->fname = NULL;

  return (new);
}


/* Note: the fname is not dup'd, the pointer is just copied */
/* return values: 0 = OK, 1 = problem */

int imgl_add (img_list_t *il, char *fname, img_t *img)
{
  unsigned long cnt;

  cnt = il->cnt + 1;

  il->fname = (char **)  realloc (il->fname, cnt * sizeof (char *));
  il->image = (img_t **) realloc (il->image,  cnt * sizeof (img_t *));

  if (il->image == NULL || il->fname == NULL) {
    printf("IMGL_ADD: Couldn't reallocate image list\n");
    return (1);
  }

  il->fname[cnt-1] = fname;
  il->image[cnt-1] = img;
  il->cnt = cnt;
  return (0);
}


void imgl_free (img_list_t *il)
{
  free((char *) il->image);
  free((char *) il);
}


char *img_basename (char *filename)
{
  char *basename;

  basename = strrchr (filename, '/');
  if (basename == NULL)
    basename = filename;
  else
    basename++;

  return (basename);
}


img_list_t *imgl_load_list (char *filename)
{
  FILE *fp;
  char buf[STR_SIZE], *image_fname;
  img_list_t *image_list;
  
  image_list = imgl_new();
  if (image_list == NULL)
    return NULL;

  if ((fp = fopen(filename, "r")) == NULL) {
    printf("Couldn't open file: %s\n", filename);
    return NULL;
  }

  while (fgets(buf, STR_SIZE, fp) != NULL) {
    image_fname = imgl_clean_str (buf);
    if (imgl_add(image_list, strdup(image_fname), NULL)) {
      printf ("Error building image list.\n");
      return NULL;
    }
  }

  fclose(fp);

  return (image_list);

}


/* clean up leading and trailing spaces of a string */

char *imgl_clean_str (char *str) {

  int i, len;
  char *new_str;
  
  len = strlen (str);
  new_str = str;

  /* remove the leading white space */

  for (i = 0 ; i < len; i++, new_str++)
    if (new_str[0] != ' ' && new_str[0] != '\t')
      break;

  /* remove the trailing white space */

  for (i = len - i - 1; i >= 0; new_str[i] = '\0', i--)
    if (new_str[i] != ' ' 
	&& new_str[i] != '\t' 
	&& new_str[i] != '\n' 
	&& new_str[i] != '\015')	/* control-M */
      break;

  return (new_str);
}


/*%************************************************************************
 *
 * Fill an image with the specified color
 *
 * The color is an array of 1 or 3 pix_val's depending
 * on whether the image is color or monochrome
 *
 **************************************************************************
 */

void img_fill (img_t *img, pix_val_t *color)
{

  img_dim_t   width, height, x, y;
  pix_val_t  *data;
  int         chan_cnt;

  /* get the image stats */

  width    = img->width;
  height   = img->height;
  chan_cnt = img->chan_cnt;

  /* copy the data over */

  data = img->data;
  for (y = 0; y < height; y++) {
    if (chan_cnt == 1) {
      for (x = 0; x < width; x++) {
	*data++ = color[0];
      }
    }
    else {
      for (x = 0; x < width; x++) {
	*data++ = color[0];
	*data++ = color[1];
	*data++ = color[2];
      }
    }
  }
    
  return;

}


/* create an empty color map of the given size */

colormap_t *img_colormap_alloc (int size, int chan_cnt)
{
  int i;
  colormap_t *new = EMALLOC (sizeof (colormap_t));
  new->size     = size;
  new->chan_cnt = chan_cnt;
  new->used     = EMALLOC (sizeof (pix_val_t) * size);
  for (i = 0; i < size; i++)
    new->used[i] = 0;
  size*= chan_cnt;
  new->data     = EMALLOC (sizeof (pix_val_t) * size);
  for (i = 0; i < size; i++)
    new->data[i] = 0;
  return (new);
}


/* free memory associated with a color map */

void img_colormap_free (colormap_t *cmap)
{
  if (cmap->data)
    efree (cmap->data);
  if (cmap->used)
    efree (cmap->used);
  efree (cmap);
}


/* find the index of a color in a colormap 
 * or -1 if not found */

int img_colormap_find (colormap_t *cmap, pix_val_t *color)
{
  int i, j;
  int size, chan_cnt;
  pix_val_t *data;

  size     = cmap->size;
  chan_cnt = cmap->chan_cnt;
  data     = cmap->data;
  for (i = 0; i < size; i++, data+= chan_cnt) {
    for (j = 0; j < chan_cnt; j++) {
      if (data[j] != color[j])
	goto next;
    }
    return (i);
  next:
    ;
  }
  return (-1);
}


/* fill in the used flags in the color map structure 
 *
 * the input image is a set of indecies, must be one channel
 *
 * return the number of unique colors used 
 */

int img_colormap_calc_used (colormap_t *cmap, img_t *img)
{
  long i, size, chan_cnt, used_cnt, cmap_size;
  pix_val_t *data, *used;

  cmap_size = cmap->size;
  used      = cmap->used;
  chan_cnt  = img->chan_cnt;
  data      = img->data;
  size      = img->width * img->height;

  if (chan_cnt != 1)
    return (0);

  for (i = 0; i < cmap_size; i++)
    used[i] = 0;

  used_cnt = 0;
  for (i = 0; i < size; i++, data++) {
    if (!used[*data]) {
      used_cnt++;
      used[*data] = 1;
      if (used_cnt == cmap_size)
	break;
    }
  }
  return (used_cnt);
}

/* check if the used colors make this image bilevel 
 *
 * if the used array is not already filled in,
 * you need to call img_color_map_calc_used before using
 * this function
 */

int img_colormap_check_used_bilevel (colormap_t *cmap)
{
  long i, chan_cnt, cmap_size;
  pix_val_t *used, *data;
  int sum;

  cmap_size = cmap->size;
  data      = cmap->data;
  used      = cmap->used;
  chan_cnt  = cmap->chan_cnt;

  for (i = 0; i < cmap_size; i++) {
    if (used[i]) {
      if (chan_cnt == 1) {
	if (data[i] != 0 && data[i] != 255) {
	  return (0);
	}
      }
      else {
	sum = (int)data[i*3]+ (int)data[i*3+1] + (int)data[i*3+2];
	if (sum != 0 && sum != (255*3)) {
	  return (0);
	}
      }
    }
  }
  return (1);
}



/*%************************************************************************
 *
 * Create a duplicate of an existing image
 *
 * all memory associated with the image is duplicated
 *
 **************************************************************************
 */

img_t *img_dup (img_t *src_img)
{

  img_t      *new_img;
  img_dim_t   i, size;
  pix_val_t  *src_data, *new_data;

  /* create a new image */
  new_img = img_create_basis (src_img,
			      src_img->width, src_img->height, 
			      src_img->chan_cnt);

  /* copy the data over */

  size = src_img->width * src_img->height * src_img->chan_cnt;
  src_data = src_img->data;
  new_data = new_img->data;
  for (i = 0; i < size; i++) {
    *new_data++ = *src_data++;
  }
    
  return (new_img);

}
