
/*
 ******************************************************************
 *
 * Convert an image from PNM, GIF, JPEG format to PS format
 *
 * Charles Rosenberg, 1/26/98
 *
 ******************************************************************
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <time.h>
#include "image.h"
#include "pnm.h"
#include "img_util.h"
#include "scale.h"
#include "jpeg_util.h"
#include "util.h"
#include "gif.h"

/*
 * NOTE
 *
 */


/* constants */

#define DEFAULT_DPI    75
#define CHARS_PER_LINE 70
#define DPI_KEYWORD    "dpi"
#define SCALE_KEYWORD  "scale"
#define MAX_FILE_CNT   100
#define THRESH         127
#define DEFAULT_PAGE_WD 8.5
#define DEFAULT_PAGE_HT 11.0

#define VERSION_STR "2001-07-20"

/* globals */

const char *hex_digits = "0123456789ABCDEF";

/* function prototypes */

void hex_encode (unsigned char *in_data,   unsigned long in_len, 
		 unsigned char **out_data, unsigned long *out_len);

void ascii85_encode (unsigned char *in_data,   unsigned long in_len, 
		     unsigned char **out_data, unsigned long *out_len);

void runlength_encode (unsigned char *in_data,   unsigned long in_len, 
		       unsigned char **out_data, unsigned long *out_len);

void bitpack_encode (unsigned char *in_data,   unsigned long in_len, 
		     unsigned char **out_data, unsigned long *out_len,
		     int bit_len, img_dim_t data_width);

void blank_adjust (img_t *image, int filter_flag, img_dim_t *rw, 
		   img_dim_t *lci, img_dim_t *rci, img_dim_t *dtc,
		   img_dim_t *dw, img_dim_t *dh, 
		   img_dim_t *xpo, img_dim_t *ypo,
		   pix_val_t blank_val);
void thresh_image (img_t *image, pix_val_t thresh);
void crop_image (pix_val_t *data_in, img_dim_t width, img_dim_t height,
		 img_dim_t left_crop_index, img_dim_t right_crop_index);

void show_help (char *cmd_name);



/*
 * the main program
 */

int main (int argc, char **argv)
{
  int        arg_index;
  char      *opt_str;
  FILE      *fptr;
  char      *in_fname, *out_fname, *fname_list[MAX_FILE_CNT];
  int        rotate_layout, bil_flag, filter_flag, multi_flag;
  int        rotate_flag;
  img_t     *image;
  img_dim_t  pixel_cnt, dpi, i, data_height, data_width, read_width;
  img_dim_t  xpos_offset, ypos_offset, data_top_crop, dpi_list[MAX_FILE_CNT];
  img_dim_t  left_crop_index, right_crop_index;
  img_dim_t  jpeg_size;
  pix_val_t *data;
  int        file_cnt, file_index, page_cnt, page_index;
  int        scale, scale_list[MAX_FILE_CNT], ftype;
  double     page_wd, page_ht;
  int        center_flag, eps_flag, cmap_flag, cmap_bit_len;
  int        jpeg_embed_flag, trim_flag;
  colormap_t *cmap;

  /* initialization */

  arg_index = 1;
  image = NULL;
  dpi = DEFAULT_DPI;
  out_fname = NULL;
  fptr = NULL;
  rotate_layout = 0;
  rotate_flag = 0;
  bil_flag = 0;
  xpos_offset = 0;
  ypos_offset = 0;
  data_top_crop = 0;
  left_crop_index = 0;		/* this is the x position of the left most 
				 * non-blank pixel */
  right_crop_index = 0;		/* this is the x position of the right most 
				 * non-blank pixel + 1 */
  filter_flag = 0;
  multi_flag = 0;
  scale = 1;
  page_wd = DEFAULT_PAGE_WD;
  page_ht = DEFAULT_PAGE_HT;
  center_flag = 1;
  eps_flag = 1;
  cmap_flag = 1;
  jpeg_embed_flag = 1;
  trim_flag = 1;

  /* to eliminate warnings */

  page_cnt = 0;
  i = 0;
  cmap_bit_len = 0;
  jpeg_size = 0;

  /* handle the command line arguments */

  while (arg_index < argc && argv[arg_index][0] == '-') {
    opt_str = argv[arg_index++];
    switch (opt_str[1]) {
    case 'd':	/* DPI */
      dpi = atoi (argv[arg_index++]);
      break;
    case 'r':	/* rotate */
      if (opt_str[2] == 'l') 
	rotate_layout = atoi (argv[arg_index++]);
      else if (opt_str[2] == '\0' || opt_str[2] == 'o') {
	rotate_flag = atoi (argv[arg_index++]);
	switch (rotate_flag) {
	case 1:
	  rotate_flag = 270;
	  break;
	case 2:
	  rotate_flag = 180;
	  break;
	case 3:
	  rotate_flag = 90;
	  break;
	}
      }
      break;
    case 'b':	/* bi-level */
      bil_flag = atoi (argv[arg_index++]);
      break;
    case 'f':	/* filter */
      filter_flag = atoi (argv[arg_index++]);
      break;
    case 'm':	/* multi */
      multi_flag = atoi (argv[arg_index++]);
      break;
    case 'c':	/* center */
      center_flag = atoi (argv[arg_index++]);
      break;
    case 'e':	/* eps */
      eps_flag = atoi (argv[arg_index++]);
      break;
    case 'j':	/* jpeg */
      jpeg_embed_flag = atoi (argv[arg_index++]);
      break;
    case 'g':	/* color map flag */
      cmap_flag = atoi (argv[arg_index++]);
      break;
    case 'w':	/* white space trim */
      trim_flag = atoi (argv[arg_index++]);
      break;
    case 's':	/* output to stdout */
      if (strcmp (argv[arg_index-1], "-sc") == 0) 
	scale = atoi (argv[arg_index++]);
      else
	fptr = stdout;
      break;
    case 'p':	/* page dimensions */
      if (strcmp (opt_str, "-pw") == 0) 
	page_wd = atof (argv[arg_index++]);
      else if (strcmp (opt_str, "-ph") == 0) 
	page_ht = atof (argv[arg_index++]);
      else
	arg_index = argc;
      break;
    default:
      arg_index = argc;
      break;
    }
  }

  if ((fptr == NULL && (argc - arg_index) < 2)
      || (argc - arg_index) < 1) {
    show_help (argv[0]);
    exit (1);
  }

  if (multi_flag)
    eps_flag = 0;

  in_fname  = argv[arg_index++];
  if (fptr == NULL)
    out_fname  = argv[arg_index++];
  if (eps_flag)
    center_flag = 0;

  if (multi_flag) {
    char temp_str[1024], *data_str;
    file_cnt = 0;
    fptr = fopen (in_fname, "r");
    if (fptr == NULL) {
      fprintf (stderr, "\nProblem opening file: %s\n\n", in_fname);
      exit (1);
    }
    page_cnt = 0;
    while (fgets (temp_str, 1024, fptr)) {
      data_str = fix_str (temp_str);
      if (data_str[0] == '#' || data_str[0] == '\0')
	continue;
      if (strncmp (data_str, DPI_KEYWORD, strlen (DPI_KEYWORD)) == 0) {
	dpi = atoi (strchr (data_str, '=') + 1);
      }
      else if (strncmp (data_str, SCALE_KEYWORD, strlen (SCALE_KEYWORD)) == 0) {
	scale = atoi (strchr (data_str, '=') + 1);
      }
      else {
	fname_list[file_cnt] = strdup (data_str);
	dpi_list[file_cnt] = dpi;
	scale_list[file_cnt] = scale;
	file_cnt++;
	page_cnt++;
      }
    }
    fclose (fptr);
  }
  else {
    file_cnt = 1;
    page_cnt = 1;
    fname_list[0] = strdup (in_fname);
    dpi_list[0] = dpi;
    scale_list[0] = scale;
  }

  /* open the file if needed */

  if (out_fname != NULL) {
    fptr = fopen (out_fname, "w");
    if (fptr == NULL) {
      fprintf (stderr, "\nProblem opening output file: %s\n\n", out_fname);
      exit (1);
    }
  }


  /* output the file header */

  if (!eps_flag)
    fprintf (fptr, "%%!PS-Adobe-2.0\n");
  else
    fprintf (fptr, "%%!PS-Adobe-2.0 EPSF-1.2\n");
  fprintf (fptr, "%%%%Pages: %d\n", page_cnt);
  fprintf (fptr, "%%%%PageOrder: Ascend\n");


  /* loop through the images */

  page_index = 0;
  for (file_index = 0; file_index < file_cnt; file_index++) {

    in_fname = fname_list[file_index];
    dpi = dpi_list[file_index];
    scale = scale_list[file_index];
    page_index++;

    xpos_offset = 0;
    ypos_offset = 0;
    data_top_crop = 0;
    left_crop_index = 0;
    right_crop_index = 0;

    /* load the input image file */

    ftype = get_img_filetype (in_fname);
    if (ftype == IMG_FTYPE_JPEG) {
      image = img_read_jpeg (in_fname);
      if (jpeg_embed_flag) {
	FILE *fptr;
	jpeg_size = get_flen (in_fname);
	if (jpeg_size == 0) {
	  fprintf (stderr, "Problem reading file: %s\n", in_fname);
	  exit (1);
	}
	image->data = EREALLOC (image->data, jpeg_size);
	fptr = EFOPEN (in_fname, "r");
	if (fread (image->data, jpeg_size, 1, fptr) != 1) {
	  fprintf (stderr, "Problem reading file: %s\n", in_fname);
	  fclose (fptr);
	  exit (1);
	}
	fclose (fptr);
	if (scale != 1) {
	  fprintf (stderr, "\nWarning scale factors cannot be applied to embedded JPEG's.\n");
	  fprintf (stderr, "\nUsing a scale factor of 1.\n");
	  scale = 1;
	}
      }
    }
    else if (ftype == IMG_FTYPE_GIF) {
      if (cmap_flag)
	image = LoadRawGifToImageBuffer (in_fname);
      else
	image = LoadGifToImageBuffer (in_fname);
    }

    else {
      image = img_read_file (in_fname);
    }
    if (image == NULL) { 
      fprintf (stderr, "\nProblem loading input file: %s\n\n", in_fname);
      exit (1);
    }
    if (jpeg_size == 0)
      jpeg_embed_flag = 0;

    /* rotate image */
    if (rotate_flag) {
      if (jpeg_embed_flag) {
	fprintf (stderr, "\nWarning rotations cannot be applied to embedded JPEG's.\n");
	fprintf (stderr, "\nYou may want to use the rotate layout option.\n");
	fprintf (stderr, "\nUsing a rotation of 0.\n");
	rotate_flag = 0;
      }
      else {
	img_t *new_image = img_simple_rotate (image, rotate_flag);
	img_free (image);
	image = new_image;
      }
    }

    /* get the color map if GIF */
    if (image->ftype == IMG_FTYPE_GIF && cmap_flag) {
      cmap = GetLastGifColorMap();
      if (cmap->chan_cnt == 1) {
	img_colormap_calc_used (cmap, image);
	bil_flag = img_colormap_check_used_bilevel (cmap);
	if (cmap->size > 16 || bil_flag) {
	  ApplyGifColorMap (image, cmap);
	  img_colormap_free (cmap);
	  cmap = NULL;
	  cmap_flag = 0;
	}
      }
    }
    else {
      cmap = NULL;
    }

    /* scale the image */

    if (scale > 1) {
      img_t *scaled_img;
      scaled_img = img_downscale_average (image, scale);
      img_free (image);
      image = scaled_img;
    }

    /* report some stats */

    fprintf (stderr, "\n");
    if (!rotate_layout) {
      fprintf (stderr, "Image width  = %4ld pixels, %5.2f inches\n", 
	       image->width, (float)image->width / (float)dpi);
      fprintf (stderr, "Image height = %4ld pixels, %5.2f inches\n",
	       image->height, (float)image->height / (float)dpi);
    }
    else {
      fprintf (stderr, "Image width  = %4ld pixels, %5.2f inches\n",
	       image->height, (float)image->height / (float)dpi);
      fprintf (stderr, "Image height = %4ld pixels, %5.2f inches\n", 
	       image->width, (float)image->width / (float)dpi);
    }
    fprintf (stderr, "\n");

    /* get the data width and height */

    data_width  = image->width;
    data_height = image->height;
    read_width  = image->width;


    /* check for blank space at the top and
     * bottom and shorten the image accordingly 
     */

    if (image->chan_cnt == 1 && !jpeg_embed_flag) {

      int blank_val;
      pix_val_t color[3] = {255, 255, 255};
      if (bil_flag) {
	thresh_image (image, THRESH);
      }
      if (cmap) {
	blank_val = img_colormap_find (cmap, color);
      }
      else {
	blank_val = 255;
      }

      if (trim_flag && blank_val >= 0) { 
	blank_adjust (image, filter_flag, &read_width,
		      &left_crop_index, &right_crop_index,
		      &data_top_crop, &data_width, &data_height,
		      &xpos_offset, &ypos_offset, blank_val);
      }
      if (bil_flag) {
	/* make sure we are a multiple of 8 */
	data_width = (read_width + 7) / 8 * 8;
      }
      else {
	data_width = read_width;
      }
    }


    /* output the image header */

    if (eps_flag) {
      if (!rotate_layout)
	fprintf (fptr, "%%%%BoundingBox: %ld %ld %ld %ld\n",
		 0L, 
		 (long)floor ((page_ht
			       - (double)image->height / (double)dpi) * 72.0), 
		 (long)ceil ((double)image->width / (double)dpi * 72.0), 
		 (long)ceil (page_ht * 72.0)
		 );
      else
	fprintf (fptr, "%%%%BoundingBox: %ld %ld %ld %ld\n",
		 0L, 
		 (long)floor ((page_wd
			       - (double)image->width / (double)dpi) * 72.0), 
		 (long)ceil ((double)image->height / (double)dpi * 72.0), 
		 (long)ceil (page_wd * 72.0)
		 );
    }
    fprintf (fptr, "%%%%EndComments\n");
    fprintf (fptr, "\n");
    {
      char  date_str[64];
      time_t time_data;
      struct tm *tm_data;
      time (&time_data);
      tm_data = localtime (&time_data);
      strftime (date_str, 64, "%Y/%m/%d at %H:%M (%Z)", tm_data);
      fprintf (fptr, "%%  Image source file: %s\n", in_fname);
      fprintf (fptr, "%%\n");
      fprintf (fptr, "%% Conversion details: performed by img2ps on %s\n", date_str);
      fprintf (fptr, "%%\n");
    }
    fprintf (fptr,   "%%    Program details: img2ps written by Charles Rosenberg (chuck@cs.cmu.edu)\n");
    fprintf (fptr,   "%%            Version: %s\n", VERSION_STR);
    fprintf (fptr, "\n");
    if (!eps_flag) {
      fprintf (fptr, "\n");
      fprintf (fptr, "%%%%Page: %d %d\n", page_index, page_index);
    }
    fprintf (fptr, "\n");
    fprintf (fptr, "/origstate save def\n");
    fprintf (fptr, "\n");
    fprintf (fptr, "%% Some interpreters don't skip the ASCII85 EOD marker.\n");
    fprintf (fptr, "%% This function lets them skip over it.\n");
    fprintf (fptr, "\n");
    fprintf (fptr, "/~ {currentfile read pop pop} def\n");
    fprintf (fptr, "\n");
    fprintf (fptr, "/image_wd    %ld def\n", data_width);
    fprintf (fptr, "/image_ht    %ld def\n", data_height);
    fprintf (fptr, "/pos_wd      %ld def\n", image->width);
    fprintf (fptr, "/pos_ht      %ld def\n", image->height);
    fprintf (fptr, "/image_dpi   %ld def\n", dpi);
    fprintf (fptr, "/image_scale 72 image_dpi div def\n");
    fprintf (fptr, "/image_chan  %d def\n", image->chan_cnt);
    fprintf (fptr, "/xpos_offset %ld image_scale mul def\n", xpos_offset);
    fprintf (fptr, "/ypos_offset %ld image_scale mul def\n", ypos_offset);
    fprintf (fptr, "/pix_buf_size %ld def\n", data_width * image->chan_cnt /
	     (bil_flag ? 8 : 1));
    fprintf (fptr, "\n");

    if (rotate_layout) {
      fprintf (fptr, "/page_ht %4g 72 mul def\n", page_wd);
      fprintf (fptr, "/page_wd %4g 72 mul def\n", page_ht);
      fprintf (fptr, "90 rotate\n");
      fprintf (fptr, "0 page_ht neg translate\n");
    }
    else {
      fprintf (fptr, "/page_ht %4g 72 mul def\n", page_ht);
      fprintf (fptr, "/page_wd %4g 72 mul def\n", page_wd);
    }


    if (center_flag) {
      fprintf (fptr, 
	       "/image_xpos page_wd pos_wd image_scale mul sub 2 div def\n");
      fprintf (fptr, 
	       "/image_ypos page_ht pos_ht image_scale mul sub 2 div def\n");
    }
    else {
      fprintf (fptr, 
	       "/image_xpos 0 def\n");
      fprintf (fptr, 
	       "/image_ypos page_ht pos_ht image_scale mul sub def\n");
    }
    fprintf (fptr, "image_xpos xpos_offset add image_ypos ypos_offset add translate\n");

    fprintf (fptr, "\n");
    fprintf (fptr, "/pix pix_buf_size string def\n");
    fprintf (fptr, "image_wd image_scale mul image_ht image_scale mul scale\n");

    fprintf (fptr, "\n");
    if (cmap == NULL) {
      if (bil_flag)
	fprintf (fptr, "image_wd image_ht 1\n");
      else
	fprintf (fptr, "image_wd image_ht 8\n");
      fprintf (fptr, "[image_wd 0 0 image_ht neg 0 image_ht]\n");
    }
    if (image->chan_cnt == 1) {

      /* monochrome */
      if (cmap == NULL) {
	fprintf (fptr, "currentfile\n");
	fprintf (fptr, "/ASCII85Decode filter\n");
	if (jpeg_embed_flag)
	  fprintf (fptr, "<< >> /DCTDecode filter\n");
	else
	  fprintf (fptr, "/RunLengthDecode filter\n");
	fprintf (fptr, "image\n");
      }

      /* color map */
      else {

	if      (cmap->size <=  2) cmap_bit_len = 1;
	else if (cmap->size <=  4) cmap_bit_len = 2;
	else if (cmap->size <= 16) cmap_bit_len = 4;
	else cmap_bit_len = 8;

	if (cmap->chan_cnt == 1) {
	  fprintf (fptr, "[/Indexed /DeviceGray %d\n", cmap->size-1);
	  fprintf (fptr, "< ");
	  for (i = 0; i < cmap->size; i++) {
	    fprintf (fptr, "%02X ", cmap->data[i]);
	    if ((i+1) % 25 == 0)
	      fprintf (fptr, "\n  ");
	  }
	}
	else {
	  fprintf (fptr, "[/Indexed /DeviceRGB %d\n", cmap->size-1);
	  fprintf (fptr, "< ");
	  for (i = 0; i < cmap->size; i++) {
	    fprintf (fptr, "%02X%02X%02X ", 
		     cmap->data[i*3], cmap->data[i*3+1], 
		     cmap->data[i*3+2]);
	    if ((i+1) % 10 == 0)
	      fprintf (fptr, "\n  ");
	  }
	}
	fprintf (fptr, ">\n");
	fprintf (fptr, "] setcolorspace\n");
	fprintf (fptr, "<<\n");
	fprintf (fptr, "  /ImageType 1\n");
	fprintf (fptr, "  /Width  image_wd\n");
	fprintf (fptr, "  /Height image_ht\n");
	fprintf (fptr, "  /BitsPerComponent %d\n", cmap_bit_len);
	fprintf (fptr, "  /Decode [0 %d]\n", (1 << cmap_bit_len) - 1);
	fprintf (fptr, "  /ImageMatrix ");
	fprintf (fptr, "[image_wd 0 0 image_ht neg 0 image_ht]\n");
	fprintf (fptr, "  /DataSource currentfile\n"); 
	fprintf (fptr, "       /ASCII85Decode filter\n");
	if (jpeg_embed_flag)
	  fprintf (fptr, "       << >> /DCTDecode filter\n");
	else
	  fprintf (fptr, "       /RunLengthDecode filter\n");
	fprintf (fptr, ">>\n");
	fprintf (fptr, "image\n");
      }
    }
    else {
      fprintf (fptr, "currentfile\n");
      fprintf (fptr, "/ASCII85Decode filter\n");
      if (jpeg_embed_flag)
	fprintf (fptr, "<< >> /DCTDecode filter\n");
      else
	fprintf (fptr, "/RunLengthDecode filter\n");
      fprintf (fptr, "false 3\n");
      /* fprintf (fptr, "{currentfile pix readhexstring pop} false 3\n"); */
      fprintf (fptr, "colorimage\n");
    }


    /* output the image data */

    data = image->data + data_top_crop * image->width * image->chan_cnt;

    /* 1 bit per pixel monochrome */

    if (bil_flag && image->chan_cnt == 1) {

      img_dim_t x, y;
      unsigned long out_size, write_size, run_size;
      int out_val;
      unsigned char *out_data, *out_ptr, *write_data, *run_data;

      out_size = (read_width + 7) / 8 * data_height;
      out_data = (unsigned char *)emalloc (out_size, "out_data");
      write_data = NULL; /* avoid warnings */

      out_val = 0;
      out_ptr = out_data;

      for (y = 0; y < data_height; y++) {
	data+= left_crop_index;
	for (x = 0; x < read_width; x++) {
	  out_val = (out_val << 1) | (*data++ > THRESH ? 1 : 0);
	  if (x % 8 == 7) {
	  last_one:
	    *out_ptr++= out_val;
	  }
	}
	if (x % 8 != 0) {
	  out_val = (out_val << (8 - (x % 8))) | (0xff >> (x % 8));
	  x+= (8 - (x % 8)) - 1;
	  goto last_one;
	}
	data+= (image->width - right_crop_index);
      }
      runlength_encode (out_data, out_size, &run_data, &run_size);
      free (out_data);
      ascii85_encode   (run_data, run_size, &write_data, &write_size);
      fwrite (write_data, write_size, 1, fptr);
      free (run_data);
      free (write_data);
    }

    /* 8 bits per pixel - monochrome or color mapped */

    else if (image->chan_cnt == 1) {

      if (left_crop_index > 0 || right_crop_index < image->width) {
	crop_image (data, image->width, data_height, 
		    left_crop_index, right_crop_index);
      }

      /* monochrome */
      if (cmap == NULL || cmap_bit_len == 8) {
	unsigned long out_size, write_size, run_size;
	unsigned char *write_data, *run_data;
	
	if (!jpeg_embed_flag) {
	  out_size = data_width * data_height;
	  runlength_encode (data, out_size, &run_data, &run_size);
	  ascii85_encode   (run_data, run_size, &write_data, &write_size);
	  fwrite (write_data, write_size, 1, fptr);
	  free (run_data);
	  free (write_data);
	}
	else {
	  ascii85_encode (data, jpeg_size, &write_data, &write_size);
	  fwrite (write_data, write_size, 1, fptr);
	  free (write_data);
	}
      }

      /* color map */
      else {
	unsigned long out_size, bit_size, write_size, run_size;
	unsigned char *bit_data, *write_data, *run_data;
	
	out_size = data_width * data_height;
	
	bitpack_encode (data, out_size, &bit_data, &bit_size, 
			cmap_bit_len, data_width);
	runlength_encode (bit_data, bit_size, &run_data, &run_size);
	ascii85_encode   (run_data, run_size, &write_data, &write_size);
	fwrite (write_data, write_size, 1, fptr);
	free (bit_data);
	free (run_data);
	free (write_data);
      }
    }

    /* 24 bits per pixel - color */

    else {
      unsigned long write_size, run_size;
      unsigned char *write_data, *run_data;

      if (!jpeg_embed_flag) {
	pixel_cnt = data_width * data_height * image->chan_cnt;
	runlength_encode (data, pixel_cnt, &run_data, &run_size);
	ascii85_encode   (run_data, run_size, &write_data, &write_size);
	fwrite (write_data, write_size, 1, fptr);
	free (run_data);
	free (write_data);
      }
      else {
	ascii85_encode (data, jpeg_size, &write_data, &write_size);
	fwrite (write_data, write_size, 1, fptr);
	free (write_data);
      }
    }


    /* output the image trailer */

    fprintf (fptr, "\n");
    if (!eps_flag) {
      fprintf (fptr, "showpage\n");
      fprintf (fptr, "\n");
    }
    fprintf (fptr, "origstate restore\n");

    img_free (image);
    if (cmap)
      img_colormap_free (cmap);

  }

  fprintf (fptr, "\n");
  fprintf (fptr, "%%%%Trailer\n");
  fprintf (fptr, "\n");
  fprintf (fptr, "%%%%EOF\n");
  fprintf (fptr, "\n");
  
  /* fprintf (fptr, "%c", 4); */
  
  fflush (fptr);

  if (out_fname != NULL)
    fclose (fptr);

  /* exit */

  exit (0);
}

/* encode a buffer of binary data into a buffer hexadecimal data 
 *
 * allocates and returns a pointer to a new buffer
 */

void hex_encode (unsigned char *in_data,   unsigned long in_len, 
		 unsigned char **out_data, unsigned long *out_len)
{
  unsigned long i, out_cnt;
  unsigned char *out_ptr;

  *out_len = in_len * 2 + in_len / (CHARS_PER_LINE / 2) + 1;
  *out_data = (unsigned char *)emalloc (*out_len, "hex_encode");
  out_ptr = *out_data;

  out_cnt = 0;
  for (i = 0; i < in_len; i++) {
    *out_ptr++ = (unsigned char)hex_digits[*in_data >> 4];
    *out_ptr++ = (unsigned char)hex_digits[*in_data++ & 0x0F];
    out_cnt+=2;
    if (i % (CHARS_PER_LINE / 2) == (CHARS_PER_LINE / 2 - 1)) {
      *out_ptr++ = '\n';
      out_cnt++;
    }
  }
  *out_len = out_cnt;
  return;
}


/* encode a buffer of binary data into ascii 85 code
 *
 * allocates and returns a pointer to a new buffer
 *
 * added fix because latex doesn't like %! as the beginning
 * of a line
 */

void ascii85_encode (unsigned char *in_data,   unsigned long in_len, 
		     unsigned char **out_data, unsigned long *out_len)
{
  unsigned long i, j, k, out_cnt, out_val, alloc_len;
  unsigned char *out_ptr, out_char;
  unsigned long val[5];

  alloc_len = (in_len / 4 + 1) * 5;
  alloc_len+= alloc_len / (CHARS_PER_LINE-1) + 1 + 4;
  *out_data = (unsigned char *)emalloc (alloc_len, "ascii85_encode");
  out_ptr = *out_data;

  out_cnt = 0;
  k = 0;

  for (i = 0; i < in_len;) {
    
    for (j = 0; j < 4; j++, i++) {
      if (i < in_len)
	val[j] = (unsigned long)*in_data++;
      else
	val[j] = 0L;
    }
    out_val = (val[0] << 24) | (val[1] << 16) | (val[2] << 8) | val[3];
    if (out_val == 0 && i <= in_len) {
      *out_ptr++ = 'z';
      out_cnt++;
      if (++k >= CHARS_PER_LINE) {
	*out_ptr++ = '\n';
	out_cnt++;
	k = 0;
      }
    }
    else {
      for (j = 0; j < 5; j++) {
	val[4-j] = out_val % 85;
	out_val/= 85;
      }
      for (j = 0; j < 5; j++) {
	out_char = (unsigned char)val[j] + 33;
	if (out_char == '%') {
	  if (out_cnt == 0 && j == 0) {
	    *out_ptr++ = ' ';
	    out_cnt++;
	  }
	  else if (*(out_ptr-1) == '\n') {
	    *(out_ptr-1) = out_char;
	    out_char = '\n';
	  }
	}
	*out_ptr++ = out_char;
	if (++k >= CHARS_PER_LINE) {
	  *out_ptr++ = '\n';
	  out_cnt++;
	  k = 0;
	}
      }
      out_cnt+= 5;
    }
  }

  /* add the end of data marker */

  *out_ptr++ = '\n';
  *out_ptr++ = '~';
  *out_ptr++ = '>';
  *out_ptr++ = '\n';
  *out_len = out_cnt + 4; 

  return;
}


/* encode a buffer of binary data with a run length code
 *
 * allocates and returns a pointer to a new buffer
 */

void runlength_encode (unsigned char *in_data,   unsigned long in_len, 
		       unsigned char **out_data, unsigned long *out_len)
{
  unsigned long i, j, out_cnt, alloc_len;
  unsigned char *out_ptr, run_val, *data_ptr, *in_ptr;
  unsigned long verbatim_len, run_len;

  /* set up the output buffer */

  alloc_len = in_len + (in_len + 127) / 128 + 1;
  *out_data = (unsigned char *)emalloc (alloc_len, "runlength_encode");
  out_ptr = *out_data;

  /* initialize the variables */
  out_cnt = 0;
  verbatim_len = 1;	/* the length of the verbatim data block */
  run_len = 0;		/* the length of the current run */
  data_ptr = in_data;	/* the start of the current verbatim data block */
  in_ptr = in_data + 1;	/* our current position in the data stream */
  run_val = *in_data;	/* the byte value of the run */

  /* iterate over all the data */

  for (i = 1; i < in_len; i++) {
    
    if (*in_ptr == run_val) {
      if (run_len == 0) {
	run_len = 1;	  
	if (verbatim_len != 0) {
	  verbatim_len--;  /* remove the previous one from the verb cnt */
	  run_len++;
	}
      }
      else {

	run_len++;

	/* output the pending verbatim data */
	if (verbatim_len != 0) {
	  *out_ptr++ = verbatim_len - 1;  /* flag the verbatim data length */
	  for (j = 0; j < verbatim_len; j++)
	    *out_ptr++ = *data_ptr++;
	  out_cnt+= verbatim_len + 1;
	  verbatim_len = 0;
	}

	/* output the pending run */
	if (run_len >= 128) {
	  *out_ptr++ = 129;	/* a run 128 bytes long */
	  *out_ptr++ = run_val;
	  out_cnt+= 2;
	  run_len-= 128;
	  data_ptr+= 128;
	}
      }
    }
    else {

      /* output the pending run */
      if (run_len > 2 || (run_len > 1 && verbatim_len == 0)) {
	*out_ptr++ = 257 - run_len;
	*out_ptr++ = run_val;
	out_cnt+= 2;
	run_len = 0;
	data_ptr = in_ptr;
      }
      verbatim_len+= 1 + run_len;
      run_len = 0;

      /* output a verbatim run */
      if (verbatim_len >= 128) {
	*out_ptr++ = 127;	/* flag 128 bytes of verbatim data */
	for (j = 0; j < 128; j++)
	  *out_ptr++ = *data_ptr++;
	out_cnt+= 129;
	verbatim_len-= 128;
      }
    }
    run_val = *in_ptr++;
  }

  /* finish off the left over data */
  if (verbatim_len > 0) {
    *out_ptr++ = verbatim_len - 1;  /* flag the verbatim data length */
    for (j = 0; j < verbatim_len; j++)
      *out_ptr++ = *data_ptr++;
    out_cnt+= verbatim_len + 1;
  }
  if (run_len > 0) {
    *out_ptr++= 257 - run_len;
    *out_ptr++ = run_val;
    out_cnt+= 2;
  }

  /* end of data marker */  
  *out_ptr++ = 128;
  *out_len = out_cnt + 1; 

  return;
}


/* pack a buffer of less than 8 bit values stored in bytes
 * to a butstream
 *
 * allocates and returns a pointer to a new buffer
 *
 * bit_len must be 1, 2, 4
 */

void bitpack_encode (unsigned char *in_data,   unsigned long in_len, 
		     unsigned char **out_data, unsigned long *out_len,
		     int bit_len, img_dim_t data_width)
{
  unsigned long i;
  unsigned char *out_ptr, out_val, bit_mask;
  int j, blk_cnt;
  img_dim_t x;

  /* set up the output buffer */

  *out_len = (in_len * bit_len + 7) / 8;
  *out_data = out_ptr = (unsigned char *)EMALLOC (*out_len);

  /* initialize the variables */

  bit_mask = 0xFF >> (8 - bit_len);
  blk_cnt = 8 / bit_len;

  /* iterate over all the data */

  x = 0;
    for (i = 0; i < in_len; i+= blk_cnt) {
    out_val = 0;
    for (j = 0; j < blk_cnt; j++, x++) {
      if (x < data_width)
	out_val = (out_val << bit_len) | (*in_data++ & bit_mask);
      else
	out_val = (out_val << bit_len);
    }
    if (x >= data_width)
      x = 0;
    *out_ptr++ = out_val;
  }

  return;
}


/* 
 * for 1 bit per pixel monochrome images 
 * check for blank space at the top and
 * bottom and shorten the image accordingly 
 */

void blank_adjust (img_t *image, int filter_flag, img_dim_t *rw, 
		   img_dim_t *lci, img_dim_t *rci, img_dim_t *dtc,
		   img_dim_t *dw, img_dim_t *dh, 
		   img_dim_t *xpo, img_dim_t *ypo,
		   pix_val_t blank_val)
{
  img_dim_t x, y;
  pix_val_t *data;
  img_dim_t left_crop_index, right_crop_index;
  img_dim_t data_height, data_width, read_width;
  img_dim_t xpos_offset, ypos_offset, data_top_crop;

  /* get the values */
  read_width = *rw;
  left_crop_index = *lci;
  right_crop_index = *rci;
  data_top_crop = *dtc;
  data_width = *dw;
  data_height = *dh;
  xpos_offset = *xpo;
  ypos_offset = *ypo;

  /* remove isolated pixels */
  if (filter_flag) {
    data = image->data;
    for (y = 0; y < image->height; y++) {
      for (x = 0; x < image->width; x++, data++) {
	if (*data != blank_val) {
	  if ((x == 0 || *(data-1) == blank_val)
	      && (x == image->width-1 || *(data+1) == blank_val)
	      && (y == 0 || *(data-image->width) == blank_val)
	      && (y == image->height-1 || *(data+image->width) == blank_val)) {
	    *data = blank_val;
	  }
	}
      }
    }
  }

  /* first the bottom */
  data = image->data + image->height * image->width - 1;
  for (y = 0; y < image->height; y++) {
    for (x = 0; x < image->width; x++) {
      if (*data-- != blank_val) {
	goto bottom_check_done;
      }
    }
    data_height--;
  }
 bottom_check_done:
  ypos_offset = image->height - data_height;

  /* now the top */
  data = image->data;
  for (y = 0; y < image->height; y++) {
    for (x = 0; x < image->width; x++) {
      if (*data++ != blank_val) {
	goto top_check_done;
      }
    }
    data_height--;
    data_top_crop++;
  }
 top_check_done:

      /* now the width */
  left_crop_index  = image->width - 1;
  right_crop_index = 0;
  for (y = 0; y < image->height; y++) {

    /* check left */
    data = image->data + y * image->width;
    for (x = 0; x < image->width; x++) {
      if (*data++ != blank_val) {
	break;
      }
    }
    if (x < left_crop_index) {
      left_crop_index = x;
    }

    /* check right */
    data = image->data + (y + 1) * image->width - 1;
    for (x = 0; x < image->width; x++) {
      if (*data-- != blank_val) {
	break;
      }
    }
    x = image->width - x;
    if (x > right_crop_index) {
      right_crop_index = x;
    }
  }
  read_width  = right_crop_index - left_crop_index;
  xpos_offset = left_crop_index;

  /* return the new values */
  *rw = read_width;
  *lci = left_crop_index;
  *rci = right_crop_index;
  *dtc = data_top_crop;
  *dw = data_width;
  *dh = data_height;
  *xpo = xpos_offset;
  *ypo = ypos_offset;
}


/* 
 * for 1 bit per pixel monochrome images stored as bytes
 * 
 * threshold the pixel values
 */

void thresh_image (img_t *image, pix_val_t thresh)
{
  img_dim_t x, y;
  pix_val_t *data;

  data = image->data;
  for (y = 0; y < image->height; y++) {
    for (x = 0; x < image->width; x++, data++) {
      *data = (*data <= THRESH) ? 0 : 255;
    }
  }
}

/* 
 * for 1 byte per pixel images 
 * 
 * re-arrange data for horizontal cropping
 */

void crop_image (pix_val_t *data_in, img_dim_t width, img_dim_t height,
		 img_dim_t left_crop_index, img_dim_t right_crop_index)
{
  img_dim_t x, y;
  pix_val_t *data_out;

  data_out = data_in;
  for (y = 0; y < height; y++) {
    data_in+= left_crop_index;
    for (x = left_crop_index; x < right_crop_index; x++) {
      *data_out++ = *data_in++;
    }
    data_in+= (width - right_crop_index);
  }
}




/* 
 * show the help info 
 */

void show_help (char *cmd_name)
{
  char *show_name = strrchr (cmd_name, '/');
  if (show_name == NULL)
    show_name = cmd_name;
  else
    show_name++;
  fprintf (stdout, "\n");
  fprintf (stdout, "Convert a PGM, PPM, GIF or JPEG format image to EPS or PS for printing.\n");
  fprintf (stdout, "\n");
    
  fprintf (stdout, "Usage: %s [opt] <in_fname> <out_fname>\n", show_name);
  fprintf (stdout, "\n");
  fprintf (stdout, "Arguments\n");
  fprintf (stdout, "   <in_fname>  = input  image file or file list (in multi-mode)\n");
  fprintf (stdout, "   <out_fname> = output postscript file\n");
  fprintf (stdout, "\n");
  fprintf (stdout, "Options\n");
  fprintf (stdout, "   -d  <int>  = dpi, default = %d\n", DEFAULT_DPI);
  fprintf (stdout, "   -s         = output data to stdout instead of a file\n");
  fprintf (stdout, "   -r  <int>  = rotate image, default = 0 (values are 0, 90, 180, 270)\n");
  fprintf (stdout, "   -c  <bool> = center flag, default = 1\n");
  fprintf (stdout, "   -e  <bool> = eps flag, default = 1\n");
  fprintf (stdout, "   -b  <bool> = bi-level mode, default = 0\n");
  fprintf (stdout, "   -f  <bool> = single pixel filter mode, default = 0\n");
  fprintf (stdout, "   -m  <bool> = multiple image mode, default = 0\n");
  fprintf (stdout, "   -sc <int>  = scale down factor, default = 1\n");
  fprintf (stdout, "   -pw <dbl>  = page width in inches, default = %g\n", DEFAULT_PAGE_WD);
  fprintf (stdout, "   -ph <dbl>  = page height in inches, default = %g\n", DEFAULT_PAGE_HT);
  fprintf (stdout, "   -g  <bool> = handle all gifs as color mapped images, default = 1\n");
  fprintf (stdout, "   -j  <bool> = embed jpegs as encoded data, default = 1\n");
  fprintf (stdout, "   -wt <bool> = trim bordering white space from image, default = 1\n");
  fprintf (stdout, "   -rl <bool> = rotate page layout, default = 0\n");
  fprintf (stdout, "\n");
  fprintf (stdout, "Note\n");
  fprintf (stdout, "   Color mapped images are only supported with GIF's.\n");
  fprintf (stdout, "\n");
  fprintf (stdout, "Written by Charles Rosenberg (chuck+@cs.cmu.edu)\n");
  fprintf (stdout, "\n");
  fprintf (stdout, "Version: %s\n", VERSION_STR);
  fprintf (stdout, "\n");
}
