/*%********************************************************************
 *
 * File: img_util.c
 *
 * Author:  Charles Rosenberg 
 * Created: May 11, 1998
 * 
 * Image processing utilities
 *
 * Modification History:
 *
 * 05/11/98, cjr: Created.
 * 09/15/99, cjr: changed img_hist_norm to img_hist_range_adjust
 * 09/15/99, cjr: added img_simple_normalize
 * 10/28/99, cjr: added some support for jpeg
 *
 **********************************************************************
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "image.h"
#include "img_util.h"
#include "gif.h"
#include "pnm.h"
#include "util.h"

/* local function prototypes */

void img_hist_range_adjust2_internal (img_t *img, int black_adj_flag);
void img_hist_range_adjust3_internal (img_t *img, int black_adj_flag);
void img_hist_range_adjust4_internal (img_t *img, int black_adj_mode,
				      int fast_flag, int fix_flag);
void img_hist_range_adjust5_internal (img_t *img, int black_adj_flag);
void img_hist_range_adjust6_internal (img_t *img, int black_adj_mode,
				      int fast_flag, int fix_flag,
				      int quant_flag);


/* function definitions */



/*
 * Does a simple high pass filter of an
 * image by removing 3x3 pixel means
 *
 * A new image is created with the filtered data.
 *
 * Differences greater than +/- 127 are truncated.
 * A difference of zero is mapped to 128, +127 to 255, and -127 to 1.
 */

img_t *img_simple_high_pass (img_t *src_img)
{

  img_t         *new_img;
  img_dim_t      width, height, x, y;
  pix_val_t     *new_data, *src_data;
  int            sum, pix_val, mx, px, my, py;

  /* create a new image */

  width  = src_img->width;
  height = src_img->height;
  new_img = img_create_basis (src_img, width, height, src_img->chan_cnt);
  if (new_img == NULL)
    return (NULL);

  /* high pass filter the image */

  src_data = src_img->data;
  new_data = new_img->data;
  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      if (x > 0) 
	mx = -1;
      else
	mx = +1;
      if (x < width-1)
	px = +1;
      else
	px = -1;
      if (y > 0) 
	my = -width;
      else
	my = +width;
      if (y < height-1)
	py = +width;
      else
	py = -width;

      sum = 0
	+ (int)*(src_data+my+mx)
	+ (int)*(src_data+my)
	+ (int)*(src_data+my+px)
	+ (int)*(src_data+mx)
	+ (int)*src_data 
	+ (int)*(src_data+px)
	+ (int)*(src_data+py+mx)
	+ (int)*(src_data+py)
	+ (int)*(src_data+py+px);
      pix_val = (int)*src_data - sum / 9 + 128;
      if (pix_val < 0) 
	pix_val = 0;
      else if (pix_val > 255)
	pix_val = 255;
      *new_data++ = pix_val;
      src_data++;
    }
  }
    
  return (new_img);

}



/*
 * A simple histogram based equalization of an image.
 *
 * A point transformation of the image such that 50% of the
 * pixel mass is in the range of 96 to 160.
 *
 * Assumes that the image mean is 128, so only scales the data,
 * does not shift it.
 *
 * The original data is modified.
 */

void img_simple_stretch (img_t *img)
{
  long        hist[256], sum;
  pix_val_t   xform[256];
  img_dim_t   width, height, x, y, pix_cnt;
  pix_val_t  *img_data;
  int         i, new_val, chan, chan_cnt;
  float       scale;

  /* loop through the color channels */

  chan_cnt = img->chan_cnt;

  for (chan = 0; chan < chan_cnt; chan++) {

    /* initialize */

    width   = img->width;
    height  = img->height;
    pix_cnt = width * height;
    for (i = 0; i < 256; i++)
      hist[i] = 0;

    /* histogram the image */
  
    img_data = img->data + chan;
    for (y = 0; y < height; y++) 
      for (x = 0; x < width; x++, img_data+= chan_cnt) 
	hist[*img_data]++;

    /* figure out the scaling factor */

    sum = hist[128];
    i = 0;
    if (sum < pix_cnt / 2) {
      for (i = 1; i <=127; i++) {
	sum+= hist[128+i] + hist[128-i];
	if (sum >= pix_cnt / 2)
	  break;
      }
    }
    if (i == 128)
      return;

    /* build the xform table */

    scale = 8.0 / (float)(i+1);
    for (i = 0; i < 256; i++) {
      new_val = (int)((float)(i - 128) * scale) + 128;
      if (new_val < 0)
	new_val = 0;
      if (new_val > 255)
	new_val = 255;
      xform[i] = new_val;
    }

    /* xform the image */
  
    img_data = img->data + chan;
    for (y = 0; y < height; y++) 
      for (x = 0; x < width; x++, img_data+= chan_cnt) 
	*img_data = xform[*img_data];
  }

  return;
}


/*
 * A simple histogram based normalization of an image.
 *
 * A point transformation of the image such that 50% of the
 * pixel mass is in the range of 128 +/- range.
 *
 * A typical value for the range could be 64
 *
 * Shifts the image mean to 128.
 *
 * The original data is modified.
 */

void img_simple_normalize (img_t *img, int range)
{
  long        hist[256], sum;
  pix_val_t   xform[256];
  img_dim_t   width, height, x, y, pix_cnt;
  pix_val_t  *img_data;
  int         i, new_val, mean, chan, chan_cnt;
  double      scale;

  /* loop through the color channels */

  chan_cnt = img->chan_cnt;

  for (chan = 0; chan < chan_cnt; chan++) {

    /* initialize */

    width   = img->width;
    height  = img->height;
    pix_cnt = width * height;
    for (i = 0; i < 256; i++)
      hist[i] = 0;

    /* histogram the image */
  
    sum = 0;
    img_data = img->data + chan;
    for (y = 0; y < height; y++) 
      for (x = 0; x < width; x++) {
	sum+= (long)*img_data;
	hist[*img_data]++;
	img_data+= chan_cnt;
      }

    /* figure out the image mean */
    mean = (int)((double)sum / (double)pix_cnt + 0.5);

    /* figure out the scaling factor */

    sum = hist[mean];
    i = 1;
    while (sum < pix_cnt / 2) {
      sum+= hist[mean+i] + hist[mean-i];
      i++;
    }
    if (i == 1)
      return;

    /* build the xform table */

    scale = (float)range / (double)i;
    for (i = 0; i < 256; i++) {
      new_val = (int)((double)(i - mean) * scale) + 128;
      if (new_val < 0)
	new_val = 0;
      if (new_val > 255)
	new_val = 255;
      xform[i] = new_val;
    }

    /* xform the image */
  
    img_data = img->data + chan;
    for (y = 0; y < height; y++) 
      for (x = 0; x < width; x++, img_data+= chan_cnt) 
	*img_data = xform[*img_data];
  }

  return;
}


/*
 * A simple histogram based normalization of an image.
 *
 * A point transformation of the image such that 50% of the
 * pixel mass is in the range of 128 +/- range.
 *
 * A typical value for the range could be 64
 *
 * This version DOES NOT shift the image mean.
 *
 * The original data is modified.
 */

void img_simple_normalize2 (img_t *img, int range)
{
  long        hist[256], sum;
  pix_val_t   xform[256];
  img_dim_t   width, height, x, y, pix_cnt;
  pix_val_t  *img_data;
  int         i, new_val, mean, chan, chan_cnt;
  double      scale;

  /* loop through the color channels */

  chan_cnt = img->chan_cnt;

  for (chan = 0; chan < chan_cnt; chan++) {

    /* initialize */

    width   = img->width;
    height  = img->height;
    pix_cnt = width * height;
    for (i = 0; i < 256; i++)
      hist[i] = 0;

    /* histogram the image */
  
    sum = 0;
    img_data = img->data + chan;
    for (y = 0; y < height; y++) 
      for (x = 0; x < width; x++) {
	sum+= (long)*img_data;
	hist[*img_data]++;
	img_data+= chan_cnt;
      }

    /* figure out the image mean */
    mean = (int)((double)sum / (double)pix_cnt + 0.5);

    /* figure out the scaling factor */

    sum = hist[mean];
    i = 1;
    while (sum < pix_cnt / 2) {
      sum+= hist[mean+i] + hist[mean-i];
      i++;
    }
    if (i == 1)
      return;

    /* build the xform table */

    scale = (float)range / (double)i;
    for (i = 0; i < 256; i++) {
      new_val = (int)((double)(i - mean) * scale) + mean;
      if (new_val < 0)
	new_val = 0;
      if (new_val > 255)
	new_val = 255;
      xform[i] = new_val;
    }

    /* xform the image */
  
    img_data = img->data + chan;
    for (y = 0; y < height; y++) 
      for (x = 0; x < width; x++, img_data+= chan_cnt) 
	*img_data = xform[*img_data];
  }

  return;
}


/*
 * A simple histogram based normalization of an image.
 *
 * A point transformation of the image such that 
 * the variance of all of the channels becomes
 * the average of the variance of all three channels.
 *
 * This version DOES NOT shift the image mean.
 *
 * This version is a NOP for monochrome data.
 *
 * The original data is modified.
 */

void img_simple_normalize3 (img_t *img)
{
  long        hist[256], sum;
  pix_val_t   xform[256];
  img_dim_t   width, height, x, y, pix_cnt;
  pix_val_t  *img_data;
  int         i, new_val, chan, chan_cnt;
  double      scale, sum2, avg_var;
  double      mean[3], var[3];

  chan_cnt = img->chan_cnt;
  width   = img->width;
  height  = img->height;
  pix_cnt = width * height;

  if (chan_cnt == 1) {
    fprintf (stderr, "This algorithm has no effect on monochrome images.\n");
    return;
  }
  if (chan_cnt > 3) {
    fprintf (stderr, "This algorithm has only handle 3 channel data.\n");
    exit (1);
  }

  /* loop through the color channels */

  avg_var = 0.0;
  for (chan = 0; chan < chan_cnt; chan++) {

    /* initialize */

    for (i = 0; i < 256; i++)
      hist[i] = 0;

    /* histogram the image */
  
    sum = 0;
    img_data = img->data + chan;
    for (y = 0; y < height; y++) 
      for (x = 0; x < width; x++) {
	sum+= (long)*img_data;
	hist[*img_data]++;
	img_data+= chan_cnt;
      }

    /* figure out the image mean */
    mean[chan] = (double)sum / (double)pix_cnt;

    /* figure out the variance */
    sum2 = 0.0;
    for (i = 0; i < 256; i++)
      sum2+= (double)hist[i] * (double)i * (double)i;
    var[chan] = (double)sum2 / (double)pix_cnt - mean[chan] * mean[chan];
    avg_var+= var[chan];
  }

  /* calculate the average variance */

  avg_var/= (double)chan_cnt;

  for (chan = 0; chan < chan_cnt; chan++) {

    /* build the xform table */

    scale = sqrt (avg_var) / sqrt (var[chan]);
    for (i = 0; i < 256; i++) {
      new_val = (int)(((double)i - mean[chan]) * scale + mean[chan] + 0.5);
      if (new_val < 0)
	new_val = 0;
      if (new_val > 255)
	new_val = 255;
      xform[i] = new_val;
    }

    /* xform the image */
  
    img_data = img->data + chan;
    for (y = 0; y < height; y++) 
      for (x = 0; x < width; x++, img_data+= chan_cnt) 
	*img_data = xform[*img_data];
  }

  return;
}


/* 
 * opens a file and tries to determine the file type
 *
 * return values:
 *	IMG_FTYPE_GIF
 *	IMG_FTYPE_PGM
 *	IMG_FTYPE_PPM
 *      IMG_FTYPE_NONE = unknown
 */

int get_img_filetype (char *filename)
{

  FILE *fptr;
  char  buf[16];
  unsigned char *ubuf = (unsigned char *)buf;

  if (strcmp (filename, "stdin") == 0
      || strcmp (filename, "-") == 0)
    return (IMG_FTYPE_NONE);
  else if ((fptr = fopen (filename, "r")) == NULL) {
    fprintf (stderr,"GET_IMG_FILETYPE: Couldn't open file: %s\n", filename);
    return (IMG_FTYPE_NONE);
  }

  if (fread (buf, 16, 1, fptr) != 1) {
    fprintf (stderr,"GET_IMG_FILETYPE: Unable to read from file: %s\n", filename);
    fclose (fptr);
    return (IMG_FTYPE_NONE);
  }
  fclose (fptr);

  /* GIF */
  if (strncmp (buf, "GIF8", 4) == 0)
    return (IMG_FTYPE_GIF);

  /* PGM (grayscale PNM) */
  else if ((strncmp (buf, "P2", 2) == 0)
      || (strncmp (buf, "P5", 2) == 0))
    return (IMG_FTYPE_PGM);

  /* PPM (color PNM) */
  else if ((strncmp (buf, "P3", 2) == 0)
      || (strncmp (buf, "P6", 2) == 0))
    return (IMG_FTYPE_PPM);

  /* JPEG (JFIF) */
  else if (ubuf[0] == 0xFF
	   && ubuf[1] == 0xD8
	   && ubuf[2] == 0xFF
	   && ubuf[3] == 0xE0
	   && buf[6] == 'J'
	   && buf[7] == 'F'
	   && buf[8] == 'I'
	   && buf[9] == 'F'
	   && buf[10] == 0x00)
    return (IMG_FTYPE_JPEG);

  /* JPEG (other) */
  else if (ubuf[0] == 0xFF
	   && ubuf[1] == 0xD8
	   && ubuf[2] == 0xFF
	   && ubuf[3] == 0xE1)
    return (IMG_FTYPE_JPEG);

  /* fprintf (stderr, "%x %x %x %x\n", ubuf[0], ubuf[1], ubuf[2], ubuf[3]); */

  return (IMG_FTYPE_NONE);

}

/*
 * tries to determine a file's type 
 * and then loads in the file
 *
 * return values: image = OK, NULL = problem 
 */

img_t *img_read_file (char *filename)
{
  img_t *retval;

  if (strcmp (filename, "stdin") == 0
      || strcmp (filename, "-") == 0) {
    retval = img_read_ppm (filename);
  }

  else {
    retval = NULL;
    switch (get_img_filetype (filename)) {
    case IMG_FTYPE_GIF:
      retval = LoadGifToImageBuffer (filename);
      break;
    case IMG_FTYPE_PGM:
    case IMG_FTYPE_PPM:
      retval = img_read_ppm (filename);
      break;
    }
  }

  return (retval);
}



/* 
 * tries to determine the file type based on the file name extension
 *
 * return values:
 *	IMG_FTYPE_GIF
 *	IMG_FTYPE_PGM
 *	IMG_FTYPE_PPM
 *      IMG_FTYPE_NONE = unknown
 */

int get_img_fname_filetype (char *filename)
{
  char *suffix = strrchr (filename, '.');
  if (suffix != NULL) {
    suffix++;
    if (strcasecmp (suffix, "JPG") == 0 || strcasecmp (suffix, "JPEG") == 0) 
      return (IMG_FTYPE_JPEG);
    else if (strcasecmp (suffix, "GIF") == 0)
      return (IMG_FTYPE_GIF);
    else if (strcasecmp (suffix , "PGM") == 0)
      return (IMG_FTYPE_PGM);
    else if (strcasecmp (suffix, "PPM") == 0)
      return (IMG_FTYPE_PPM);
  }
  return (IMG_FTYPE_NONE);
}


/*
 * Histogram equalization of an image
 *
 * Modified version of original code from Henry Rowley, har+@cs.cmu.edu
 *
 * The original data is modified.
 *
 */

void img_hist_eq (img_t *img)
{
  long       size;
  pix_val_t  *ptr;
  long        hist[256];
  long        sum, i;
  double      scale;
  int         chan, chan_cnt;

  size = img->width * img->height;

  /* loop through the color channels */

  chan_cnt = img->chan_cnt;

  for (chan = 0; chan < chan_cnt; chan++) {

    /* Compute histogram */

    ptr = img->data + chan;
    for (i=0; i<256; i++) 
      hist[i]=0;
    for (i=0; i<size; i++, ptr+= chan_cnt) 
      hist[*ptr]++;

    /* Compute cumulative histogram and map */

    sum = 0;
    scale = (double)size / 255.0;
    for (i = 0; i < 256; i++) {
      sum+=hist[i];
      hist[i] = (int)floor ((double)sum / scale + 0.5);
      if (hist[i] > 255) hist[i] = 255;
      if (hist[i] < 0)   hist[i] = 0;
    }

    /* Apply histogram equalization to image */

    ptr = img->data + chan;
    for (i=0; i<size; i++, ptr+= chan_cnt) 
      *ptr=hist[*ptr];
  }

  return;
}



/*
 * Histogram dynamic range adjustment of an image
 *
 * Takes bottom 2% and top 2% of the image pixels and sets those
 * values to 0 and 255, respectively.
 *
 * If chan_flag is true, each color channel is processed
 * indepedently in the case of color images.
 *
 * The original data is modified.
 *
 */

void img_hist_range_adjust (img_t *img, int chan_flag)
{
  long       size;
  pix_val_t  *ptr, min_val, max_val;
  long        hist[256];
  long        sum, i;
  long        chunk;
  double      scale;
  int         chan, chan_cnt;

  /* initialization */

  size = img->width * img->height;
  if (!chan_flag)
    size*= img->chan_cnt;

  /* to remove warnings */

  min_val = max_val = 0;

  /* loop through the color channels */

  if (chan_flag)
    chan_cnt = img->chan_cnt;
  else
    chan_cnt = 1;

  for (chan = 0; chan < chan_cnt; chan++) {

    /* Compute histogram */

    ptr = img->data + chan;
    for (i=0; i<256; i++) 
      hist[i]=0;
    for (i=0; i<size; i++, ptr+= chan_cnt) 
      hist[*ptr]++;

    /* Find the min and max points */

    chunk = (double)size * 0.02;
    sum = 0;
    for (i = 0; i < 256; i++) {
      sum+=hist[i];
      if (sum >= chunk) {
	min_val = i;
	break;
      }
    }
    sum = 0;
    for (i = 255; i >= 0; i--) {
      sum+=hist[i];
      if (sum >= chunk) {
	max_val = i;
	break;
      }
    }

    fprintf (stderr, "Min val: %d, Max val: %d\n", min_val, max_val);

    /* check for zero range */
  
    if (max_val == min_val)
      continue;

    /* Compute the transformation map */

    scale = 255.0 / (double)(max_val - min_val);
    for (i = 0; i < 256; i++) {
      hist[i] = (int)((double)(i - min_val) * scale + 0.5);
      if (hist[i] > 255) hist[i] = 255;
      if (hist[i] < 0)   hist[i] = 0;
    }

    /* Apply transformation to image */

    ptr = img->data + chan;
    for (i=0; i<size; i++, ptr+= chan_cnt) 
      *ptr=hist[*ptr];
  }

  return;
}


/* color to gray table based conversion support routines */

static int gray_tables_flag = 0;
static int red2gray[256];
static int grn2gray[256];
static int blu2gray[256];

/* build tables if not already built */
void build_gray_tables () 
{
  if (!gray_tables_flag) {
    int i;
    for (i = 0; i < 256; i++) {
      red2gray[i] = (int)((double)i * 0.3 + 0.5);
      grn2gray[i] = (int)((double)i * 0.6 + 0.5);
      blu2gray[i] = (int)((double)i * 0.1 + 0.5);
    }
    gray_tables_flag = 1;
  }
}

#define HIRES_GTBL_SZ 1024
#define MCVT(x) ((double)(x)/(double)(HIRES_GTBL_SZ-1)*255.0)
#define VCVT(x) ((double)(x)/((double)(HIRES_GTBL_SZ-1)*(double)(HIRES_GTBL_SZ-1))*255.0*255.0)

static int hires_gray_tables_flag = 0;
static int red2hrgry[HIRES_GTBL_SZ];
static int grn2hrgry[HIRES_GTBL_SZ];
static int blu2hrgry[HIRES_GTBL_SZ];

/* build tables if not already built */
void build_hires_gray_tables () 
{
  if (!hires_gray_tables_flag) {
    int i;
    for (i = 0; i < 256; i++) {
      red2hrgry[i] = (int)((double)i / 255.0 * (double)(HIRES_GTBL_SZ-1) 
			   * 0.3 + 0.5);
      grn2hrgry[i] = (int)((double)i / 255.0 * (double)(HIRES_GTBL_SZ-1) 
			   * 0.6 + 0.5);
      blu2hrgry[i] = (int)((double)i / 255.0 * (double)(HIRES_GTBL_SZ-1) 
			   * 0.1 + 0.5);
    }
    hires_gray_tables_flag = 1;
  }
}


/*
 * Histogram dynamic range adjustment of an image
 *
 * Takes bottom 0.1% and top 1% of the image pixels and sets those
 * values to 0 and 255, respectively.
 *
 * This is biased for images with white backgrounds
 *
 * This image treats monochrome images the same as
 * img_hist_range_adjust.
 *
 * In the case of color images, it histograms the grayscale
 * version of the image.  Then it maps all three channels
 * indepdently based on the average lightest and darkest colors
 * for that gray level.
 *
 * The original data is modified.
 *
 */

void img_hist_range_adjust2 (img_t *img){
  img_hist_range_adjust2_internal (img, 1);
}

void img_hist_range_adjust2_white_only (img_t *img) {
  img_hist_range_adjust2_internal (img, 0);
}

void img_hist_range_adjust2_internal (img_t *img, int black_adj_flag)
{

  /* handle monochrome images */

  if (img->chan_cnt == 1) {

    long       size;
    pix_val_t  *ptr, min_val, max_val;
    long        hist[256];
    long        sum, i;
    long        chunk;
    double      scale;

    /* initialization */

    size = img->width * img->height;
    min_val = 0;
    max_val = 255;

    /* to remove warnings */

    /* Compute histogram */

    ptr = img->data;
    for (i=0; i<256; i++) 
      hist[i]=0;
    for (i=0; i<size; i++)
      hist[*ptr++]++;

    /* Find the min and max points */

    if (black_adj_flag) {
      chunk = (double)size * 0.001;
      sum = 0;
      for (i = 0; i < 256; i++) {
	sum+=hist[i];
	if (sum >= chunk) {
	  min_val = i;
	  break;
	}
      }
    }
    chunk = (double)size * 0.01;
    sum = 0;
    for (i = 255; i >= 0; i--) {
      sum+=hist[i];
      if (sum >= chunk) {
	max_val = i;
	break;
      }
    }

    fprintf (stderr, "Min val: %d, Max val: %d\n", min_val, max_val);

    /* check for zero range */
  
    if (max_val == min_val)
      return;

    /* Compute the transformation map */

    scale = 255.0 / (double)(max_val - min_val);
    for (i = 0; i < 256; i++) {
      hist[i] = (int)((double)(i - min_val) * scale + 0.5);
      if (hist[i] > 255) hist[i] = 255;
      if (hist[i] < 0)   hist[i] = 0;
    }

    /* Apply transformation to image */

    ptr = img->data;
    for (i=0; i<size; i++)
      *ptr=hist[*ptr++];
  }

  /* handle color images */

  else {

    long       size;
    pix_val_t  *ptr, min_val, max_val;
    pix_val_t   min[3], max[3];
    long        hist[256];
    pix_val_t   red_min[256], grn_min[256], blu_min[256];
    pix_val_t   red_max[256], grn_max[256], blu_max[256];
    long        sum, i;
    long        chunk;
    double      scale;
    int         chan, chan_cnt, val;

    /* build tables if not already built */

    build_gray_tables ();

    /* initialization */

    chan_cnt = img->chan_cnt;
    size = img->width * img->height;
    min_val = 0;
    max_val = 255;

    /* to remove warnings */

    /* Compute histogram */

    ptr = img->data;
    for (i = 0; i < 256; i++) {
      hist[i] = 0;
      red_min[i] = grn_min[i] = blu_min[i] = 255;
      red_max[i] = grn_max[i] = blu_max[i] = 0;
    }
    for (i = 0; i < size; i++, ptr+= chan_cnt) {
      val = red2gray[*ptr] + grn2gray[*(ptr+1)] + blu2gray[*(ptr+2)];
      if (val > 255)
	val = 255;
      hist[val]++;
      if (*ptr < red_min[val])
	red_min[val] = *ptr;
      if (*ptr > red_max[val])
	red_max[val] = *ptr;
      if (*(ptr+1) < grn_min[val])
	grn_min[val] = *(ptr+1);
      if (*(ptr+1) > grn_max[val])
	grn_max[val] = *(ptr+1);
      if (*(ptr+2) < blu_min[val])
	blu_min[val] = *(ptr+2);
      if (*(ptr+2) > blu_max[val])
	blu_max[val] = *(ptr+2);
    }

    /* Find the min and max points */

    if (black_adj_flag) {
      chunk = (double)size * 0.001;
      sum = 0;
      for (i = 0; i < 256; i++) {
	sum+=hist[i];
	if (sum >= chunk) {
	  min_val = i;
	  break;
	}
      }
    }
    else {
      red_min[0] = grn_min[0] = blu_min[0] = 0;
    }
    chunk = (double)size * 0.01;
    sum = 0;
    for (i = 255; i >= 0; i--) {
      sum+=hist[i];
      if (sum >= chunk) {
	max_val = i;
	break;
      }
    }

    /* compute the separate channel adjustments */

    min[0] = red_min[min_val];
    min[1] = grn_min[min_val];
    min[2] = blu_min[min_val];
    max[0] = red_max[max_val];
    max[1] = grn_max[max_val];
    max[2] = blu_max[max_val];

    fprintf (stderr, 
	     "Min val: %d (%d,%d,%d); Max val: %d (%d,%d,%d)\n", 
	     min_val, min[0], min[1], min[2],
	     max_val, max[0], max[1], max[2]);

    /* loop through the color channels */

    for (chan = 0; chan < chan_cnt; chan++) {

      min_val = min[chan];
      max_val = max[chan];

      /* check for zero range */
  
      if (max_val == min_val)
	continue;

      /* Compute the transformation map */

      scale = 255.0 / (double)(max_val - min_val);
      for (i = 0; i < 256; i++) {
	hist[i] = (int)((double)(i - min_val) * scale + 0.5);
	if (hist[i] > 255) hist[i] = 255;
	if (hist[i] < 0)   hist[i] = 0;
      }

      /* Apply transformation to image */

      ptr = img->data + chan;
      for (i=0; i<size; i++, ptr+= chan_cnt) 
	*ptr=hist[*ptr];
    }
  }

  return;
}


/*
 * Histogram equalization of an image
 *
 * Main difference from img_hist_eq is that this version zeroes out 
 * the bottom 2% and top 2% of the histogram (based on counts)
 * this allows it to handle very light and dark backgrounds better.
 *
 * The original data is modified.
 *
 */

void img_hist_eq2 (img_t *img)
{
  long       size;
  pix_val_t  *ptr;
  long        hist[256];
  long        sum, i;
  double      scale;
  long        chunk;
  int         chan, chan_cnt;

  size = img->width * img->height;

  /* loop through the color channels */

  chan_cnt = img->chan_cnt;

  for (chan = 0; chan < chan_cnt; chan++) {

    /* Compute histogram */

    ptr = img->data + chan;
    for (i=0; i<256; i++) 
      hist[i]=0;
    for (i=0; i<size; i++, ptr+= chan_cnt) 
      hist[*ptr]++;

    /* Zero out the bottom and top 2% of the counts */

    chunk = (double)size * 0.02;
    sum = 0;
    for (i = 0; i < 256; i++) {
      sum+=  hist[i];
      size-= hist[i];
      hist[i] = 0;
      if (sum >= chunk)
	break;
    }
    sum = 0;
    for (i = 255; i >= 0; i--) {
      sum+=  hist[i];
      size-= hist[i];
      hist[i] = 0;
      if (sum >= chunk)
	break;
    }

    /* check for zero pixels left */
  
    if (size == 0)
      return;

    /* Compute cumulative histogram and map */

    sum = 0;
    scale = (double)size / 255.0;
    for (i = 0; i < 256; i++) {
      sum+=hist[i];
      hist[i] = (int)floor ((double)sum / scale + 0.5);
      if (hist[i] > 255) hist[i] = 255;
      if (hist[i] < 0)   hist[i] = 0;
    }

    /* Apply histogram equalization to image */

    ptr = img->data + chan;
    size = img->width * img->height;
    for (i=0; i<size; i++, ptr+= chan_cnt) 
      *ptr=hist[*ptr];
  }

  return;
}



/*
 * Load in and scale a set of face images
 *
 * Args:    ip_mode: 
 *		0 = do nothing, returns original image pointer
 *		1 = high pass the image, frees orig image & returns new one
 *		2 = high pass and simple stretch the image, 
 *		    frees orig image & returns new one
 *		3 = standard hist equalization,  returns orig image ptr
 *		4 = standard hist range adjustment, returns orig image ptr
 *		5 = hist eq2 for very light and dark backgrounds, 
 *		    returns orig image ptr
 *	    other = print error message and exit
 *
 * Return:  A pointer to the image.
 * Errors:  Exits on errors
 *
 */

img_t *img_process (img_t *image, int ip_mode) {
  return (img_process_args (image, ip_mode, NULL));
}

#define arg_ival(x) (args ? arg_ivalue : (x))
#define arg_fval(x) (args ? arg_fvalue : (x))

img_t *img_process_args (img_t *image, int ip_mode, double *args)
{

  int    arg_ivalue;
  double arg_fvalue;

  if (args) {
    arg_ivalue = *args + 0.5;
    arg_fvalue = *args;
  }
  else {
    arg_ivalue = 0;
    arg_fvalue = 0.0;
  }

  switch (ip_mode) {
  case 0:
    break;
  case 1:
  case 2:
    {
      img_t *ip_image;
      ip_image = img_simple_high_pass (image);
      if (ip_image == NULL) {
	fprintf (stderr, "\nProblem processing image: %s.\n\n", image->name);
	exit (1);
      }
      img_free (image);
      image = ip_image;
    }
    if (ip_mode == 2)
      img_simple_stretch (image);
    break;
  case 3:
    img_hist_eq (image);
    break;
  case 4:
    img_hist_range_adjust (image, 1);
    break;
  case 5:
    img_hist_eq2 (image);
    break;
  case 6:
    img_simple_normalize (image, arg_ival (64));
    break;
  case 7:
    img_hist_range_adjust (image, 1);
    img_simple_normalize (image, arg_ival (64));
    break;
  case 8:
    img_simple_normalize (image, arg_ival (48));
    break;
  case 9:
    img_hist_range_adjust (image, 0);
    break;
  case 10:
    img_hist_range_adjust2 (image);
    break;
  case 11:
    img_hist_range_adjust2_white_only (image);
    break;
  case 12:
    img_hist_range_adjust3 (image);
    break;
  case 13:
    img_hist_range_adjust3_white_only (image);
    break;
  case 14:
    img_hist_range_adjust4 (image);
    break;
  case 15:
    img_hist_range_adjust4_white_only (image);
    break;
  case 16:
    img_hist_range_adjust5 (image);
    break;
  case 17:
    img_hist_range_adjust5_white_only (image);
    break;
  case 18:
    img_hist_range_adjust4_simple_black (image);
    break;
  case 19:
    img_hist_range_adjust4_var_black (image);
    break;
  case 20:
    img_hist_range_adjust4_mix0 (image);
    break;
  case 21:
    img_hist_range_adjust4_mix1 (image);
    break;
  case 22:
    img_hist_range_adjust4_mix2 (image);
    break;
  case 23:
    img_hist_range_adjust4_fast (image);
    break;
  case 24:
    img_hist_range_adjust4_fast_mix0 (image);
    break;
  case 25:
    img_simple_normalize2 (image, arg_ival (64));
    break;
  case 26:
    img_hist_range_adjust (image, 1);
    img_simple_normalize2 (image, arg_ival (64));
    break;
  case 27:
    img_simple_normalize3 (image);
    break;
  case 28:
    img_hist_range_adjust (image, 1);
    img_simple_normalize3 (image);
    break;
  case 29:
    img_hist_range_adjust4_fast_nofix (image);
    break;
  case 30:
    img_hist_range_adjust6 (image);
    break;
  case 31:
    img_hist_range_adjust6_fast (image);
    break;
  case 32:
    img_hist_quant_adjust6 (image);
    break;
  case 33:
    img_hist_quant_adjust6_fast (image);
    break;
  case 34:
    img_mean_scale (image, arg_ival (128));
    break;
  case 35:
    img_mean_scale (image, arg_ival (96));
    break;
  case 36:
    img_mean_scale (image, arg_ival (64));
    break;
  case 37:
    img_mean_scale (image, arg_ival (48));
    break;
  case 38:
    img_mean_scale (image, arg_ival (32));
    break;
  case 39:
  case 40:
  case 41:
  case 42:
    {
      img_t *ip_image;
      double weight;
      switch (ip_mode) {
      case 39:
	weight = 0.25;
	break;
      case 40:
	weight = 0.50;
	break;
      case 41:
	weight = 0.75;
	break;
      default:
	weight = 1.00;
	break;
      }
      ip_image = img_simple_sharpen (image, arg_fval (weight));
      if (ip_image == NULL) {
	fprintf (stderr, "\nProblem processing image: %s.\n\n", image->name);
	exit (1);
      }
      img_free (image);
      image = ip_image;
    }
    break;
  case 43:
    {
      img_t *ip_image;
      ip_image = img_simple_sharpen_bound (image, arg_fval (0.5));
      if (ip_image == NULL) {
	fprintf (stderr, "\nProblem processing image: %s.\n\n", image->name);
	exit (1);
      }
      img_free (image);
      image = ip_image;
    }
    break;
  default:
    fprintf (stderr, "\nUnknown processing mode %d for image: %s.\n\n", 
	     ip_mode, image->name);
    exit (1);
  }
 
  return (image);
}



/*
 * Dump a description of the image processing options to stderr
 *
 */


void ip_help (FILE *stream, char *indent) 
{
  fprintf (stream, "%s 0 - none\n", indent);
  fprintf (stream, "%s 1 - high pass\n", indent);
  fprintf (stream, "%s 2 - high pass and stretch\n", indent);
  fprintf (stream, "%s 3 - histogram equalize\n", indent);
  fprintf (stream, "%s 4 - histogram range adjust\n", indent);
  fprintf (stream, "%s 5 - histogram equalize with light and dark adjustment\n", indent);
  fprintf (stream, "%s 6 - histogram normalize with range 64\n", indent);
  fprintf (stream, "%s 7 - range adjust and then histogram normalize with range 64\n", indent);
  fprintf (stream, "%s 8 - histogram normalize with range 48\n", indent);
  fprintf (stream, "%s 9 - histogram range adjust, all color channels grouped together\n", indent);
  fprintf (stream, "%s10 - histogram range adjust, better handling of color channels\n", indent);
  fprintf (stream, "%s11 - same as 10, but only adjusts white point\n", indent);
  fprintf (stream, "%s12 - histogram range adjust, functional but theoretically flawed EM based\n", indent);
  fprintf (stream, "%s13 - same as 12, but only adjusts white point\n", indent);
  fprintf (stream, "%s14 - histogram range adjust, EM based\n", indent);
  fprintf (stream, "%s15 - same as 14, but only adjusts white point\n", indent);
  fprintf (stream, "%s16 - histogram range adjust, EM based on grayscale histogram\n", indent);
  fprintf (stream, "%s17 - same as 16, but only adjusts white point\n", indent);
  fprintf (stream, "%s18 - same as 14, but simple black point adjustment\n", indent);
  fprintf (stream, "%s19 - same as 14, but variance based black point adjustment\n", indent);
  fprintf (stream, "%s20 - same as 14, but only uses adjust 0\n", indent);
  fprintf (stream, "%s21 - same as 14, but only uses adjust 1\n", indent);
  fprintf (stream, "%s22 - same as 14, but only uses adjust 2\n", indent);
  fprintf (stream, "%s23 - same as 14, but fast approximation\n", indent);
  fprintf (stream, "%s24 - same as 23, but only uses adjust 0\n", indent);
  fprintf (stream, "%s25 - same as  6, but does not change channel means\n", indent);
  fprintf (stream, "%s26 - same as 25, but range adjust first\n", indent);
  fprintf (stream, "%s27 - same as 25, but normalizes channel variances\n", indent);
  fprintf (stream, "%s28 - same as 27, but range adjust first\n", indent);
  fprintf (stream, "%s29 - same as 23, but does not patch light and dark colors\n", indent);
  fprintf (stream, "%s30 - histogram range adjust, multiple cluster EM based\n", indent);
  fprintf (stream, "%s31 - same as 30, but fast approximation\n", indent);
  fprintf (stream, "%s32 - image color quantization, multiple cluster EM based\n", indent);
  fprintf (stream, "%s33 - same as 32, but fast approximation\n", indent);
  fprintf (stream, "%s34 - apply a single scale factor to all channels so the mean is 128\n", indent);
  fprintf (stream, "%s35 - apply a single scale factor to all channels so the mean is 96\n", indent);  
 fprintf (stream, "%s36 - apply a single scale factor to all channels so the mean is 64\n", indent);  
  fprintf (stream, "%s37 - apply a single scale factor to all channels so the mean is 48\n", indent);  
  fprintf (stream, "%s38 - apply a single scale factor to all channels so the mean is 32\n", indent);  
  fprintf (stream, "%s39 - light  sharpening (wt = 0.25)\n", indent);  
  fprintf (stream, "%s40 - medium sharpening (wt = 0.50)\n", indent);  
  fprintf (stream, "%s41 - medium sharpening (wt = 0.75)\n", indent);  
  fprintf (stream, "%s42 - heavy  sharpening (wt = 1.00)\n", indent);  
  fprintf (stream, "%s43 - medium bounded sharpening (wt = 0.50)\n", indent);  
  fprintf (stream, "%s * - note that for color images, each color channel\n", indent);
  fprintf (stream, "%s     is processed independently, except where noted otherwise\n", indent);
  return;
}



/*%************************************************************************
 *
 * Blit an image
 *
 * Copies the enitre contents of one image to a subregion of another.
 *
 * Currently does not do any bounds checking or cropping, the src image must
 * fit in the dest image.
 *
 * properly handles blitting between combinations of color 
 * and monochrome images
 *
 **************************************************************************
 */

void img_blit (img_t *dest_img, img_t *src_img,
	       img_dim_t x_start, img_dim_t y_start)
{

  img_dim_t      src_width, src_height, src_stride, dest_stride, x, y;
  pix_val_t     *src_data, *dest_data, *dest_line_ptr;
  int            src_chan_cnt, dest_chan_cnt;
  int            r_val, g_val, b_val;

  /* get the image stats */

  src_width  = src_img->width;
  src_height = src_img->height;
  src_stride = src_img->stride;
  src_chan_cnt  = src_img->chan_cnt;
  dest_stride   = dest_img->stride;
  dest_chan_cnt = dest_img->chan_cnt;

  /* copy the data over */

  src_data = src_img->data;
  dest_line_ptr = dest_img->data + y_start * dest_img->stride 
    + x_start * dest_chan_cnt;
  for (y = 0; y < src_height; y++, dest_line_ptr+= dest_img->stride) {
    dest_data = dest_line_ptr;
    
    /* images have the same channel count */
    if (src_chan_cnt == dest_chan_cnt) {
      for (x = 0; x < src_stride; x++) {
	*dest_data++ = *src_data++;
      }
    }

    /* images have different channel counts */
    else {
      for (x = 0; x < src_width; x++) {
	if (src_chan_cnt == 1) {
	  r_val = g_val = b_val = *src_data++;
	}
	else {
	  r_val = *src_data++;
	  g_val = *src_data++;
	  b_val = *src_data++;
	}
	if (dest_chan_cnt == 1) {
	  *dest_data++ = (r_val * 2 + g_val * 7 + b_val) / 10;
	}
	else {
	  *dest_data++ = r_val;
	  *dest_data++ = g_val;
	  *dest_data++ = b_val;
	}
      }
    }

  }
    
  return;

}



/*
 * Crops an image.
 *
 * A new image is created with the cropped region.
 */

img_t *img_crop (img_t *img, 
		 img_dim_t x_start, img_dim_t y_start,
		 img_dim_t x_end, img_dim_t y_end)
{

  img_t         *new_img;
  img_dim_t      new_width, new_height, x, y, new_stride, old_stride;
  pix_val_t     *new_data, *src_data, *src_line_ptr;

  /* create a new image of the correct size */

  old_stride = img->width * img->chan_cnt;
  new_width  = x_end - x_start + 1;
  new_height = y_end - y_start + 1;
  new_stride = new_width * img->chan_cnt;
  new_img = img_create_basis (img, new_width, new_height, img->chan_cnt);
  if (new_img == NULL)
    return (NULL);
  new_data = new_img->data;

  /* copy the cropped data over */

  src_line_ptr = img->data + y_start * old_stride + x_start * img->chan_cnt;
  for (y = 0; y < new_height; y++, src_line_ptr+= old_stride) {
    src_data = src_line_ptr;
    for (x = 0; x < new_stride; x++) {
      *new_data++ = *src_data++;
    }
  }
    
  return (new_img);

}


/*
 * Pads an image.
 *
 * A new image is created with the specified dimensions
 *
 * background is the background color, the first byte is used
 * for grayscale images, all three for color
 */

img_t *img_pad (img_t *img, 
		img_dim_t new_width, img_dim_t new_height,
		pix_val_t *background)
{

  img_t         *new_img;
  img_dim_t      x, y, new_stride, old_stride;
  img_dim_t      x_start, y_start, old_height, old_width;
  pix_val_t     *src_ptr, *dest_ptr;
  int            c, chan_cnt;

  /* get dimensions */

  chan_cnt = img->chan_cnt;
  old_width = img->width;
  old_height = img->height;
  old_stride = img->width * chan_cnt;

  /* if any new dimension is smaller, then it is increased to the 
   * current size */

  if (new_width < old_width)
    new_width = old_width;
  if (new_height < old_height)
    new_height = old_height;

  /* create a new image of the correct size */

  new_stride = new_width * chan_cnt;
  new_img = img_create_basis (img, new_width, new_height, chan_cnt);
  if (new_img == NULL)
    return (NULL);

  /* copy the padded data over */

  x_start = (new_width - old_width) / 2;
  y_start = (new_height - old_height) / 2;

  src_ptr  = img->data;
  dest_ptr = new_img->data;

  /* add padding */
  for (y = 0; y < y_start; y++) {
    for (x = 0; x < new_width; x++) {
      for (c = 0; c < chan_cnt; c++)
	*dest_ptr++ = background[c];
    }
  }

  /* copy src data */
  for (y = 0; y < old_height; y++) {
    for (x = 0; x < x_start; x++) {
      for (c = 0; c < chan_cnt; c++) 
	*dest_ptr++ = background[c];
    }
    for (x = 0; x < old_stride; x++) {
      *dest_ptr++ = *src_ptr++;
    }
    for (x = old_width + x_start; x < new_width; x++) {
      for (c = 0; c < chan_cnt; c++) 
	*dest_ptr++ = background[c];
    }
  }

  /* add padding */
  for (y = old_height + y_start; y < new_height; y++) {
    for (x = 0; x < new_width; x++) {
      for (c = 0; c < chan_cnt; c++)
	*dest_ptr++ = background[c];
    }
  }
    
  return (new_img);

}




/*
 * Histogram dynamic range adjustment of an image
 *
 * This is EM based.  Starts with a hump and at the min, mid, and max.
 *
 * The three humps are then independent.  The fact that it get stuck
 * in a local minima is what makes this work.  But this also means
 * that since the results of the middle bump calculation are not
 * used, they are not necessary.
 * 
 * It works well, but the probability estimates aren't correct.
 * It seems like the fact that it works is a fluke.
 *
 * The monochrome implementation was never completed.
 *
 * Takes bottom 0.1% and top 0.1% of the image pixels are used for
 * initial means estimates.
 *
 * This image treats monochrome images the same as
 * img_hist_range_adjust.
 *
 * In the case of color images, it histograms the grayscale
 * version of the image.  Then it maps all three channels
 * indepdently based on the average lightest and darkest colors
 * for that gray level.
 *
 * The original data is modified.
 *
 */

void img_hist_range_adjust3 (img_t *img){
  img_hist_range_adjust3_internal (img, 1);
}

void img_hist_range_adjust3_white_only (img_t *img) {
  img_hist_range_adjust3_internal (img, 0);
}

void img_hist_range_adjust3_internal (img_t *img, int black_adj_flag)
{

  /* TO DO: complete code to handle monochrome images */

  /* handle monochrome images */

  if (img->chan_cnt == 1) {

    long       size;
    pix_val_t  *ptr, min_val, max_val;
    long        hist[256];
    long        sum, i;
    long        chunk;
    double      scale, tot, tot2, init_var;
    double      lo_mean, md_mean, hi_mean;
    double      lo_var, md_var, hi_var;

    fprintf (stderr, "WARNING: monochrome processing implementation not complete.\n");

    /* initialization */

    size = img->width * img->height;

    /* to remove warnings */

    /* Compute histogram */

    ptr = img->data;
    for (i = 0; i < 256; i++)
      hist[i] = 0;
    for (i = 0; i < size; i++)
      hist[*ptr++]++;

    /*** generate initial mean and variance estimates ***/

    /* Find the min and max points */

    min_val = 0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 0; i < 256; i++) {
      sum+=hist[i];
      if (sum >= chunk) {
	min_val = i;
	break;
      }
    }
    max_val = 255;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 255; i >= 0; i--) {
      sum+=hist[i];
      if (sum >= chunk) {
	max_val = i;
	break;
      }
    }
    tot = tot2 = 0.0;
    for (i = 0; i < 256; i++) {
      tot+= (double)hist[i] * (double)i;
      tot2+= (double)hist[i] * (double)hist[i] * (double)i;
    }

    /* generate initial estimates */

    init_var = (tot2 - tot * tot / (double)size) / (double)size;
    if (init_var <= 0.0)
      init_var = 0.01;
    lo_mean = (double)min_val;
    md_mean = tot / (double)size;
    hi_mean = (double)max_val;
    lo_var = init_var * 0.125;
    md_var = init_var * 0.25;
    hi_var = init_var * 0.125;

    fprintf (stderr, "Min val: %d, Max val: %d\n", min_val, max_val);

    /* check for zero range */
  
    if (max_val == min_val)
      return;

    /* Compute the transformation map */

    scale = 255.0 / (double)(max_val - min_val);
    for (i = 0; i < 256; i++) {
      hist[i] = (int)((double)(i - min_val) * scale + 0.5);
      if (hist[i] > 255) hist[i] = 255;
      if (hist[i] < 0)   hist[i] = 0;
    }

    /* Apply transformation to image */

    ptr = img->data;
    for (i=0; i<size; i++)
      *ptr=hist[*ptr++];
  }

  /* handle color images */

  else {

    pix_val_t  *ptr;
    int         min[3], max[3], val, min_val, max_val;
    int         chan, chan_cnt, loop_cnt;
    int         r_ival, g_ival, b_ival;
    long        size, chunk, sum, i;
    long        hist[256];
    long        red_sum[256], grn_sum[256], blu_sum[256];
    long        red_tot, grn_tot, blu_tot;
    double      red_tot2, grn_tot2, blu_tot2;
    double      scale;
    double      rval, gval, bval;
    double      rval2, gval2, bval2;
    double      prob;
    double      rdiff, gdiff, bdiff;
    double      lo_rmean, lo_gmean, lo_bmean;
    double      md_rmean, md_gmean, md_bmean;
    double      hi_rmean, hi_gmean, hi_bmean;
    double      lo_rvar, lo_gvar, lo_bvar;
    double      md_rvar, md_gvar, md_bvar;
    double      hi_rvar, hi_gvar, hi_bvar;
    double      lo_sum, md_sum, hi_sum;
    double      lo_rsum, lo_gsum, lo_bsum;
    double      md_rsum, md_gsum, md_bsum;
    double      hi_rsum, hi_gsum, hi_bsum;
    double      lo_rsum2, lo_gsum2, lo_bsum2;
    double      md_rsum2, md_gsum2, md_bsum2;
    double      hi_rsum2, hi_gsum2, hi_bsum2;
    double      init_rvar, init_gvar, init_bvar;


    /* initialization */

    chan_cnt = img->chan_cnt;
    size = img->width * img->height;

    /* build tables if not already built */

    build_gray_tables ();

    /*** generate initial mean and variance estimates ***/

    /* note that this is needed or the algorithm my latch on to 
     * a group of red pixels which have zero g and b values 
     */

    /* Compute histogram */

    ptr = img->data;
    for (i = 0; i < 256; i++) {
      hist[i] = 0;
      red_sum[i] = grn_sum[i] = blu_sum[i] = 0;
    }
    red_tot2 = grn_tot2 = blu_tot2 = 0.0;
    for (i = 0; i < size; i++) {
      r_ival = *ptr++;
      g_ival = *ptr++;
      b_ival = *ptr++;
      val = red2gray[r_ival] + grn2gray[g_ival] + blu2gray[b_ival];
      if (val > 255)
	val = 255;
      hist[val]++;
      red_sum[val]+= r_ival;
      grn_sum[val]+= g_ival;
      blu_sum[val]+= b_ival;
      red_tot2+= (double)(r_ival * r_ival);
      grn_tot2+= (double)(g_ival * g_ival);
      blu_tot2+= (double)(b_ival * b_ival);
    }

    /* Find the min and max points */

    min_val = 0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 0; i < 256; i++) {
      sum+=hist[i];
      if (sum >= chunk) {
	min_val = i;
	break;
      }
    }
    max_val = 255;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 255; i >= 0; i--) {
      sum+=hist[i];
      if (sum >= chunk) {
	max_val = i;
	break;
      }
    }
    red_tot = grn_tot = blu_tot = 0;
    for (i = 0; i < 256; i++) {
      red_tot+= red_sum[i];
      grn_tot+= grn_sum[i];
      blu_tot+= blu_sum[i];
    }


    /* generate initial estimates */

    init_rvar = (red_tot2 - (double)red_tot * (double)red_tot / (double)size) / (double)size;
    init_gvar = (grn_tot2 - (double)grn_tot * (double)grn_tot / (double)size) / (double)size;
    init_bvar = (blu_tot2 - (double)blu_tot * (double)blu_tot / (double)size) / (double)size;
    if (init_rvar <= 0.0)
      init_rvar = 0.01;
    if (init_gvar <= 0.0)
      init_gvar = 0.01;
    if (init_bvar <= 0.0)
      init_bvar = 0.01;
    lo_rmean = (double)red_sum[min_val] / (double)hist[min_val];
    lo_gmean = (double)grn_sum[min_val] / (double)hist[min_val];
    lo_bmean = (double)blu_sum[min_val] / (double)hist[min_val];
    md_rmean = (double)red_tot          / (double)size;
    md_gmean = (double)grn_tot          / (double)size;
    md_bmean = (double)blu_tot          / (double)size;
    hi_rmean = (double)red_sum[max_val] / (double)hist[max_val];
    hi_gmean = (double)grn_sum[max_val] / (double)hist[max_val];
    hi_bmean = (double)blu_sum[max_val] / (double)hist[max_val];
    lo_rvar = init_rvar * 0.125;
    lo_gvar = init_gvar * 0.125;
    lo_bvar = init_bvar * 0.125;
    md_rvar = init_rvar * 0.25;
    md_gvar = init_gvar * 0.25;
    md_bvar = init_bvar * 0.25;
    hi_rvar = init_rvar * 0.125;
    hi_gvar = init_gvar * 0.125;
    hi_bvar = init_bvar * 0.125;


    /*** loop through and perform EM iterations ***/

    for (loop_cnt = -1; loop_cnt < 5; loop_cnt++) {

      /* Compute the contributions from each pixel */

      if (loop_cnt >= 0) {
	lo_sum = md_sum = hi_sum = 0.0;
	lo_rsum = lo_gsum = lo_bsum = 0.0;
	md_rsum = md_gsum = md_bsum = 0.0;
	hi_rsum = hi_gsum = hi_bsum = 0.0;
	lo_rsum2 = lo_gsum2 = lo_bsum2 = 0.0;
	md_rsum2 = md_gsum2 = md_bsum2 = 0.0;
	hi_rsum2 = hi_gsum2 = hi_bsum2 = 0.0;
	ptr = img->data;
	for (i = 0; i < size; i++) {
	  rval = *ptr++;
	  gval = *ptr++;
	  bval = *ptr++;
	  rval2 = rval * rval;
	  gval2 = gval * gval;
	  bval2 = bval * bval;

	  rdiff = rval - lo_rmean;
	  gdiff = gval - lo_gmean;
	  bdiff = bval - lo_bmean;
	  prob = exp (-0.5 * (rdiff * rdiff / lo_rvar 
			      + gdiff * gdiff / lo_gvar
			      + bdiff * bdiff / lo_bvar));
	  lo_rsum+= rval * prob;
	  lo_gsum+= gval * prob;
	  lo_bsum+= bval * prob;
	  lo_rsum2+= rval2 * prob;
	  lo_gsum2+= gval2 * prob;
	  lo_bsum2+= bval2 * prob;
	  lo_sum+= prob;

	  rdiff = rval - md_rmean;
	  gdiff = gval - md_gmean;
	  bdiff = bval - md_bmean;
	  prob = exp (-0.5 * (rdiff * rdiff / md_rvar 
			      + gdiff * gdiff / md_gvar
			      + bdiff * bdiff / md_bvar));
	  md_rsum+= rval * prob;
	  md_gsum+= gval * prob;
	  md_bsum+= bval * prob;
	  md_rsum2+= rval2 * prob;
	  md_gsum2+= gval2 * prob;
	  md_bsum2+= bval2 * prob;
	  md_sum+= prob;

	  rdiff = rval - hi_rmean;
	  gdiff = gval - hi_gmean;
	  bdiff = bval - hi_bmean;
	  prob = exp (-0.5 * (rdiff * rdiff / hi_rvar 
			      + gdiff * gdiff / hi_gvar
			      + bdiff * bdiff / hi_bvar));
	  hi_rsum+= rval * prob;
	  hi_gsum+= gval * prob;
	  hi_bsum+= bval * prob;
	  hi_rsum2+= rval2 * prob;
	  hi_gsum2+= gval2 * prob;
	  hi_bsum2+= bval2 * prob;
	  hi_sum+= prob;
	}

	/* now calculate the updated means and variances */

	lo_rmean = lo_rsum / lo_sum;
	lo_gmean = lo_gsum / lo_sum;
	lo_bmean = lo_bsum / lo_sum;
	lo_rvar  = lo_rsum2 / lo_sum - lo_rmean * lo_rmean;
	lo_gvar  = lo_gsum2 / lo_sum - lo_gmean * lo_gmean;
	lo_bvar  = lo_bsum2 / lo_sum - lo_bmean * lo_bmean;
	if (lo_rvar <= 0.0)
	  lo_rvar = 0.01;
	if (lo_gvar <= 0.0)
	  lo_gvar = 0.01;
	if (lo_bvar <= 0.0)
	  lo_bvar = 0.01;

	md_rmean = md_rsum / md_sum;
	md_gmean = md_gsum / md_sum;
	md_bmean = md_bsum / md_sum;
	md_rvar  = md_rsum2 / md_sum - md_rmean * md_rmean;
	md_gvar  = md_gsum2 / md_sum - md_gmean * md_gmean;
	md_bvar  = md_bsum2 / md_sum - md_bmean * md_bmean;
	if (md_rvar <= 0.0)
	  md_rvar = 0.01;
	if (md_gvar <= 0.0)
	  md_gvar = 0.01;
	if (md_bvar <= 0.0)
	  md_bvar = 0.01;

	hi_rmean = hi_rsum / hi_sum;
	hi_gmean = hi_gsum / hi_sum;
	hi_bmean = hi_bsum / hi_sum;
	hi_rvar  = hi_rsum2 / hi_sum - hi_rmean * hi_rmean;
	hi_gvar  = hi_gsum2 / hi_sum - hi_gmean * hi_gmean;
	hi_bvar  = hi_bsum2 / hi_sum - hi_bmean * hi_bmean;
	if (hi_rvar <= 0.0)
	  hi_rvar = 0.01;
	if (hi_gvar <= 0.0)
	  hi_gvar = 0.01;
	if (hi_bvar <= 0.0)
	  hi_bvar = 0.01;
      }

      fprintf (stderr, 
	       "Iter %d: Lo Mean: r=%5.1f, g=%5.1f, b=%5.1f; Lo Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	       loop_cnt + 1 ,
	       lo_rmean, lo_gmean, lo_bmean,
	       lo_rvar, lo_gvar, lo_bvar);
      fprintf (stderr, 
	       "        Md Mean: r=%5.1f, g=%5.1f, b=%5.1f; Md Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	       md_rmean, md_gmean, md_bmean,
	       md_rvar, md_gvar, md_bvar);
      fprintf (stderr, 
	       "        Hi Mean: r=%5.1f, g=%5.1f, b=%5.1f; Hi Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	       hi_rmean, hi_gmean, hi_bmean,
	       hi_rvar, hi_gvar, hi_bvar);

    }

    /* compute the separate channel adjustments */
    if (black_adj_flag) {
      min[0] = lo_rmean - sqrt (lo_rvar) * 0.5;
      min[1] = lo_gmean - sqrt (lo_gvar) * 0.5;
      min[2] = lo_bmean - sqrt (lo_bvar) * 0.5;
      if (min[0] < 0)
	min[0] = 0;
      if (min[1] < 0)
	min[1] = 0;
      if (min[2] < 0)
	min[2] = 0;
    }
    else {
      min[0] = 0;
      min[1] = 0;
      min[2] = 0;
    }
    max[0] = hi_rmean + sqrt (hi_rvar) * 0.5;
    max[1] = hi_gmean + sqrt (hi_gvar) * 0.5;
    max[2] = hi_bmean + sqrt (hi_bvar) * 0.5;
    if (max[0] > 255)
      max[0] = 255;
    if (max[1] > 255)
      max[1] = 255;
    if (max[2] > 255)
      max[2] = 255;

    /* loop through the color channels */

    for (chan = 0; chan < chan_cnt; chan++) {

      min_val = min[chan];
      max_val = max[chan];

      /* check for zero range */
  
      if (max_val == min_val)
	continue;

      /* Compute the transformation map */

      scale = 255.0 / (double)(max_val - min_val);
      for (i = 0; i < 256; i++) {
	hist[i] = (int)((double)(i - min_val) * scale + 0.5);
	if (hist[i] > 255) hist[i] = 255;
	if (hist[i] < 0)   hist[i] = 0;
      }

      /* Apply transformation to image */

      ptr = img->data + chan;
      for (i=0; i<size; i++, ptr+= chan_cnt) 
	*ptr=hist[*ptr];
    }
  }

  return;
}




/*
 * Histogram dynamic range adjustment of an image
 *
 * This is EM based.  Starts with a hump and at the min, mid, and max.
 *
 * The monochrome implementation was never completed.
 *
 * Takes bottom 0.1% and top 0.1% of the image pixels are used for
 * initial means estimates.
 *
 * This image treats monochrome images the same as
 * img_hist_range_adjust.
 *
 * In the case of color images, it histograms the grayscale
 * version of the image.  Then it maps all three channels
 * indepdently based on the average lightest and darkest colors
 * for that gray level.
 *
 * The original data is modified.
 *
 */

#define ADJUST4_LOOP_CNT 25

void img_hist_range_adjust4 (img_t *img) {
  img_hist_range_adjust4_internal (img, 1, 0, 1);
}

void img_hist_range_adjust4_white_only (img_t *img) {
  img_hist_range_adjust4_internal (img, 0, 0, 1);
}

void img_hist_range_adjust4_simple_black (img_t *img) {
  img_hist_range_adjust4_internal (img, 2, 0, 1);
}

void img_hist_range_adjust4_var_black (img_t *img) {
  img_hist_range_adjust4_internal (img, 3, 0, 1);
}

void img_hist_range_adjust4_mix0 (img_t *img) {
  img_hist_range_adjust4_internal (img, 4, 0, 1);
}

void img_hist_range_adjust4_mix1 (img_t *img) {
  img_hist_range_adjust4_internal (img, 5, 0, 1);
}

void img_hist_range_adjust4_mix2 (img_t *img) {
  img_hist_range_adjust4_internal (img, 6, 0, 1);
}

void img_hist_range_adjust4_fast (img_t *img) {
  img_hist_range_adjust4_internal (img, 1, 1, 1);
}

void img_hist_range_adjust4_fast_mix0 (img_t *img) {
  img_hist_range_adjust4_internal (img, 4, 1, 1);
}

void img_hist_range_adjust4_fast_nofix (img_t *img) {
  img_hist_range_adjust4_internal (img, 1, 1, 0);
}

double var_adjust (double rvar, double gvar, double bvar)
{
  return (pow (sqrt(rvar) * 0.3 + sqrt(gvar) * 0.6 + sqrt(bvar) * 0.1, 2.0));
}


void img_hist_range_adjust4_internal (img_t *img, int black_adj_mode,
				      int fast_flag, int fix_flag)
{

  /* TO DO: complete code to handle monochrome images */

  /* handle monochrome images */

  if (img->chan_cnt == 1) {

    long       size;
    pix_val_t  *ptr, min_val, max_val;
    long        hist[256];
    long        sum, i;
    long        chunk;
    double      scale, tot, tot2, init_var;
    double      lo_mean, md_mean, hi_mean;
    double      lo_var, md_var, hi_var;

    fprintf (stderr, "WARNING: monochrome processing implementation not complete.\n");

    /* initialization */

    size = img->width * img->height;

    /* to remove warnings */

    /* Compute histogram */

    ptr = img->data;
    for (i = 0; i < 256; i++)
      hist[i] = 0;
    for (i = 0; i < size; i++)
      hist[*ptr++]++;

    /*** generate initial mean and variance estimates ***/

    /* Find the min and max points */

    min_val = 0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 0; i < 256; i++) {
      sum+=hist[i];
      if (sum >= chunk) {
	min_val = i;
	break;
      }
    }
    max_val = 255;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 255; i >= 0; i--) {
      sum+=hist[i];
      if (sum >= chunk) {
	max_val = i;
	break;
      }
    }
    tot = tot2 = 0.0;
    for (i = 0; i < 256; i++) {
      tot+= (double)hist[i] * (double)i;
      tot2+= (double)hist[i] * (double)hist[i] * (double)i;
    }

    /* generate initial estimates */

    init_var = (tot2 - tot * tot / (double)size) / (double)size;
    if (init_var <= 1.0)
      init_var = 0.01;
    lo_mean = (double)min_val;
    md_mean = tot / (double)size;
    hi_mean = (double)max_val;
    lo_var = init_var * 0.125;
    md_var = init_var * 0.25;
    hi_var = init_var * 0.125;

    fprintf (stderr, "Min val: %d, Max val: %d\n", min_val, max_val);

    /* check for zero range */
  
    if (max_val == min_val)
      return;

    /* Compute the transformation map */

    scale = 255.0 / (double)(max_val - min_val);
    for (i = 0; i < 256; i++) {
      hist[i] = (int)((double)(i - min_val) * scale + 0.5);
      if (hist[i] > 255) hist[i] = 255;
      if (hist[i] < 0)   hist[i] = 0;
    }

    /* Apply transformation to image */

    ptr = img->data;
    for (i=0; i<size; i++)
      *ptr=hist[*ptr++];
  }

  /* handle color images */

  else {

    pix_val_t  *ptr;
    int         val;
    double      min[3], max[3], min_val, max_val, offset;
    double      sfact[3];
    double      post_offset, post_scale;
    int         chan, chan_cnt, loop_cnt;
    int         r_ival, g_ival, b_ival;
    int         adjust_flag;
    long        size, chunk, sum, i;
    long        hist[HIRES_GTBL_SZ];
    long        red_sum[HIRES_GTBL_SZ], grn_sum[HIRES_GTBL_SZ], blu_sum[HIRES_GTBL_SZ];
    long        red_tot, grn_tot, blu_tot;
    double      red_sum2[HIRES_GTBL_SZ], grn_sum2[HIRES_GTBL_SZ], blu_sum2[HIRES_GTBL_SZ];
    double      red_tot2, grn_tot2, blu_tot2;
    double      scale;
    double      rval, gval, bval;
    double      rval2, gval2, bval2;
    double      prob;
    double      lo_prob, md_prob, hi_prob;
    double      rdiff, gdiff, bdiff;
    double      lo_rmean, lo_gmean, lo_bmean;
    double      md_rmean, md_gmean, md_bmean;
    double      hi_rmean, hi_gmean, hi_bmean;
    double      lo_rvar, lo_gvar, lo_bvar;
    double      md_rvar, md_gvar, md_bvar;
    double      hi_rvar, hi_gvar, hi_bvar;
    double      lo_sum, md_sum, hi_sum;
    double      lo_rsum, lo_gsum, lo_bsum;
    double      md_rsum, md_gsum, md_bsum;
    double      hi_rsum, hi_gsum, hi_bsum;
    double      lo_rsum2, lo_gsum2, lo_bsum2;
    double      md_rsum2, md_gsum2, md_bsum2;
    double      hi_rsum2, hi_gsum2, hi_bsum2;
    double      init_rvar, init_gvar, init_bvar;
    double      lo_pow, md_pow, hi_pow;
    double      lo_part, md_part, hi_part;

    /* initialization */

    chan_cnt = img->chan_cnt;
    size = img->width * img->height;

    /* handle warnings */
    lo_part = md_part = hi_part = 0.0;

    /* build tables if not already built */

     build_hires_gray_tables ();

    /*** generate initial mean and variance estimates ***/

    /* note that this is needed or the algorithm my latch on to 
     * a group of red pixels which have zero g and b values 
     */

    /* Compute histogram */

    ptr = img->data;
    for (i = 0; i <  HIRES_GTBL_SZ; i++) {
      hist[i] = 0;
      red_sum[i] = grn_sum[i] = blu_sum[i] = 0;
      red_sum2[i] = grn_sum2[i] = blu_sum2[i] = 0.0;
    }
    red_tot2 = grn_tot2 = blu_tot2 = 0.0;
    for (i = 0; i < size; i++) {
      r_ival = *ptr++;
      g_ival = *ptr++;
      b_ival = *ptr++;
      val = red2hrgry[r_ival] + grn2hrgry[g_ival] + blu2hrgry[b_ival];
      if (val >  HIRES_GTBL_SZ-1)
	val = HIRES_GTBL_SZ-1;
      hist[val]++;
      red_sum[val]+= r_ival;
      grn_sum[val]+= g_ival;
      blu_sum[val]+= b_ival;
      red_sum2[val]+= (double)(r_ival * r_ival);
      grn_sum2[val]+= (double)(g_ival * g_ival);
      blu_sum2[val]+= (double)(b_ival * b_ival);
    }

    /*** generate initial estimates ***/


    /* find the min point */

    red_tot = grn_tot = blu_tot = 0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 0; i <  HIRES_GTBL_SZ; i++) {
      sum+=hist[i];
      red_tot+= red_sum[i];
      grn_tot+= grn_sum[i];
      blu_tot+= blu_sum[i];
      if (sum >= chunk) {
	break;
      }
    }
    lo_rmean = (double)red_tot / (double)sum;
    lo_gmean = (double)grn_tot / (double)sum;
    lo_bmean = (double)blu_tot / (double)sum;

    /* find the max point */

    red_tot = grn_tot = blu_tot = 0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i =  HIRES_GTBL_SZ-1; i >= 0; i--) {
      sum+=hist[i];
      red_tot+= red_sum[i];
      grn_tot+= grn_sum[i];
      blu_tot+= blu_sum[i];
      if (sum >= chunk) {
	break;
      }
    }
    hi_rmean = (double)red_tot / (double)sum;
    hi_gmean = (double)grn_tot / (double)sum;
    hi_bmean = (double)blu_tot / (double)sum;

    /* find the overall mean */

    red_tot = grn_tot = blu_tot = 0;
    red_tot2 = grn_tot2 = blu_tot2 = 0;
    for (i = 0; i <  HIRES_GTBL_SZ; i++) {
      red_tot+= red_sum[i];
      grn_tot+= grn_sum[i];
      blu_tot+= blu_sum[i];
      red_tot2+= red_sum2[i];
      grn_tot2+= grn_sum2[i];
      blu_tot2+= blu_sum2[i];
    }

    init_rvar = (red_tot2 - (double)red_tot * (double)red_tot / (double)size) / (double)size;
    init_gvar = (grn_tot2 - (double)grn_tot * (double)grn_tot / (double)size) / (double)size;
    init_bvar = (blu_tot2 - (double)blu_tot * (double)blu_tot / (double)size) / (double)size;
    if (init_rvar <= 1.0)
      init_rvar = 1.0;
    if (init_gvar <= 1.0)
      init_gvar = 1.0;
    if (init_bvar <= 1.0)
      init_bvar = 1.0;
    md_rmean = (double)red_tot / (double)size;
    md_gmean = (double)grn_tot / (double)size;
    md_bmean = (double)blu_tot / (double)size;
    lo_rvar = init_rvar * 0.125;
    lo_gvar = init_gvar * 0.125;
    lo_bvar = init_bvar * 0.125;
    md_rvar = init_rvar * 0.5;
    md_gvar = init_gvar * 0.5;
    md_bvar = init_bvar * 0.5;
    hi_rvar = init_rvar * 0.125;
    hi_gvar = init_gvar * 0.125;
    hi_bvar = init_bvar * 0.125;

    if (fast_flag) {
      lo_pow = hi_pow = 3.0;
      md_pow = 0.5;
    }
    else {
      lo_pow = hi_pow = 3.0;
      md_pow = 0.25;
    }


    /*** loop through and perform EM iterations ***/

    for (loop_cnt = -1; loop_cnt < ADJUST4_LOOP_CNT; loop_cnt++) {

      /* Compute the contributions from each pixel */

      if (loop_cnt >= 0) {
	lo_sum = md_sum = hi_sum = 0.0;
	lo_rsum = lo_gsum = lo_bsum = 0.0;
	md_rsum = md_gsum = md_bsum = 0.0;
	hi_rsum = hi_gsum = hi_bsum = 0.0;
	lo_rsum2 = lo_gsum2 = lo_bsum2 = 0.0;
	md_rsum2 = md_gsum2 = md_bsum2 = 0.0;
	hi_rsum2 = hi_gsum2 = hi_bsum2 = 0.0;

	/* just use the results stored in the histogram */
	if (fast_flag) {
	  for (i = 0; i <  HIRES_GTBL_SZ; i++) {

	    if (hist[i] == 0)
	      continue;

	    rval  = (double)red_sum[i]  / (double)hist[i];
	    gval  = (double)grn_sum[i]  / (double)hist[i];
	    bval  = (double)blu_sum[i]  / (double)hist[i];
	    rval2 = (double)red_sum2[i] / (double)hist[i];
	    gval2 = (double)grn_sum2[i] / (double)hist[i];
	    bval2 = (double)blu_sum2[i] / (double)hist[i];

	    rdiff = rval - lo_rmean;
	    gdiff = gval - lo_gmean;
	    bdiff = bval - lo_bmean;
	    lo_prob = exp (-0.5 * (rdiff * rdiff / lo_rvar 
				   + gdiff * gdiff / lo_gvar
				   + bdiff * bdiff / lo_bvar)
			   * lo_pow) * sqrt (lo_pow);
	    /* lo_prob = lo_prob * lo_prob * lo_prob * sqrt (3.0); */
	    lo_prob/= (sqrt (lo_rvar) * sqrt (lo_gvar) * sqrt (lo_bvar));

	    rdiff = rval - md_rmean;
	    gdiff = gval - md_gmean;
	    bdiff = bval - md_bmean;
	    md_prob = exp (-0.5 * (rdiff * rdiff / md_rvar 
				   + gdiff * gdiff / md_gvar
				   + bdiff * bdiff / md_bvar)
			   * md_pow) * sqrt (md_pow);
	    md_prob/= (sqrt (md_rvar) * sqrt (md_gvar) * sqrt (md_bvar));

	    rdiff = rval - hi_rmean;
	    gdiff = gval - hi_gmean;
	    bdiff = bval - hi_bmean;
	    hi_prob = exp (-0.5 * (rdiff * rdiff / hi_rvar 
				   + gdiff * gdiff / hi_gvar
				   + bdiff * bdiff / hi_bvar)
			   * hi_pow) * sqrt (hi_pow);
	    /* hi_prob = hi_prob * hi_prob * hi_prob * sqrt (3.0); */
	    hi_prob/= (sqrt (hi_rvar) * sqrt (hi_gvar) * sqrt (hi_bvar));

	    prob = lo_prob + md_prob + hi_prob;
	    if (prob == 0.0) {
	      prob = 3.0;
	      lo_prob = md_prob = hi_prob = 1.0;
	    }
	    lo_prob = lo_prob / prob * (double)hist[i];
	    md_prob = md_prob / prob * (double)hist[i];
	    hi_prob = hi_prob / prob * (double)hist[i];

	    lo_rsum+= rval * lo_prob;
	    lo_gsum+= gval * lo_prob;
	    lo_bsum+= bval * lo_prob;
	    lo_rsum2+= rval2 * lo_prob;
	    lo_gsum2+= gval2 * lo_prob;
	    lo_bsum2+= bval2 * lo_prob;
	    lo_sum+= lo_prob;

	    md_rsum+= rval * md_prob;
	    md_gsum+= gval * md_prob;
	    md_bsum+= bval * md_prob;
	    md_rsum2+= rval2 * md_prob;
	    md_gsum2+= gval2 * md_prob;
	    md_bsum2+= bval2 * md_prob;
	    md_sum+= md_prob;

	    hi_rsum+= rval * hi_prob;
	    hi_gsum+= gval * hi_prob;
	    hi_bsum+= bval * hi_prob;
	    hi_rsum2+= rval2 * hi_prob;
	    hi_gsum2+= gval2 * hi_prob;
	    hi_bsum2+= bval2 * hi_prob;
	    hi_sum+= hi_prob;
	  }
	}

	/* loop over all of the image pixels */
	else {
	  ptr = img->data;
	  for (i = 0; i < size; i++) {
	    rval = *ptr++;
	    gval = *ptr++;
	    bval = *ptr++;
	    rval2 = rval * rval;
	    gval2 = gval * gval;
	    bval2 = bval * bval;

	    rdiff = rval - lo_rmean;
	    gdiff = gval - lo_gmean;
	    bdiff = bval - lo_bmean;
	    lo_prob = exp (-0.5 * (rdiff * rdiff / lo_rvar 
				   + gdiff * gdiff / lo_gvar
				   + bdiff * bdiff / lo_bvar)
			   * lo_pow) * sqrt (lo_pow);
	    /* lo_prob = lo_prob * lo_prob * lo_prob * sqrt (3.0); */
	    lo_prob/= (sqrt (lo_rvar) * sqrt (lo_gvar) * sqrt (lo_bvar));

	    rdiff = rval - md_rmean;
	    gdiff = gval - md_gmean;
	    bdiff = bval - md_bmean;
	    md_prob = exp (-0.5 * (rdiff * rdiff / md_rvar 
				   + gdiff * gdiff / md_gvar
				   + bdiff * bdiff / md_bvar)
			   * md_pow) * sqrt (md_pow);
	    md_prob/= (sqrt (md_rvar) * sqrt (md_gvar) * sqrt (md_bvar));

	    rdiff = rval - hi_rmean;
	    gdiff = gval - hi_gmean;
	    bdiff = bval - hi_bmean;
	    hi_prob = exp (-0.5 * (rdiff * rdiff / hi_rvar 
				   + gdiff * gdiff / hi_gvar
				   + bdiff * bdiff / hi_bvar)
			   * hi_pow) * sqrt (hi_pow);
	    /* hi_prob = hi_prob * hi_prob * hi_prob * sqrt (3.0); */
	    hi_prob/= (sqrt (hi_rvar) * sqrt (hi_gvar) * sqrt (hi_bvar));

	    prob = lo_prob + md_prob + hi_prob;
	    if (prob == 0.0) {
	      prob = 3.0;
	      lo_prob = md_prob = hi_prob = 1.0;
	    }
	    lo_prob = lo_prob / prob;
	    md_prob = md_prob / prob;
	    hi_prob = hi_prob / prob;

	    lo_rsum+= rval * lo_prob;
	    lo_gsum+= gval * lo_prob;
	    lo_bsum+= bval * lo_prob;
	    lo_rsum2+= rval2 * lo_prob;
	    lo_gsum2+= gval2 * lo_prob;
	    lo_bsum2+= bval2 * lo_prob;
	    lo_sum+= lo_prob;

	    md_rsum+= rval * md_prob;
	    md_gsum+= gval * md_prob;
	    md_bsum+= bval * md_prob;
	    md_rsum2+= rval2 * md_prob;
	    md_gsum2+= gval2 * md_prob;
	    md_bsum2+= bval2 * md_prob;
	    md_sum+= md_prob;

	    hi_rsum+= rval * hi_prob;
	    hi_gsum+= gval * hi_prob;
	    hi_bsum+= bval * hi_prob;
	    hi_rsum2+= rval2 * hi_prob;
	    hi_gsum2+= gval2 * hi_prob;
	    hi_bsum2+= bval2 * hi_prob;
	    hi_sum+= hi_prob;
	  }
	}

	/* now calculate the updated means and variances */

	if (lo_sum == 0.0)
	  lo_sum = 1;
	if (md_sum == 0.0)
	  md_sum = 1;
	if (hi_sum == 0.0)
	  hi_sum = 1;
	prob = lo_sum + md_sum + hi_sum;
	lo_part = lo_sum / prob;
	md_part = md_sum / prob;
	hi_part = hi_sum / prob;
	/*
	if (lo_part * 2.0 > md_part) {
	  lo_pow*= 1.5;
	  fprintf (stderr, "Adjust lo_pow - increase (part)\n");
	}
	if (hi_part * 2.0 > md_part) {
	  hi_pow*= 1.5;
	  fprintf (stderr, "Adjust hi_pow - increase (part)\n");
	}
	*/

	md_rmean = md_rsum / md_sum;
	md_gmean = md_gsum / md_sum;
	md_bmean = md_bsum / md_sum;
	md_rvar  = md_rsum2 / md_sum - md_rmean * md_rmean;
	md_gvar  = md_gsum2 / md_sum - md_gmean * md_gmean;
	md_bvar  = md_bsum2 / md_sum - md_bmean * md_bmean;
	if (md_rvar <= 1.0)
	  md_rvar = 1.0;
	if (md_gvar <= 1.0)
	  md_gvar = 1.0;
	if (md_bvar <= 1.0)
	  md_bvar = 1.0;

	lo_rmean = lo_rsum / lo_sum;
	lo_gmean = lo_gsum / lo_sum;
	lo_bmean = lo_bsum / lo_sum;
	lo_rvar  = lo_rsum2 / lo_sum - lo_rmean * lo_rmean;
	lo_gvar  = lo_gsum2 / lo_sum - lo_gmean * lo_gmean;
	lo_bvar  = lo_bsum2 / lo_sum - lo_bmean * lo_bmean;
	if (loop_cnt < ADJUST4_LOOP_CNT-4) {
	  adjust_flag = 0;
	  if (lo_rvar <= 0.1) {
	    lo_rvar = (lo_gvar + lo_bvar) * 0.5;
	    adjust_flag = 1;
	  }
	  if (lo_gvar <= 0.1) {
	    lo_gvar = (lo_rvar + lo_bvar) * 0.5;
	    adjust_flag = 1;
	  }
	  if (lo_bvar <= 0.1) {
	    lo_bvar = (lo_rvar + lo_gvar) * 0.5;
	    adjust_flag = 1;
	  }
	  if (adjust_flag) {
	    lo_pow/= 1.5;
	    /*
	      lo_rvar = lo_gvar = lo_bvar =
	      var_adjust (lo_rvar, lo_gvar, lo_bvar);
	      */
	    /*
	    lo_rmean = lo_gmean = lo_bmean = 
	      lo_rmean * 0.3 + lo_gmean * 0.6 + lo_bmean * 0.1;
	      */
	    fprintf (stderr, "lo_var - too low, increasing\n");
	    fprintf (stderr, "lo_pow - decrease\n");
	  }
	}

	hi_rmean = hi_rsum / hi_sum;
	hi_gmean = hi_gsum / hi_sum;
	hi_bmean = hi_bsum / hi_sum;
	hi_rvar  = hi_rsum2 / hi_sum - hi_rmean * hi_rmean;
	hi_gvar  = hi_gsum2 / hi_sum - hi_gmean * hi_gmean;
	hi_bvar  = hi_bsum2 / hi_sum - hi_bmean * hi_bmean;
	if (loop_cnt < ADJUST4_LOOP_CNT-4) {
	  adjust_flag = 0;
	  if (hi_rvar <= 0.1) {
	    hi_rvar = (hi_gvar + hi_bvar) * 0.5;
	    adjust_flag = 1;
	  }
	  if (hi_gvar <= 0.1) {
	    hi_gvar = (hi_rvar + hi_bvar) * 0.5;
	    adjust_flag = 1;
	  }
	  if (hi_bvar <= 0.1) {
	    hi_bvar = (hi_rvar + hi_gvar) * 0.5;
	    adjust_flag = 1;
	  }
	  if (adjust_flag) {
	    hi_pow/= 1.5;
	    /*
	    hi_rvar = hi_gvar = hi_bvar =
	      hi_rvar * 0.3 * 0.3 + hi_gvar * 0.6 * 0.6 + hi_bvar * 0.1 * 0.1;
	      */
	    /*
	    hi_rvar = hi_gvar = hi_bvar = 
	      var_adjust (hi_rvar, hi_gvar, hi_bvar);
	    */
	    /*
	    hi_rmean = hi_gmean = hi_bmean = 
	      hi_rmean * 0.3 + hi_gmean * 0.6 + hi_bmean * 0.1;
	      */
	    fprintf (stderr, "hi_var - too low, increasing\n");
	    fprintf (stderr, "hi_pow - decrease\n");
	  }
	}

	/*
	if (loop_cnt != ADJUST4_LOOP_CNT-1) {
	  if (lo_rvar > md_rvar)
	    lo_rmean = 0.0;
	  if (lo_gvar > md_gvar)
	    lo_gmean = 0.0;
	  if (lo_bvar > md_bvar)
	    lo_bmean = 0.0;
	  if (hi_rvar > md_rvar)
	    hi_rmean = 255.0;
	  if (hi_gvar > md_gvar)
	    hi_gmean = 255.0;
	  if (hi_bvar > md_bvar)
	    hi_bmean = 255.0;
	}
	*/

	if (loop_cnt < ADJUST4_LOOP_CNT-4) {
	  adjust_flag = 0;
	  if (lo_rvar > md_rvar
	      || lo_rmean + sqrt (lo_rvar) > md_rmean) {
	    fprintf (stderr, "Adjust lo_rvar\n");
	    /* md_rmean = (md_rmean + lo_rmean) / 2.0; */
	    // md_rvar = lo_rvar;
	    /* lo_rmean = lo_gmean = lo_bmean = 
	      (lo_rmean * 0.3 + lo_gmean * 0.6 + lo_bmean * 0.1); */
	    // lo_rvar*= 0.5;
	    adjust_flag = 1;
	  }
	  if (lo_gvar > md_gvar
	      || lo_gmean + sqrt (lo_gvar) > md_gmean) {
	    fprintf (stderr, "Adjust lo_gvar\n");
	    /* md_gmean = (md_gmean + lo_gmean) / 2.0; */
	    // md_gvar = lo_gvar;
	    /* lo_rmean = lo_gmean = lo_bmean = 
	      (lo_rmean * 0.3 + lo_gmean * 0.6 + lo_bmean * 0.1); */
	    // lo_gvar*= 0.5;
	    adjust_flag = 1;
	  }
	  if (lo_bvar > md_bvar
	      || lo_bmean + sqrt (lo_bvar) > md_bmean) {
	    fprintf (stderr, "Adjust lo_bvar\n");
	    /* md_bmean = (md_bmean + lo_bmean) / 2.0; */
	    // md_bvar = lo_bvar;
	    /* lo_rmean = lo_gmean = lo_bmean = 
	      (lo_rmean * 0.3 + lo_gmean * 0.6 + lo_bmean * 0.1); */
	    // lo_bvar*= 0.5;
	    adjust_flag = 1;
	  }
	  if (!adjust_flag) {
	    if (lo_rmean - sqrt (lo_rvar) > md_rmean - sqrt (md_rvar)) {
	      fprintf (stderr, "Adjust lo_rvar 2\n");
	      // lo_rmean*= 0.5;
	      // lo_rvar*= 0.5;
	      // md_rvar*= 0.5;
	      // hi_rvar*= 0.5;
	      adjust_flag = 2;
	    }
	    if (lo_gmean - sqrt (lo_gvar) > md_gmean - sqrt (lo_gvar)) {
	      fprintf (stderr, "Adjust lo_gvar 2\n");
	      // lo_gmean*= 0.5;
	      // lo_gvar*= 0.5;
	      // md_gvar*= 0.5;
	      // hi_gvar*= 0.5;
	      adjust_flag = 2;
	    }
	    if (lo_bmean - sqrt (lo_bvar) > md_bmean - sqrt (lo_bvar)) {
	      fprintf (stderr, "Adjust lo_bvar 2\n");
	      // lo_bmean*= 0.5;
	      // lo_bvar*= 0.5;
	      // md_bvar*= 0.5;
	      // hi_bvar*= 0.5;
	      adjust_flag = 2;
	    }
	  }
	  if (adjust_flag == 1) {
	    lo_pow*= 1.5;
	    lo_rvar = lo_gvar = lo_bvar =
	      var_adjust (lo_rvar, lo_gvar, lo_bvar);

	    lo_rmean = lo_gmean = lo_bmean = 
	      lo_rmean * 0.3 + lo_gmean * 0.6 + lo_bmean * 0.1;

	    /*
	    md_rmean+= sqrt (lo_rvar) * 0.25;
	    md_gmean+= sqrt (lo_gvar) * 0.25;
	    md_bmean+= sqrt (lo_bvar) * 0.25;
	    */
	    fprintf (stderr, "Adjust lo_pow - increase\n");
	  }
	  if (adjust_flag == 2) {
	    lo_rvar = lo_gvar = lo_bvar =
	      var_adjust (lo_rvar, lo_gvar, lo_bvar);
	    lo_rmean = lo_gmean = lo_bmean = 
	      lo_rmean * 0.3 + lo_gmean * 0.6 + lo_bmean * 0.1 
	      - sqrt (lo_rvar) * 0.5;
	    // lo_pow/= 1.5; 
	    // fprintf (stderr, "Adjust lo_pow - decrease\n");
	    fprintf (stderr, "Adjust lo_mean and var\n");
	  }
	  adjust_flag = 0;
	  if (hi_rvar > md_rvar
	      || hi_rmean -  sqrt (hi_rvar) < md_rmean) {
	    fprintf (stderr, "Adjust hi_rvar\n");
	    /* md_rmean = (md_rmean + hi_rmean) / 2.0; */
	    // md_rvar = hi_rvar;
	    /* hi_rmean = hi_gmean = hi_bmean = 
	      (hi_rmean * 0.3 + hi_gmean * 0.6 + hi_bmean * 0.1); */
	    // hi_rvar*= 0.5;
	    adjust_flag = 1;
	  }
	  if (hi_gvar > md_gvar
	      || hi_gmean -  sqrt (hi_gvar) < md_gmean) {
	    fprintf (stderr, "Adjust hi_gvar\n");
	    /* md_gmean = (md_gmean + hi_gmean) / 2.0; */
	    // md_gvar = hi_gvar;
	    /* hi_rmean = hi_gmean = hi_bmean = 
	      (hi_rmean * 0.3 + hi_gmean * 0.6 + hi_bmean * 0.1); */
	    // hi_gvar*= 0.5;
	    adjust_flag = 1;
	  }
	  if (hi_bvar > md_bvar
	      || hi_bmean - sqrt (hi_bvar) < md_bmean) {
	    fprintf (stderr, "Adjust hi_bvar\n");
	    /* md_bmean = (md_bmean + hi_bmean) / 2.0; */
	    // md_bvar = hi_bvar;
	    /* hi_rmean = hi_gmean = hi_bmean = 
	      (hi_rmean * 0.3 + hi_gmean * 0.6 + hi_bmean * 0.1); */
	    // hi_bvar*= 0.5;
	    adjust_flag = 1;
	  }
	  if (!adjust_flag) {
	    if (hi_rmean + sqrt (hi_rvar) < md_rmean + sqrt (md_rvar)) {
	      fprintf (stderr, "Adjust hi_rvar 2\n");
	      // hi_rmean = 255.0 - (255.0 - hi_rmean) * 0.5;
	      // lo_rvar*= 0.5;
	      // md_rvar*= 0.5;
	      // hi_rvar*= 0.5;
	      adjust_flag = 2;
	    }
	    if (hi_gmean + sqrt (hi_gvar) < md_gmean + sqrt (hi_gvar)) {
	      fprintf (stderr, "Adjust hi_gvar 2\n");
	      // hi_gmean = 255.0 - (255.0 - hi_gmean) * 0.5;
	      // lo_gvar*= 0.5;
	      // md_gvar*= 0.5;
	      // hi_gvar*= 0.5;
	      adjust_flag = 2;
	    }
	    if (hi_bmean + sqrt (hi_bvar) < md_bmean + sqrt (hi_bvar)) {
	      fprintf (stderr, "Adjust hi_bvar 2\n");
	      // hi_bmean = 255.0 - (255.0 - hi_bmean) * 0.5;
	      // lo_bvar*= 0.5;
	      // md_bvar*= 0.5;
	      // hi_bvar*= 0.5;
	      adjust_flag = 2;
	    }
	  }	  
	  if (adjust_flag == 1) {
	    hi_pow*= 1.5;
	    fprintf (stderr, "Adjust hi_pow - increase\n");
	    hi_rvar = hi_gvar = hi_bvar = 
	      var_adjust (hi_rvar, hi_gvar, hi_bvar);

	    hi_rmean = hi_gmean = hi_bmean = 
 	      hi_rmean * 0.3 + hi_gmean * 0.6 + hi_bmean * 0.1; 

	    /*
	    md_rmean-= sqrt (hi_rvar) * 0.25;
	    md_gmean-= sqrt (hi_gvar) * 0.25;
	    md_bmean-= sqrt (hi_bvar) * 0.25;
	    */
	  }
	  if (adjust_flag == 2) {
	    hi_rvar = hi_gvar = hi_bvar = 
	      var_adjust (hi_rvar, hi_gvar, hi_bvar);
	    hi_rmean = hi_gmean = hi_bmean = 
	      hi_rmean * 0.3 + hi_gmean * 0.6 + hi_bmean * 0.1 + sqrt (hi_rvar) * 0.5;
	    // hi_pow/= 1.5; 
	    // fprintf (stderr, "Adjust hi_pow - decrease\n");
	    fprintf (stderr, "Adjust hi_mean and var\n");
	  }
	}

	if (lo_rvar < 0.1)
	  lo_rvar = 0.1;
	if (lo_gvar < 0.1) 
	  lo_gvar = 0.1;
	if (lo_bvar < 0.1) 
	  lo_bvar = 0.1;

	if (hi_rvar < 0.1) 
	  hi_rvar = 0.1;
	if (hi_gvar < 0.1) 
	  hi_gvar = 0.1;
	if (hi_bvar < 0.1) 
	  hi_bvar = 0.1;

      }

      fprintf (stderr, 
	       "Iter%3d: L Mean: r=%5.1f, g=%5.1f, b=%5.1f; L Var: r=%6.1f, g=%6.1f, b=%6.1f\n",
	       loop_cnt + 1 ,
	       lo_rmean, lo_gmean, lo_bmean,
	       lo_rvar, lo_gvar, lo_bvar);
      fprintf (stderr, 
	       "         M Mean: r=%5.1f, g=%5.1f, b=%5.1f; M Var: r=%6.1f, g=%6.1f, b=%6.1f\n",
	       md_rmean, md_gmean, md_bmean,
	       md_rvar, md_gvar, md_bvar);
      fprintf (stderr, 
	       "         H Mean: r=%5.1f, g=%5.1f, b=%5.1f; H Var: r=%6.1f, g=%6.1f, b=%6.1f\n",
	       hi_rmean, hi_gmean, hi_bmean,
	       hi_rvar, hi_gvar, hi_bvar);

      fprintf (stderr,
	       "          Parts: l=%5.3f, m=%5.3f, h=%5.3f\n",
	       lo_part, md_part, hi_part);
    }

    /* check for failure */
    if (lo_rmean + sqrt (lo_rvar) * 0.1 > md_rmean 
	|| lo_gmean + sqrt (lo_gvar) * 0.1 > md_gmean
	|| lo_bmean + sqrt (lo_bvar) * 0.1 > md_bmean) {
      fprintf (stderr, "Low cluster problem, aborting adjustment\n");
      return;
    }
    if (hi_rmean - sqrt (hi_rvar) * 0.1 < md_rmean 
	|| hi_gmean - sqrt (hi_gvar) * 0.1 < md_gmean
	|| hi_bmean - sqrt (hi_bvar) * 0.1 < md_bmean) {
      fprintf (stderr, "High cluster problem, aborting adjustment\n");
      return;
    }

    /* compute the separate channel adjustments */

    /* full black adjustment */
    if (black_adj_mode == 1
	|| black_adj_mode == 4
	|| black_adj_mode == 5
	|| black_adj_mode == 6) {
      
      double lo_mean, md_mean, hi_mean; 
      double a0_var, a1_var, a2_var, tot_var;
      double a0_mean;
      /* double a0_mean, a1_mean, a2_mean; */
      double scale_factor;
      double red_offset[9], grn_offset[9], blu_offset[9];
      double red_scale[9], grn_scale[9], blu_scale[9];
      double adj_var[9];
      double max_range;
      /* double min_range, max_range; */
      long   j, k, adj_ind[3];

      lo_mean = lo_rmean * 0.3 + lo_gmean * 0.6 + lo_bmean * 0.1;
      md_mean = md_rmean * 0.3 + md_gmean * 0.6 + md_bmean * 0.1;
      hi_mean = hi_rmean * 0.3 + hi_gmean * 0.6 + hi_bmean * 0.1;

      /* use the low and high points for calculations */
      red_offset[0] = lo_rmean;
      grn_offset[0] = lo_gmean;
      blu_offset[0] = lo_bmean;
      red_scale[0]  = 1.0 / (hi_rmean - lo_rmean);
      grn_scale[0]  = 1.0 / (hi_gmean - lo_gmean);
      blu_scale[0]  = 1.0 / (hi_bmean - lo_bmean);

      /* use the low and mid points for calculations */
      red_offset[1] = lo_rmean;
      grn_offset[1] = lo_gmean;
      blu_offset[1] = lo_bmean;
      scale_factor  = (md_mean - lo_mean) / (hi_mean - lo_mean);
      red_scale[1]  = scale_factor / (md_rmean - lo_rmean);
      grn_scale[1]  = scale_factor / (md_gmean - lo_gmean);
      blu_scale[1]  = scale_factor / (md_bmean - lo_bmean);

      /* use the mid and high points for calculations */
      scale_factor  = (hi_mean - md_mean) / (hi_mean - lo_mean);
      red_scale[2]  = scale_factor / (hi_rmean - md_rmean);
      grn_scale[2]  = scale_factor / (hi_gmean - md_gmean);
      blu_scale[2]  = scale_factor / (hi_bmean - md_bmean);
      scale_factor  = (md_mean - lo_mean) / (hi_mean - lo_mean);
      red_offset[2] = md_rmean - scale_factor / red_scale[2];
      grn_offset[2] = md_gmean - scale_factor / grn_scale[2];
      blu_offset[2] = md_bmean - scale_factor / blu_scale[2];

      /* use the low and high points for calculations */
      /* hi replaced by average */
      red_offset[3] = lo_rmean;
      grn_offset[3] = lo_gmean;
      blu_offset[3] = lo_bmean;
      red_scale[3]  = 1.0 / (hi_mean - lo_rmean);
      grn_scale[3]  = 1.0 / (hi_mean - lo_gmean);
      blu_scale[3]  = 1.0 / (hi_mean - lo_bmean);

      /* use the low and high points for calculations */
      /* lo replaced by average */
      red_offset[4] = lo_mean;
      grn_offset[4] = lo_mean;
      blu_offset[4] = lo_mean;
      red_scale[4]  = 1.0 / (hi_rmean - lo_mean);
      grn_scale[4]  = 1.0 / (hi_gmean - lo_mean);
      blu_scale[4]  = 1.0 / (hi_bmean - lo_mean);

      /* use the low and mid points for calculations */
      /* md replaced by average */
      red_offset[5] = lo_rmean;
      grn_offset[5] = lo_gmean;
      blu_offset[5] = lo_bmean;
      scale_factor  = (md_mean - lo_mean) / (hi_mean - lo_mean);
      red_scale[5]  = scale_factor / (md_mean - lo_rmean);
      grn_scale[5]  = scale_factor / (md_mean - lo_gmean);
      blu_scale[5]  = scale_factor / (md_mean - lo_bmean);

      /* use the low and mid points for calculations */
      /* lo replaced by average */
      red_offset[6] = lo_mean;
      grn_offset[6] = lo_mean;
      blu_offset[6] = lo_mean;
      scale_factor  = (md_mean - lo_mean) / (hi_mean - lo_mean);
      red_scale[6]  = scale_factor / (md_rmean - lo_mean);
      grn_scale[6]  = scale_factor / (md_gmean - lo_mean);
      blu_scale[6]  = scale_factor / (md_bmean - lo_mean);

      /* use the mid and high points for calculations */
      /* hi replaced by mean */
      scale_factor  = (hi_mean - md_mean) / (hi_mean - lo_mean);
      red_scale[7]  = scale_factor / (hi_mean - md_rmean);
      grn_scale[7]  = scale_factor / (hi_mean - md_gmean);
      blu_scale[7]  = scale_factor / (hi_mean - md_bmean);
      scale_factor  = (md_mean - lo_mean) / (hi_mean - lo_mean);
      red_offset[7] = md_rmean - scale_factor / red_scale[7];
      grn_offset[7] = md_gmean - scale_factor / grn_scale[7];
      blu_offset[7] = md_bmean - scale_factor / blu_scale[7];

      /* use the mid and high points for calculations */
      /* md replaced by mean */
      scale_factor  = (hi_mean - md_mean) / (hi_mean - lo_mean);
      red_scale[8]  = scale_factor / (hi_rmean - md_mean);
      grn_scale[8]  = scale_factor / (hi_gmean - md_mean);
      blu_scale[8]  = scale_factor / (hi_bmean - md_mean);
      scale_factor  = (md_mean - lo_mean) / (hi_mean - lo_mean);
      red_offset[8] = md_mean - scale_factor / red_scale[8];
      grn_offset[8] = md_mean - scale_factor / grn_scale[8];
      blu_offset[8] = md_mean - scale_factor / blu_scale[8];

      for (i = 0; i < 9; i++) {
	a0_mean = (red_scale[i] + grn_scale[i] + blu_scale[i]) / 3.0;
	a0_var = (fabs (red_scale[i] - a0_mean) 
		  + fabs (grn_scale[i] - a0_mean)
		  + fabs (blu_scale[i] - a0_mean)) / 3.0;
	fprintf (stderr, "offset[%ld]=(%.1f,%.1f,%.1f), scale[%ld]=(%.5f,%.5f,%.5f), var=%f\n",
		 i, red_offset[i], grn_offset[i], blu_offset[i],
		 i, red_scale[i], grn_scale[i], blu_scale[i],
		 a0_var);
	adj_var[i] = a0_var;
      }

      for (i = 0; i < 3; i++) {
	adj_ind[i] = i;
	j = i * 2 + 3;
	for (k = j; k <= j + 1; k++) {
	  if (k == 5 || k == 8)
	    continue;
	  if (adj_var[k] < adj_var[i]) {
	    adj_var[i] = adj_var[k];
	    red_scale[i] = red_scale[k];
	    grn_scale[i] = grn_scale[k];
	    blu_scale[i] = blu_scale[k];
	    red_offset[i] = red_offset[k];
	    grn_offset[i] = grn_offset[k];
	    blu_offset[i] = blu_offset[k];
	    adj_ind[i] = k;
	  }
	}
      }


      /*
      a0_mean = (red_scale[0] + grn_scale[0] + blu_scale[0]) / 3.0;
      a1_mean = (red_scale[1] + grn_scale[1] + blu_scale[1]) / 3.0;
      a2_mean = (red_scale[2] + grn_scale[2] + blu_scale[2]) / 3.0;
      */
      /*
      a0_var = (fabs (red_scale[0] - a0_mean) 
		+ fabs (grn_scale[0] - a0_mean)
		+ fabs (blu_scale[0] - a0_mean)) / 3.0;
      a1_var = (fabs (red_scale[1] - a1_mean) 
		+ fabs (grn_scale[1] - a1_mean)
		+ fabs (blu_scale[1] - a1_mean)) / 3.0;
      a2_var = (fabs (red_scale[2] - a2_mean) 
		+ fabs (grn_scale[2] - a2_mean)
		+ fabs (blu_scale[2] - a2_mean)) / 3.0;
		*/
      a0_var = adj_var[0];
      a1_var = adj_var[1];
      a2_var = adj_var[2];

      if (!fix_flag) {
	adj_ind[0] = 0;
	adj_ind[1] = 1;
	adj_ind[2] = 2;
      }
      fprintf (stderr, "Ind: a0: %ld, a1: %ld, a2: %ld\n",
	       adj_ind[0], adj_ind[1], adj_ind[2]);

      fprintf (stderr, "Var: a0: %g, a1: %g, a2: %g\n",
	       a0_var, a1_var, a2_var);

      if (a0_var < 1.0e-6)
	a0_var = 1.0e-6;
      if (a1_var < 1.0e-6)
	a1_var = 1.0e-6;
      if (a2_var < 1.0e-6)
	a2_var = 1.0e-6;
      a0_var = 1.0 / a0_var;
      a1_var = 1.0 / a1_var;
      a2_var = 1.0 / a2_var;
      /*
      a0_var = 1.0 / sqrt (a0_var);
      a1_var = 1.0 / sqrt (a1_var);
      a2_var = 1.0 / sqrt (a2_var);
      */
      tot_var = a0_var + a1_var + a2_var;
      a0_var = a0_var / tot_var;
      a1_var = a1_var / tot_var;
      a2_var = a2_var / tot_var;

      /* debug */
      /*
      a0_var = 0.0;
      a1_var = 0.0;
      a2_var = 1.0;
      */

      fprintf (stderr, "Mix: a0: %g, a1: %g, a2: %g\n",
	       a0_var, a1_var, a2_var);

      switch (black_adj_mode) {
      case 4:
	a0_var = 1.0;
	a1_var = 0.0;
	a2_var = 0.0;
	break;
      case 5:
	a0_var = 0.0;
	a1_var = 1.0;
	a2_var = 0.0;
	break;
      case 6:
	a0_var = 0.0;
	a1_var = 0.0;
	a2_var = 1.0;
	break;
      }

      if (black_adj_mode != 1) {
	fprintf (stderr, "Forced mix: a0: %g, a1: %g, a2: %g\n",
		 a0_var, a1_var, a2_var);
      }

      min[0] = (red_offset[0]   * red_scale[0] * a0_var
		+ red_offset[1] * red_scale[1] * a1_var
		+ red_offset[2] * red_scale[2] * a2_var);
      min[1] = (grn_offset[0]   * grn_scale[0] * a0_var
		+ grn_offset[1] * grn_scale[1] * a1_var
		+ grn_offset[2] * grn_scale[2] * a2_var);
      min[2] = (blu_offset[0]   * blu_scale[0] * a0_var
		+ blu_offset[1] * blu_scale[1] * a1_var
		+ blu_offset[2] * blu_scale[2] * a2_var);
      sfact[0] = (  red_scale[0] * a0_var
		  + red_scale[1] * a1_var
		  + red_scale[2] * a2_var);
      sfact[1] = (  grn_scale[0] * a0_var
		  + grn_scale[1] * a1_var
		  + grn_scale[2] * a2_var);
      sfact[2] = (  blu_scale[0] * a0_var
		  + blu_scale[1] * a1_var
		  + blu_scale[2] * a2_var);
      min[0]/= sfact[0];
      min[1]/= sfact[1];
      min[2]/= sfact[2];
      max[0] = min[0] + 1.0 / sfact[0];
      max[1] = min[1] + 1.0 / sfact[1];
      max[2] = min[2] + 1.0 / sfact[2];

      /* decide on final scaling factor */
      {
	double rlo, glo, blo;
	double rhi, ghi, bhi;
	double lo_red, lo_grn, lo_blu;
	double hi_red, hi_grn, hi_blu;
	double min_lo, max_hi;

	rlo = lo_rmean - sqrt (lo_rvar) * 1.5;
	glo = lo_gmean - sqrt (lo_gvar) * 1.5;
	blo = lo_bmean - sqrt (lo_bvar) * 1.5;
	min_lo = rlo;
	if (glo < min_lo)
	  min_lo = glo;
	if (blo < min_lo)
	  min_lo = blo;

	rhi = hi_rmean + sqrt (hi_rvar) * 1.5;
	ghi = hi_gmean + sqrt (hi_gvar) * 1.5;
	bhi = hi_bmean + sqrt (hi_bvar) * 1.5;
	max_hi = rhi;
	if (ghi > max_hi)
	  max_hi = ghi;
	if (bhi > max_hi)
	  max_hi = bhi;

	lo_red = lo_grn = lo_blu = 0.0;
	hi_red = hi_grn = hi_blu = 0.0;
	switch (adj_ind[0]) {
	case 0:		/* standard lo and hi */
	  lo_red+= rlo * a0_var;
	  lo_grn+= glo * a0_var;
	  lo_blu+= blo * a0_var;
	  hi_red+= rhi * a0_var;
	  hi_grn+= ghi * a0_var;
	  hi_blu+= bhi * a0_var;
	  break;
	case 3:		/* hi replaced by average */
	  lo_red+= rlo * a0_var;
	  lo_grn+= glo * a0_var;
	  lo_blu+= blo * a0_var;
	  hi_red+= max_hi * a0_var;
	  hi_grn+= max_hi * a0_var;
	  hi_blu+= max_hi * a0_var;
	  break;
	case 4:		/* lo replaced by average */
	  lo_red+= min_lo * a0_var;
	  lo_grn+= min_lo * a0_var;
	  lo_blu+= min_lo * a0_var;
	  hi_red+= rhi * a0_var;
	  hi_grn+= ghi * a0_var;
	  hi_blu+= bhi* a0_var;
	  break;
	}
	switch (adj_ind[1]) {
	case 1:		/* standard lo and md */
	case 5:		 /* md replaced by average */
	  lo_red+= rlo * a1_var;
	  lo_grn+= glo * a1_var;
	  lo_blu+= blo * a1_var;
	  hi_red+= rhi * a1_var;
	  hi_grn+= ghi * a1_var;
	  hi_blu+= bhi * a1_var;
	  break;
	case 6:		 /* lo replaced by average */
	  lo_red+= min_lo * a1_var;
	  lo_grn+= min_lo * a1_var;
	  lo_blu+= min_lo * a1_var;
	  hi_red+= rhi * a1_var;
	  hi_grn+= ghi * a1_var;
	  hi_blu+= bhi * a1_var;
	  break;
	}
	switch (adj_ind[2]) {
	case 2:		/* standard md and hi */
	case 8:		/* md replaced by mean */
	  lo_red+= rlo * a2_var;
	  lo_grn+= glo * a2_var;
	  lo_blu+= blo * a2_var;
	  hi_red+= rhi * a2_var;
	  hi_grn+= ghi * a2_var;
	  hi_blu+= bhi * a2_var;
	  break;
	case 7:		 /* hi replaced by mean */
	  lo_red+= rlo * a2_var;
	  lo_grn+= glo * a2_var;
	  lo_blu+= blo * a2_var;
	  hi_red+= max_hi * a2_var;
	  hi_grn+= max_hi * a2_var;
	  hi_blu+= max_hi * a2_var;
	  break;
	}

	post_offset = (min[0] - lo_red + min[1] - lo_grn + min[2] - lo_blu) 
	  / 3.0;
	/* to each of the lo_xmean's I need to add post_offset
	 * and subtract it, so the net result is that I do
	 * nothing */
	/*
	min_range = ((lo_red - min[0]) * sfact[0]
		     + (lo_grn - min[1]) * sfact[1]
		     + (lo_blu - min[2]) * sfact[2]) / 3.0;
		     */
	max_range = ((hi_red + post_offset - min[0]) * sfact[0]
		     + (hi_grn + post_offset - min[1]) * sfact[1] 
		     + (hi_blu  + post_offset - min[2]) * sfact[2]) 
	  / 3.0;
	post_scale = 1.0 / max_range;
      }
    }

    /* simple black adjustment */
    else if (black_adj_mode == 2) {
      min[0] = lo_rmean - sqrt (lo_rvar);
      min[1] = lo_gmean - sqrt (lo_gvar);
      min[2] = lo_bmean - sqrt (lo_bvar);
      if (min[1] < min[0])
	min[0] = min[1];
      if (min[2] < min[0])
	min[0] = min[2];
      min[2] = min[1] = min[0];
      max[0] = hi_rmean + sqrt (hi_rvar);
      max[1] = hi_gmean + sqrt (hi_gvar);
      max[2] = hi_bmean + sqrt (hi_bvar);

      post_offset = sqrt (lo_rvar);
      if (sqrt (lo_gvar) > post_offset)
	post_offset = sqrt (lo_gvar);
      if (sqrt (lo_bvar) > post_offset)
	post_offset = sqrt (lo_bvar);
      /* post_offset*= 1.5; */
      post_scale = sqrt (hi_rvar);
      if (sqrt (hi_gvar) > post_scale)
	post_scale = sqrt (hi_gvar);
      if (sqrt (hi_bvar) > post_scale)
	post_scale = sqrt (hi_bvar);
      post_scale = 255.0 / (255.0 + post_scale + post_offset);
    }

    /* variance based black adjustment */
    else if (black_adj_mode == 3) {
      min[0] = lo_rmean - sqrt (lo_rvar);
      min[1] = lo_gmean - sqrt (lo_gvar);
      min[2] = lo_bmean - sqrt (lo_bvar);
      post_offset = sqrt (lo_rvar);
      if (sqrt (lo_gvar) > post_offset)
	post_offset = sqrt (lo_gvar);
      if (sqrt (lo_bvar) > post_offset)
	post_offset = sqrt (lo_bvar);
      /* post_offset*= 1.5; */
      max[0] = hi_rmean + sqrt (hi_rvar);
      max[1] = hi_gmean + sqrt (hi_gvar);
      max[2] = hi_bmean + sqrt (hi_bvar);
      post_scale = sqrt (hi_rvar);
      if (sqrt (hi_gvar) > post_scale)
	post_scale = sqrt (hi_gvar);
      if (sqrt (hi_bvar) > post_scale)
	post_scale = sqrt (hi_bvar);
      post_scale = 255.0 / (255.0 + post_scale + post_offset);
    }

    /* no black adjustment */
    else {
      min[0] = 0.0;
      min[1] = 0.0;
      min[2] = 0.0;
      post_offset = 0.0;
      max[0] = hi_rmean + sqrt (hi_rvar);
      max[1] = hi_gmean + sqrt (hi_gvar);
      max[2] = hi_bmean + sqrt (hi_bvar);
      post_scale = sqrt (hi_rvar);
      if (sqrt (hi_gvar) > post_scale)
	post_scale = sqrt (hi_gvar);
      if (sqrt (hi_bvar) > post_scale)
	post_scale = sqrt (hi_bvar);
      post_scale = 255.0 / (255.0 + post_scale + post_offset);
    }

    fprintf (stderr, 
	     "Final: r=[%.1f,%.1f]->[%.1f,%.1f], g=[%.1f,%.1f]->[%.1f,%.1f], b=[%.1f,%.1f]->[%.1f,%.1f]\n", 
	     min[0], max[0], post_offset, post_scale * 255.0 + post_offset,
	     min[1], max[1], post_offset, post_scale * 255.0 + post_offset,
	     min[2], max[2], post_offset, post_scale * 255.0 + post_offset);

    /* loop through the color channels */

    for (chan = 0; chan < chan_cnt; chan++) {

      min_val = min[chan];
      max_val = max[chan];

      /* check for zero range */
  
      if (max_val == min_val)
	continue;

      /* Compute the transformation map */

      scale = (255.0 / (max_val - min_val)) * post_scale;
      offset = min_val - post_offset;
      for (i = 0; i < 256; i++) {
	hist[i] = (int)(((double)i - offset) * scale + 0.5);
	if (hist[i] > 255) hist[i] = 255;
	if (hist[i] < 0)   hist[i] = 0;
      }

      /* Apply transformation to image */

      ptr = img->data + chan;
      for (i=0; i<size; i++, ptr+= chan_cnt) 
	*ptr=hist[*ptr];
    }
  }

  return;
}


/*
 * Histogram dynamic range adjustment of an image
 *
 * This is EM based.  Starts with a hump and at the min, mid, and max.
 *
 * This version runs EM in the grayscale domain and then uses 
 * color information captured for each gray level to do the
 * adjustment.
 *
 * The monochrome implementation was never completed.
 *
 * Takes bottom 0.1% and top 0.1% of the image pixels are used for
 * initial means estimates.
 *
 * This image treats monochrome images the same as
 * img_hist_range_adjust.
 *
 * In the case of color images, it histograms the grayscale
 * version of the image.  Then it maps all three channels
 * indepdently based on the average lightest and darkest colors
 * for that gray level.
 *
 * The original data is modified.
 *
 */

#define ADJUST5_LOOP_CNT 20

void img_hist_range_adjust5 (img_t *img){
  img_hist_range_adjust5_internal (img, 1);
}

void img_hist_range_adjust5_white_only (img_t *img) {
  img_hist_range_adjust5_internal (img, 0);
}

void img_hist_range_adjust5_internal (img_t *img, int black_adj_flag)
{

  /* TO DO: complete code to handle monochrome images */

  /* handle monochrome images */

  if (img->chan_cnt == 1) {

    long       size;
    pix_val_t  *ptr, min_val, max_val;
    long        hist[256];
    long        sum, i;
    long        chunk;
    double      scale, tot, tot2, init_var;
    double      lo_mean, md_mean, hi_mean;
    double      lo_var, md_var, hi_var;

    fprintf (stderr, "WARNING: monochrome processing implementation not complete.\n");

    /* initialization */

    size = img->width * img->height;

    /* to remove warnings */

    /* Compute histogram */

    ptr = img->data;
    for (i = 0; i < 256; i++)
      hist[i] = 0;
    for (i = 0; i < size; i++)
      hist[*ptr++]++;

    /*** generate initial mean and variance estimates ***/

    /* Find the min and max points */

    min_val = 0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 0; i < 256; i++) {
      sum+=hist[i];
      if (sum >= chunk) {
	min_val = i;
	break;
      }
    }
    max_val = 255;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 255; i >= 0; i--) {
      sum+=hist[i];
      if (sum >= chunk) {
	max_val = i;
	break;
      }
    }
    tot = tot2 = 0.0;
    for (i = 0; i < 256; i++) {
      tot+= (double)hist[i] * (double)i;
      tot2+= (double)hist[i] * (double)hist[i] * (double)i;
    }

    /* generate initial estimates */

    init_var = (tot2 - tot * tot / (double)size) / (double)size;
    if (init_var <= 1.0)
      init_var = 1.0;
    lo_mean = (double)min_val;
    md_mean = tot / (double)size;
    hi_mean = (double)max_val;
    lo_var = init_var * 0.125;
    md_var = init_var * 0.25;
    hi_var = init_var * 0.125;

    fprintf (stderr, "Min val: %d, Max val: %d\n", min_val, max_val);

    /* check for zero range */
  
    if (max_val == min_val)
      return;

    /* Compute the transformation map */

    scale = 255.0 / (double)(max_val - min_val);
    for (i = 0; i < 256; i++) {
      hist[i] = (int)((double)(i - min_val) * scale + 0.5);
      if (hist[i] > 255) hist[i] = 255;
      if (hist[i] < 0)   hist[i] = 0;
    }

    /* Apply transformation to image */

    ptr = img->data;
    for (i=0; i<size; i++)
      *ptr=hist[*ptr++];
  }

  /* handle color images */

  else {

    pix_val_t  *ptr;
    int         min[3], max[3], min_val, max_val, ival;
    int         min_tar, max_tar, offset;
    int         chan, chan_cnt, loop_cnt;
    int         r_ival, g_ival, b_ival;
    long        size, chunk, sum, i;
    long        hist[HIRES_GTBL_SZ];
    double      val, val2;
    double      red_sum[HIRES_GTBL_SZ], grn_sum[HIRES_GTBL_SZ], blu_sum[HIRES_GTBL_SZ];
    double      red_sum2[HIRES_GTBL_SZ], grn_sum2[HIRES_GTBL_SZ], blu_sum2[HIRES_GTBL_SZ];
    double      tot, tot2;
    double      scale, diff;
    double      rval, gval, bval;
    double      rval2, gval2, bval2;
    double      prob, init_var;
    double      lo_prob, md_prob, hi_prob;
    double      lo_mean,  md_mean, hi_mean;
    double      lo_var, md_var, hi_var;  
    double      lo_tot, md_tot, hi_tot;
    double      lo_tot2, md_tot2, hi_tot2;
    double      lo_rmean, lo_gmean, lo_bmean;
    double      md_rmean, md_gmean, md_bmean;
    double      hi_rmean, hi_gmean, hi_bmean;
    double      lo_rvar, lo_gvar, lo_bvar;
    double      md_rvar, md_gvar, md_bvar;
    double      hi_rvar, hi_gvar, hi_bvar;
    double      lo_sum, md_sum, hi_sum;
    double      lo_rtot, lo_gtot, lo_btot;
    double      md_rtot, md_gtot, md_btot;
    double      hi_rtot, hi_gtot, hi_btot;
    double      lo_rtot2, lo_gtot2, lo_btot2;
    double      md_rtot2, md_gtot2, md_btot2;
    double      hi_rtot2, hi_gtot2, hi_btot2;
    double      rdiff, gdiff, bdiff;
    double      rmean, gmean, bmean;

    /* initialization */

    chan_cnt = img->chan_cnt;
    size = img->width * img->height;

    /* build tables if not already built */

    build_hires_gray_tables ();

    /*** generate initial mean and variance estimates ***/

    /* note that this is needed or the algorithm my latch on to 
     * a group of red pixels which have zero g and b values 
     */

    /* Compute histogram */

    for (i = 0; i < HIRES_GTBL_SZ; i++) {
      hist[i] = 0;
      red_sum[i] = grn_sum[i] = blu_sum[i] = 0.0;
      red_sum2[i] = grn_sum2[i] = blu_sum2[i] = 0.0;
    }
    ptr = img->data;
    for (i = 0; i < size; i++) {
      r_ival = *ptr++;
      g_ival = *ptr++;
      b_ival = *ptr++;
      ival = red2hrgry[r_ival] + grn2hrgry[g_ival] + blu2hrgry[b_ival];
      if (ival >  HIRES_GTBL_SZ-1) {
	ival =  HIRES_GTBL_SZ-1;
      }
      /* fprintf (stderr, "%d %d %d %d\n", r_ival, g_ival, b_ival, ival); */
      hist[ival]++;
      red_sum[ival]+= (double)r_ival;
      grn_sum[ival]+= (double)g_ival;
      blu_sum[ival]+= (double)b_ival;
      red_sum2[ival]+= (double)(r_ival * r_ival);
      grn_sum2[ival]+= (double)(g_ival * g_ival);
      blu_sum2[ival]+= (double)(b_ival * b_ival);
    }

    /*** generate initial estimates ***/


    /* find the min point */

    tot = 0.0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 0; i < HIRES_GTBL_SZ; i++) {
      sum+= hist[i];
      tot+= (double)(hist[i] * i);
      if (sum >= chunk) {
	break;
      }
    }
    lo_mean = tot / (double)sum;

    /* find the max point */

    tot = 0.0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = HIRES_GTBL_SZ-1; i >= 0; i--) {
      sum+=hist[i];
      tot+= (double)(hist[i] * i);
      if (sum >= chunk) {
	break;
      }
    }
    hi_mean = tot / (double)sum;

    /* find the overall mean */

    tot = 0.0;
    tot2 = 0.0;
    for (i = 0; i < HIRES_GTBL_SZ; i++) {
      tot+= (double)hist[i] * (double)i;
      tot2+= (double)hist[i] * (double)i * (double)i;
    }
    md_mean = tot / (double)size;

    init_var = (tot2 - tot * tot / (double)size) / (double)size;
    if (init_var <= 1.0)
      init_var =  HIRES_GTBL_SZ;
    lo_var = init_var * 0.25; /* 0.125; */
    md_var = init_var * 0.5;
    hi_var = init_var * 0.25; /* 0.125; */

    /*** loop through and perform EM iterations ***/

    for (loop_cnt = -1; loop_cnt < ADJUST5_LOOP_CNT;  loop_cnt++) {

      /* Compute the contributions from each pixel */

      if (loop_cnt >= 0) {

	lo_sum = md_sum = hi_sum = 0.0;
	lo_tot = md_tot = hi_tot = 0.0;
	lo_tot2 = md_tot2 = hi_tot2 = 0.0;

	for (i = 0; i < HIRES_GTBL_SZ; i++) {
	  val = (double)i;
	  val2 = val * val;

	  diff = val - lo_mean;
	  lo_prob = exp (-0.5 * (diff * diff / lo_var));
	  /* lo_prob = lo_prob * lo_prob * sqrt (2.0); */
	  lo_prob = lo_prob * lo_prob * lo_prob * sqrt (3.0);
	  lo_prob/= sqrt (lo_var);

	  diff = val - md_mean;
	  md_prob = exp (-0.5 * (diff * diff / md_var)) / sqrt (md_var);

	  diff = val - hi_mean;
	  hi_prob = exp (-0.5 * (diff * diff / hi_var));
	  /* hi_prob = hi_prob * hi_prob * sqrt (2.0); */
	  hi_prob = hi_prob * hi_prob * hi_prob * sqrt (3.0);
	  hi_prob/= sqrt (hi_var);

	  prob = lo_prob + md_prob + hi_prob;
	  lo_prob = lo_prob / prob * hist[i];
	  md_prob = md_prob / prob * hist[i];
	  hi_prob = hi_prob / prob * hist[i];

	  lo_tot+= val * lo_prob;
	  lo_tot2+= val2 * lo_prob;
	  lo_sum+= lo_prob;

	  md_tot+= val * md_prob;
	  md_tot2+= val2 * md_prob;
	  md_sum+= md_prob;

	  hi_tot+= val * hi_prob;
	  hi_tot2+= val2 * hi_prob;
	  hi_sum+= hi_prob;
	}

	/* now calculate the updated means and variances */

	if (lo_sum == 0.0)
	  lo_sum = 1.0;
	lo_mean = lo_tot / lo_sum;
	lo_var  = lo_tot2 / lo_sum - lo_mean * lo_mean;
	if (lo_var <= 1.0) {
	  /*
	  lo_var = md_var * 0.125 * 0.5;
	   lo_mean*= 0.5;
	  if (lo_var <= 0.0)
	  */
	    lo_var = 1.0;
	}

	if (md_sum == 0.0)
	  md_sum = 1.0;
	md_mean = md_tot / md_sum;
	md_var  = md_tot2 / md_sum - md_mean * md_mean;
	if (md_var <= 1.0)
	  md_var = 1.0;

	if (hi_sum == 0.0)
	  hi_sum = 1.0;
	hi_mean = hi_tot / hi_sum;
	hi_var  = hi_tot2 / hi_sum - hi_mean * hi_mean;
	if (hi_var <= 1.0) {
	    hi_var = 1.0;
	}	  

	if (loop_cnt != ADJUST5_LOOP_CNT-1) {
	  if (lo_var > md_var) {
	    lo_mean = 0.0;
	  }
	  if (hi_var > md_var) {
	    hi_mean = HIRES_GTBL_SZ-1;
	  }
	}

	/* 
	if (lo_var > md_var) {
	  lo_var = md_var;
	}
	if (hi_var > md_var) {
	  hi_var = md_var; 
	}
	*/

      }

      fprintf (stderr, 
	       "Iter %d: Lo Mean=%5.1f, Lo Var=%5.1f\n",
	       loop_cnt+1 , MCVT(lo_mean), VCVT(lo_var));
      fprintf (stderr,
	       "        Md Mean=%5.1f, Md Var=%5.1f\n",
	       MCVT(md_mean), VCVT(md_var));
      fprintf (stderr,
	       "        Hi Mean=%5.1f, Hi Var=%5.1f\n",
	       MCVT(hi_mean), VCVT(hi_var));
    }


    /*** compute the "colorized" versions of the means and variances ***/

    lo_sum = md_sum = hi_sum = 0.0;
    lo_rtot = lo_gtot = lo_btot = 0.0;
    md_rtot = md_gtot = md_btot = 0.0;
    hi_rtot = hi_gtot = hi_btot = 0.0;
    lo_rtot2 = lo_gtot2 = lo_btot2 = 0.0;
    md_rtot2 = md_gtot2 = md_btot2 = 0.0;
    hi_rtot2 = hi_gtot2 = hi_btot2 = 0.0;

    for (i = 0; i <  HIRES_GTBL_SZ; i++) {

      if (hist[i] == 0)
	continue;

      rval = red_sum[i] / (double)hist[i];
      gval = grn_sum[i] / (double)hist[i];
      bval = blu_sum[i] / (double)hist[i];
      rval2 = red_sum2[i] / (double)hist[i];
      gval2 = grn_sum2[i] / (double)hist[i];
      bval2 = blu_sum2[i] / (double)hist[i];

      val = (double)i;

      diff = val - lo_mean;
      lo_prob = exp (-0.5 * (diff * diff / lo_var));
      /* lo_prob = lo_prob * lo_prob * sqrt (2.0); */
      lo_prob = lo_prob * lo_prob * lo_prob * sqrt (3.0);
      lo_prob/= sqrt (lo_var);

      diff = val - md_mean;
      md_prob = exp (-0.5 * (diff * diff / md_var));
      
      diff = val - hi_mean;
      hi_prob = exp (-0.5 * (diff * diff / hi_var));
      /* hi_prob = hi_prob * hi_prob * sqrt (2.0); */
      hi_prob = hi_prob * hi_prob * hi_prob * sqrt (3.0);
      hi_prob/= sqrt (hi_var);

      prob = lo_prob + md_prob + hi_prob;
      lo_prob = lo_prob / prob * hist[i];
      md_prob = md_prob / prob * hist[i];
      hi_prob = hi_prob / prob * hist[i];

      lo_rtot+= rval * lo_prob;
      lo_gtot+= gval * lo_prob;
      lo_btot+= bval * lo_prob;
      lo_rtot2+= rval2 * lo_prob;
      lo_gtot2+= gval2 * lo_prob;
      lo_btot2+= bval2 * lo_prob;
      lo_sum+= lo_prob;

      md_rtot+= rval * md_prob;
      md_gtot+= gval * md_prob;
      md_btot+= bval * md_prob;
      md_rtot2+= rval2 * md_prob;
      md_gtot2+= gval2 * md_prob;
      md_btot2+= bval2 * md_prob;
      md_sum+= md_prob;

      hi_rtot+= rval * hi_prob;
      hi_gtot+= gval * hi_prob;
      hi_btot+= bval * hi_prob;
      hi_rtot2+= rval2 * hi_prob;
      hi_gtot2+= gval2 * hi_prob;
      hi_btot2+= bval2 * hi_prob;
      hi_sum+= hi_prob;
    }


    /* now calculate the updated means and variances */

    if (lo_sum == 0.0)
      lo_sum = 1.0;
    lo_rmean = lo_rtot / lo_sum;
    lo_gmean = lo_gtot / lo_sum;
    lo_bmean = lo_btot / lo_sum;
    lo_rvar  = lo_rtot2 / lo_sum - lo_rmean * lo_rmean;
    lo_gvar  = lo_gtot2 / lo_sum - lo_gmean * lo_gmean;
    lo_bvar  = lo_btot2 / lo_sum - lo_bmean * lo_bmean;

    if (md_sum == 0.0)
      md_sum = 1.0;
    md_rmean = md_rtot / md_sum;
    md_gmean = md_gtot / md_sum;
    md_bmean = md_btot / md_sum;
    md_rvar  = md_rtot2 / md_sum - md_rmean * md_rmean;
    md_gvar  = md_gtot2 / md_sum - md_gmean * md_gmean;
    md_bvar  = md_btot2 / md_sum - md_bmean * md_bmean;

    if (hi_sum == 0.0)
      hi_sum = 1.0;
    hi_rmean = hi_rtot / hi_sum;
    hi_gmean = hi_gtot / hi_sum;
    hi_bmean = hi_btot / hi_sum;
    hi_rvar  = hi_rtot2 / hi_sum - hi_rmean * hi_rmean;
    hi_gvar  = hi_gtot2 / hi_sum - hi_gmean * hi_gmean;
    hi_bvar  = hi_btot2 / hi_sum - hi_bmean * hi_bmean;

    fprintf (stderr, 
	     " Color: Lo Mean: r=%5.1f, g=%5.1f, b=%5.1f;",
	     lo_rmean, lo_gmean, lo_bmean);
    fprintf (stderr, 
	     " Lo Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     lo_rvar, lo_gvar, lo_bvar);
    fprintf (stderr, 
	     "        Md Mean: r=%5.1f, g=%5.1f, b=%5.1f;",
	     md_rmean, md_gmean, md_bmean);
    fprintf (stderr, 
	     " Md Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     md_rvar, md_gvar, md_bvar);
    fprintf (stderr, 
	     "        Hi Mean: r=%5.1f, g=%5.1f, b=%5.1f;",
	     hi_rmean, hi_gmean, hi_bmean);
    fprintf (stderr, 
	     " Hi Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     hi_rvar, hi_gvar, hi_bvar);


    /*** iterate to find more accurate versions of the variances */

    for (loop_cnt = 0; loop_cnt < 10; loop_cnt++) {

      lo_sum = md_sum = hi_sum = 0.0;
      lo_rtot = lo_gtot = lo_btot = 0.0;
      md_rtot = md_gtot = md_btot = 0.0;
      hi_rtot = hi_gtot = hi_btot = 0.0;
      lo_rtot2 = lo_gtot2 = lo_btot2 = 0.0;
      md_rtot2 = md_gtot2 = md_btot2 = 0.0;
      hi_rtot2 = hi_gtot2 = hi_btot2 = 0.0;

      for (i = 0; i <  HIRES_GTBL_SZ; i++) {

	if (hist[i] == 0)
	  continue;

	rval = red_sum[i] / (double)hist[i];
	gval = grn_sum[i] / (double)hist[i];
	bval = blu_sum[i] / (double)hist[i];
	rval2 = red_sum2[i] / (double)hist[i];
	gval2 = grn_sum2[i] / (double)hist[i];
	bval2 = blu_sum2[i] / (double)hist[i];

	rdiff = rval - lo_rmean;
	gdiff = gval - lo_gmean;
	bdiff = bval - lo_bmean;
	lo_prob = exp (-0.5 * (rdiff * rdiff / lo_rvar 
			       + gdiff * gdiff / lo_gvar
			       + bdiff * bdiff / lo_bvar));
	/* lo_prob = lo_prob * lo_prob * sqrt (2.0); */
	lo_prob = lo_prob * lo_prob * lo_prob * sqrt (3.0);
	lo_prob/= (sqrt (lo_rvar) * sqrt (lo_gvar) * sqrt (lo_bvar));

	rdiff = rval - md_rmean;
	gdiff = gval - md_gmean;
	bdiff = bval - md_bmean;
	md_prob = exp (-0.5 * (rdiff * rdiff / md_rvar 
			       + gdiff * gdiff / md_gvar
			       + bdiff * bdiff / md_bvar));
	md_prob/= (sqrt (md_rvar) * sqrt (md_gvar) * sqrt (md_bvar));
      
	rdiff = rval - hi_rmean;
	gdiff = gval - hi_gmean;
	bdiff = bval - hi_bmean;
	hi_prob = exp (-0.5 * (rdiff * rdiff / hi_rvar 
			       + gdiff * gdiff / hi_gvar
			       + bdiff * bdiff / hi_bvar));
	/* hi_prob = hi_prob * hi_prob * sqrt (2.0); */
	hi_prob = hi_prob * hi_prob * hi_prob * sqrt (3.0);
	hi_prob/= (sqrt (hi_rvar) * sqrt (hi_gvar) * sqrt (hi_bvar));

	prob = lo_prob + md_prob + hi_prob;
	lo_prob = lo_prob / prob * hist[i];
	md_prob = md_prob / prob * hist[i];
	hi_prob = hi_prob / prob * hist[i];

	lo_rtot+= rval * lo_prob;
	lo_gtot+= gval * lo_prob;
	lo_btot+= bval * lo_prob;
	lo_rtot2+= rval2 * lo_prob;
	lo_gtot2+= gval2 * lo_prob;
	lo_btot2+= bval2 * lo_prob;
	lo_sum+= lo_prob;

	md_rtot+= rval * md_prob;
	md_gtot+= gval * md_prob;
	md_btot+= bval * md_prob;
	md_rtot2+= rval2 * md_prob;
	md_gtot2+= gval2 * md_prob;
	md_btot2+= bval2 * md_prob;
	md_sum+= md_prob;

	hi_rtot+= rval * hi_prob;
	hi_gtot+= gval * hi_prob;
	hi_btot+= bval * hi_prob;
	hi_rtot2+= rval2 * hi_prob;
	hi_gtot2+= gval2 * hi_prob;
	hi_btot2+= bval2 * hi_prob;
	hi_sum+= hi_prob;
      }

      /* now calculate the updated variances */
      
      if (lo_sum == 0.0)
	lo_sum = 1.0;
      rmean = lo_rtot / lo_sum;
      gmean = lo_gtot / lo_sum;
      bmean = lo_btot / lo_sum;
      lo_rvar = lo_rtot2 / lo_sum - rmean * rmean;
      lo_gvar = lo_gtot2 / lo_sum - gmean * gmean;
      lo_bvar = lo_btot2 / lo_sum - bmean * bmean;
      if (lo_rvar <= 1.0) 
	lo_rvar = 1.0;
      if (lo_gvar <= 1.0) 
	lo_gvar = 1.0;
      if (lo_bvar <= 1.0) 
	lo_bvar = 1.0;
      
      if (md_sum == 0.0)
	md_sum = 1.0;
      rmean = md_rtot / md_sum;
      gmean = md_gtot / md_sum;
      bmean = md_btot / md_sum;
      md_rvar = md_rtot2 / md_sum - rmean * rmean;
      md_gvar = md_gtot2 / md_sum - gmean * gmean;
      md_bvar = md_btot2 / md_sum - bmean * bmean;
      if (md_rvar <= 1.0)
	md_rvar = 1.0;
      if (md_gvar <= 1.0)
	md_gvar = 1.0;
      if (md_bvar <= 1.0)
	md_bvar = 1.0;

      if (hi_sum == 0.0)
	hi_sum = 1.0;
      rmean = hi_rtot / hi_sum;
      gmean = hi_gtot / hi_sum;
      bmean = hi_btot / hi_sum;
      hi_rvar = hi_rtot2 / hi_sum - rmean * rmean;
      hi_gvar = hi_gtot2 / hi_sum - gmean * gmean;
      hi_bvar = hi_btot2 / hi_sum - bmean * bmean;
      if (hi_rvar <= 1.0) 
	hi_rvar = 1.0;
      if (hi_gvar <= 1.0) 
	hi_gvar = 1.0;
      if (hi_bvar <= 1.0) 
	hi_bvar = 1.0;

    fprintf (stderr, 
	     " Lo Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     lo_rvar, lo_gvar, lo_bvar);
    fprintf (stderr, 
	     " Md Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     md_rvar, md_gvar, md_bvar);
    fprintf (stderr, 
	     " Hi Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     hi_rvar, hi_gvar, hi_bvar);

    }


    fprintf (stderr, 
	     " Color: Lo Mean: r=%5.1f, g=%5.1f, b=%5.1f;",
	     lo_rmean, lo_gmean, lo_bmean);
    fprintf (stderr, 
	     " Lo Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     lo_rvar, lo_gvar, lo_bvar);
    fprintf (stderr, 
	     "        Md Mean: r=%5.1f, g=%5.1f, b=%5.1f;",
	     md_rmean, md_gmean, md_bmean);
    fprintf (stderr, 
	     " Md Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     md_rvar, md_gvar, md_bvar);
    fprintf (stderr, 
	     "        Hi Mean: r=%5.1f, g=%5.1f, b=%5.1f;",
	     hi_rmean, hi_gmean, hi_bmean);
    fprintf (stderr, 
	     " Hi Var: r=%5.1f, g=%5.1f, b=%5.1f\n",
	     hi_rvar, hi_gvar, hi_bvar);



    /* compute the separate channel adjustments */

    if (black_adj_flag) {
      min[0] = lo_rmean;
      min[1] = lo_gmean;
      min[2] = lo_bmean;
      if (min[0] < 0)
	min[0] = 0;
      if (min[1] < 0)
	min[1] = 0;
      if (min[2] < 0)
	min[2] = 0;
      min_tar = sqrt (lo_rvar);
      if (sqrt (lo_gvar) > min_tar)
	min_tar = sqrt (lo_gvar);
      if (sqrt (lo_bvar) > min_tar)
	min_tar = sqrt (lo_bvar);
      min_tar*= 1.5;
    }
    else {
      min[0] = 0;
      min[1] = 0;
      min[2] = 0;
      min_tar = 0;
    }
    max[0] = hi_rmean;
    max[1] = hi_gmean;
    max[2] = hi_bmean;
    if (max[0] > 255)
      max[0] = 255;
    if (max[1] > 255)
      max[1] = 255;
    if (max[2] > 255)
      max[2] = 255;
      max_tar = sqrt (hi_rvar);
      if (sqrt (hi_gvar) > max_tar)
	max_tar = sqrt (hi_gvar);
      if (sqrt (hi_bvar) > max_tar)
	max_tar = sqrt (hi_bvar);
      max_tar = 255 - max_tar * 1.5;

    /*
    if (lo_var > md_var) {
      min_tar = 0.0;
    }
    if (hi_var > md_var) {
      max_tar = 255.0;
    }
    */

    fprintf (stderr, "Max tar: %d, Min tar: %d\n", min_tar, max_tar);

    /* loop through and process the color channels */

    for (chan = 0; chan < chan_cnt; chan++) {

      min_val = min[chan];
      max_val = max[chan];

      /* check for zero range */
  
      if (max_val == min_val)
	continue;

      /* Compute the transformation map */

      scale = (double)(max_tar - min_tar) / (double)(max_val - min_val);
      offset = min_val - min_tar;
      for (i = 0; i < 256; i++) {
	hist[i] = (int)((double)(i - offset) * scale + 0.5);
	if (hist[i] > 255) hist[i] = 255;
	if (hist[i] < 0)   hist[i] = 0;
      }

      /* Apply transformation to image */

      ptr = img->data + chan;
      for (i=0; i<size; i++, ptr+= chan_cnt) 
	*ptr=hist[*ptr];
    }
  }

  return;
}



/*
 * Histogram dynamic range adjustment of an image
 *
 * This is EM based.  Starts with a hump and at the min, mid, and max.
 *
 * The monochrome implementation was never completed.
 *
 * Takes bottom 0.1% and top 0.1% of the image pixels are used for
 * initial means estimates.
 *
 * This image treats monochrome images the same as
 * img_hist_range_adjust.
 *
 * In the case of color images, it histograms the grayscale
 * version of the image.  Then it maps all three channels
 * indepdently based on the average lightest and darkest colors
 * for that gray level.
 *
 * The original data is modified.
 *
 */

#define ADJUST6_LOOP_CNT  25
#define ADJUST6_CLUST_CNT 8

void img_hist_range_adjust6 (img_t *img) {
  img_hist_range_adjust6_internal (img, 1, 0, 1, 0);
}

void img_hist_range_adjust6_white_only (img_t *img) {
  img_hist_range_adjust6_internal (img, 0, 0, 1, 0);
}

void img_hist_range_adjust6_simple_black (img_t *img) {
  img_hist_range_adjust6_internal (img, 2, 0, 1, 0);
}

void img_hist_range_adjust6_var_black (img_t *img) {
  img_hist_range_adjust6_internal (img, 3, 0, 1, 0);
}

void img_hist_range_adjust6_mix0 (img_t *img) {
  img_hist_range_adjust6_internal (img, 4, 0, 1, 0);
}

void img_hist_range_adjust6_mix1 (img_t *img) {
  img_hist_range_adjust6_internal (img, 5, 0, 1, 0);
}

void img_hist_range_adjust6_mix2 (img_t *img) {
  img_hist_range_adjust6_internal (img, 6, 0, 1, 0);
}

void img_hist_range_adjust6_fast (img_t *img) {
  img_hist_range_adjust6_internal (img, 1, 1, 1, 0);
}

void img_hist_range_adjust6_fast_mix0 (img_t *img) {
  img_hist_range_adjust6_internal (img, 4, 1, 1, 0);
}

void img_hist_range_adjust6_fast_nofix (img_t *img) {
  img_hist_range_adjust6_internal (img, 1, 1, 0, 0);
}

void img_hist_quant_adjust6 (img_t *img) {
  img_hist_range_adjust6_internal (img, 1, 0, 0, 1);
}

void img_hist_quant_adjust6_fast (img_t *img) {
  img_hist_range_adjust6_internal (img, 1, 1, 0, 1);
}


void img_hist_range_adjust6_internal (img_t *img, int black_adj_mode,
				      int fast_flag, int fix_flag,
				      int quant_flag)
{

  /* TO DO: complete code to handle monochrome images */

  /* handle monochrome images */

  if (img->chan_cnt == 1) {

    fprintf (stderr, "WARNING: monochrome processing implementation not implemented.\n");

  }

  /* handle color images */

  else {

    pix_val_t  *ptr;
    int         val;
    double      min[3], max[3], min_val, max_val, offset;
    double      sfact[3];
    double      post_offset, post_scale;
    int         chan, chan_cnt, loop_cnt;
    int         r_ival, g_ival, b_ival;
    long        size, chunk, sum, i, j;
    long        hist[HIRES_GTBL_SZ];
    long        red_sum[HIRES_GTBL_SZ], grn_sum[HIRES_GTBL_SZ], blu_sum[HIRES_GTBL_SZ];
    long        red_tot, grn_tot, blu_tot;
    double      red_sum2[HIRES_GTBL_SZ], grn_sum2[HIRES_GTBL_SZ], blu_sum2[HIRES_GTBL_SZ];
    double      red_tot2, grn_tot2, blu_tot2;
    double      scale;
    double      rval, gval, bval;
    double      rval2, gval2, bval2;
    double      rdiff, gdiff, bdiff;
    double      lo_rmean, lo_gmean, lo_bmean;
    double      hi_rmean, hi_gmean, hi_bmean;
    double      lo_rvar, lo_gvar, lo_bvar;
    double      hi_rvar, hi_gvar, hi_bvar;
    double      init_rvar, init_gvar, init_bvar;
    double      lo_part, md_part, hi_part;
    double      rmean[ADJUST6_CLUST_CNT], gmean[ADJUST6_CLUST_CNT];
    double      bmean[ADJUST6_CLUST_CNT];
    double      rvar[ADJUST6_CLUST_CNT], gvar[ADJUST6_CLUST_CNT];
    double      bvar[ADJUST6_CLUST_CNT];
    double      dist_pow[ADJUST6_CLUST_CNT];
    double      rsum[ADJUST6_CLUST_CNT], gsum[ADJUST6_CLUST_CNT];
    double      bsum[ADJUST6_CLUST_CNT];
    double      rsum2[ADJUST6_CLUST_CNT], gsum2[ADJUST6_CLUST_CNT];
    double      bsum2[ADJUST6_CLUST_CNT];
    double      prob[ADJUST6_CLUST_CNT], prob_sum[ADJUST6_CLUST_CNT];
    double      tot_prob, curr_prob;
    double      gray, lo_gray, hi_gray;
    int         lo_index, hi_index;
    double      lo_prob, hi_prob;

    /* initialization */

    chan_cnt = img->chan_cnt;
    size = img->width * img->height;

    /* handle warnings */
    lo_part = md_part = hi_part = 0.0;

    /* build tables if not already built */

     build_hires_gray_tables ();

    /*** generate initial mean and variance estimates ***/

    /* note that this is needed or the algorithm my latch on to 
     * a group of red pixels which have zero g and b values 
     */

    /* Compute histogram */

    ptr = img->data;
    for (i = 0; i < HIRES_GTBL_SZ; i++) {
      hist[i] = 0;
      red_sum[i] = grn_sum[i] = blu_sum[i] = 0;
      red_sum2[i] = grn_sum2[i] = blu_sum2[i] = 0.0;
    }
    red_tot2 = grn_tot2 = blu_tot2 = 0.0;
    for (i = 0; i < size; i++) {
      r_ival = *ptr++;
      g_ival = *ptr++;
      b_ival = *ptr++;
      val = (double)(r_ival + g_ival + b_ival) / 3.0 / 256.0 * (HIRES_GTBL_SZ-1);;
      if (val > HIRES_GTBL_SZ-1)
	val = HIRES_GTBL_SZ-1;
      hist[val]++;
      red_sum[val]+= r_ival;
      grn_sum[val]+= g_ival;
      blu_sum[val]+= b_ival;
      red_sum2[val]+= (double)(r_ival * r_ival);
      grn_sum2[val]+= (double)(g_ival * g_ival);
      blu_sum2[val]+= (double)(b_ival * b_ival);
    }

    /*** generate initial estimates ***/


    /* find the min point */

    red_tot = grn_tot = blu_tot = 0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i = 0; i <  HIRES_GTBL_SZ; i++) {
      sum+=hist[i];
      red_tot+= red_sum[i];
      grn_tot+= grn_sum[i];
      blu_tot+= blu_sum[i];
      if (sum >= chunk) {
	break;
      }
    }
    lo_rmean = (double)red_tot / (double)sum;
    lo_gmean = (double)grn_tot / (double)sum;
    lo_bmean = (double)blu_tot / (double)sum;

    /* find the max point */

    red_tot = grn_tot = blu_tot = 0;
    chunk = (double)size * 0.001;
    sum = 0;
    for (i =  HIRES_GTBL_SZ-1; i >= 0; i--) {
      sum+=hist[i];
      red_tot+= red_sum[i];
      grn_tot+= grn_sum[i];
      blu_tot+= blu_sum[i];
      if (sum >= chunk) {
	break;
      }
    }
    hi_rmean = (double)red_tot / (double)sum;
    hi_gmean = (double)grn_tot / (double)sum;
    hi_bmean = (double)blu_tot / (double)sum;

    /* lay out the centers equally between the lo and hi means */
    for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
      rmean[j] = lo_rmean + (hi_rmean - lo_rmean) * drandom ();
      gmean[j] = lo_gmean + (hi_gmean - lo_gmean) * drandom ();
      bmean[j] = lo_bmean + (hi_bmean - lo_bmean) * drandom ();
    }

    /* find the overall mean */

    red_tot = grn_tot = blu_tot = 0;
    red_tot2 = grn_tot2 = blu_tot2 = 0;
    for (i = 0; i <  HIRES_GTBL_SZ; i++) {
      red_tot+= red_sum[i];
      grn_tot+= grn_sum[i];
      blu_tot+= blu_sum[i];
      red_tot2+= red_sum2[i];
      grn_tot2+= grn_sum2[i];
      blu_tot2+= blu_sum2[i];
    }

    init_rvar = (red_tot2 - (double)red_tot * (double)red_tot / (double)size) / (double)size;
    init_gvar = (grn_tot2 - (double)grn_tot * (double)grn_tot / (double)size) / (double)size;
    init_bvar = (blu_tot2 - (double)blu_tot * (double)blu_tot / (double)size) / (double)size;
    if (init_rvar <= 1.0)
      init_rvar = 1.0;
    if (init_gvar <= 1.0)
      init_gvar = 1.0;
    if (init_bvar <= 1.0)
      init_bvar = 1.0;

    for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
      rvar[j] = init_rvar / ADJUST6_CLUST_CNT / ADJUST6_CLUST_CNT;
      gvar[j] = init_gvar / ADJUST6_CLUST_CNT / ADJUST6_CLUST_CNT;
      bvar[j] = init_bvar / ADJUST6_CLUST_CNT / ADJUST6_CLUST_CNT;
    }

    for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
      dist_pow[j] = 3.0;
    }

    /*** loop through and perform EM iterations ***/

    for (loop_cnt = -1; loop_cnt < ADJUST6_LOOP_CNT; loop_cnt++) {

      /* Compute the contributions from each pixel */

      if (loop_cnt >= 0) {
	for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
	  rsum[j] = gsum[j] = bsum[j] = 0.0;
	  rsum2[j] = gsum2[j] = bsum2[j] = 0.0;
	  prob_sum[j] = 0.0;
	}

	/* just use the results stored in the histogram */
	if (fast_flag) {
	  for (i = 0; i <  HIRES_GTBL_SZ; i++) {

	    if (hist[i] == 0)
	      continue;

	    rval  = (double)red_sum[i]  / (double)hist[i];
	    gval  = (double)grn_sum[i]  / (double)hist[i];
	    bval  = (double)blu_sum[i]  / (double)hist[i];
	    rval2 = (double)red_sum2[i] / (double)hist[i];
	    gval2 = (double)grn_sum2[i] / (double)hist[i];
	    bval2 = (double)blu_sum2[i] / (double)hist[i];

	    tot_prob = 0.0;
	    for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
	      rdiff = rval - rmean[j];
	      gdiff = gval - gmean[j];
	      bdiff = bval - bmean[j];
	      prob[j] = exp (-0.5 * (rdiff * rdiff / rvar[j] 
				     + gdiff * gdiff / gvar[j]
				     + bdiff * bdiff / bvar[j])
			     * dist_pow[j]) * sqrt (dist_pow[j]);
	      prob[j]/= (sqrt (rvar[j]) * sqrt (gvar[j]) * sqrt (bvar[j]));
	      tot_prob+= prob[j];
	    }
	    /* if (quant_flag) { */
	    if (0) {
	      double max_prob = -1.0;
	      long   max_index = 0;
	      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
		if (prob[j] > max_prob) {
		  max_prob  = prob[j];
		  max_index = j;
		}
	      }
	      tot_prob = max_prob;
	      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
		if (j != max_index)
		  prob[j] = 0.0;
	      }
	    }
	    if (tot_prob != 0.0) {
	      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
		curr_prob = prob[j] / tot_prob * (double)hist[i];
		prob_sum[j]+= curr_prob;
		rsum[j]+= rval * curr_prob;
		gsum[j]+= gval * curr_prob;
		bsum[j]+= bval * curr_prob;
		rsum2[j]+= rval2 * curr_prob;
		gsum2[j]+= gval2 * curr_prob;
		bsum2[j]+= bval2 * curr_prob;
	      }
	    }

	  }
	}

	/* loop over all of the image pixels */
	else {

	  ptr = img->data;
	  for (i = 0; i < size; i++) {
	    rval = *ptr++;
	    gval = *ptr++;
	    bval = *ptr++;
	    rval2 = rval * rval;
	    gval2 = gval * gval;
	    bval2 = bval * bval;

	    tot_prob = 0.0;
	    for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
	      rdiff = rval - rmean[j];
	      gdiff = gval - gmean[j];
	      bdiff = bval - bmean[j];
	      prob[j] = exp (-0.5 * (rdiff * rdiff / rvar[j] 
				     + gdiff * gdiff / gvar[j]
				     + bdiff * bdiff / bvar[j])
			     * dist_pow[j]) * sqrt (dist_pow[j]);
	      prob[j]/= (sqrt (rvar[j]) * sqrt (gvar[j]) * sqrt (bvar[j]));
	      tot_prob+= prob[j];
	    }
	    /* if (quant_flag) { */
	    if (0) {
	      double max_prob = -1.0;
	      long   max_index = 0;
	      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
		if (prob[j] > max_prob) {
		  max_prob  = prob[j];
		  max_index = j;
		}
	      }
	      tot_prob = max_prob;
	      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
		if (j != max_index)
		  prob[j] = 0.0;
	      }
	    }
	    if (tot_prob != 0.0) {
	      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
		curr_prob = prob[j] / tot_prob;
		prob_sum[j]+= curr_prob;
		rsum[j]+= rval * curr_prob;
		gsum[j]+= gval * curr_prob;
		bsum[j]+= bval * curr_prob;
		rsum2[j]+= rval2 * curr_prob;
		gsum2[j]+= gval2 * curr_prob;
		bsum2[j]+= bval2 * curr_prob;
	      }
	    }

	  }

	}

	/* now calculate the updated means and variances */

	for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
	  if (prob_sum[j] == 0.0) {
	    if (loop_cnt < ADJUST6_LOOP_CNT - 4) {
	      rmean[j] = lo_rmean + (hi_rmean - lo_rmean) * drandom ();
	      gmean[j] = lo_gmean + (hi_gmean - lo_gmean) * drandom ();
	      bmean[j] = lo_bmean + (hi_bmean - lo_bmean) * drandom ();
	      rvar[j] = init_rvar / ADJUST6_CLUST_CNT / ADJUST6_CLUST_CNT;
	      gvar[j] = init_gvar / ADJUST6_CLUST_CNT / ADJUST6_CLUST_CNT;
	      bvar[j] = init_bvar / ADJUST6_CLUST_CNT / ADJUST6_CLUST_CNT;
	    }
	    else {
	      rmean[j] = gmean[j] = bmean[j] = 0.0;
	      rvar[j] = gvar[j] = bvar[j] = 0.1;
	    }
	  }
	  else {
	    rmean[j] = rsum[j] / prob_sum[j];
	    gmean[j] = gsum[j] / prob_sum[j];
	    bmean[j] = bsum[j] / prob_sum[j];
	    rvar[j]  = rsum2[j] / prob_sum[j] - rmean[j] * rmean[j];
	    gvar[j]  = gsum2[j] / prob_sum[j] - gmean[j] * gmean[j];
	    bvar[j]  = bsum2[j] / prob_sum[j] - bmean[j] * bmean[j];
	    if (rvar[j] < 0.1)
	      rvar[j] = 0.1;
	    if (gvar[j] < 0.1)
	      gvar[j] = 0.1;
	    if (bvar[j] < 0.1)
	      bvar[j] = 0.1;
	  }
	}
      }

      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
	if (j == 0) {
	  fprintf (stderr, "Iter%3d: ", loop_cnt+1);
	}
	else {
	  fprintf (stderr, "         ");
	}
	fprintf (stderr, 
		 "%2ld Mean: r=%5.1f, g=%5.1f, b=%5.1f; %2ld Var: r=%6.1f, g=%6.1f, b=%6.1f\n",
		 j, rmean[j], gmean[j], bmean[j],
		 j, rvar[j], gvar[j], bvar[j]);
      }
    }

    /* do color compensation */

    if (quant_flag == 0) {

      /* compute the separate channel adjustments */

      /* find the darkest and lightest cluster */

      lo_gray = 500.0;
      hi_gray = -1.0;
      lo_index = hi_index = 0.0;
      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
	if (prob_sum[j] > 1.0) {
	  gray = rmean[j] * 0.3 + gmean[j] * 0.6 + bmean[j] * 0.1; 
	   /* gray = (rmean[j] + gmean[j] + bmean[j]) / 3.0; */
	  if (gray < lo_gray) {
	    lo_gray = gray;
	    lo_index = j;
	  }
	  if (gray > hi_gray) {
	    hi_gray = gray;
	    hi_index = j;
	  }
	}
      }
      fprintf (stderr, "Lo index: %d, Hi index: %d\n", lo_index, hi_index);

      /* mix overlapping clusters into the low and high means */

      lo_rmean = rsum[lo_index];
      lo_gmean = gsum[lo_index];
      lo_bmean = bsum[lo_index];
      lo_rvar  = rsum2[lo_index];
      lo_gvar  = gsum2[lo_index];
      lo_bvar  = bsum2[lo_index];
      lo_prob  = prob_sum[lo_index];
      hi_rmean = rsum[hi_index];
      hi_gmean = gsum[hi_index];
      hi_bmean = bsum[hi_index];
      hi_rvar  = rsum2[hi_index];
      hi_gvar  = gsum2[hi_index];
      hi_bvar  = bsum2[hi_index];
      hi_prob  = prob_sum[hi_index];
      for (j = 0; j < ADJUST6_CLUST_CNT; j++) {

        if (prob_sum[j] > 0.0) { 

	  int lo_flag = 0;
	  int hi_flag = 0;
	  int rflag, gflag, bflag;

	  /* check for low overlap */

	  if (j != lo_index) {

	    rdiff = fabs (rmean[j] - rmean[lo_index]);
	    gdiff = fabs (gmean[j] - gmean[lo_index]);
	    bdiff = fabs (bmean[j] - bmean[lo_index]);
	    rflag = (rdiff < sqrt (rvar[j]) * 1.5
		     || rdiff < sqrt (rvar[lo_index]) * 1.5);
	    gflag = (gdiff < sqrt (gvar[j]) * 1.5
		     || gdiff < sqrt (gvar[lo_index]) * 1.5);
	    bflag = (bdiff < sqrt (bvar[j]) * 1.5
		     || bdiff < sqrt (bvar[lo_index]) * 1.5);
	    if ((rdiff < 1.0 && (gflag | bflag))
		|| (gdiff < 1.0 && (rflag | bflag))
		|| (bdiff < 1.0 && (gflag | bflag))) {
	      lo_flag = 1;
	    }
	  }
	  
	  /* check for high overlap */

	  if (j != hi_index) {

	    rdiff = fabs (rmean[j] - rmean[hi_index]);
	    gdiff = fabs (gmean[j] - gmean[hi_index]);
	    bdiff = fabs (bmean[j] - bmean[hi_index]);
	    rflag = (rdiff < sqrt (rvar[j]) * 1.5
		     || rdiff < sqrt (rvar[hi_index]) * 1.5);
	    gflag = (gdiff < sqrt (gvar[j]) * 1.5
		     || gdiff < sqrt (gvar[hi_index]) * 1.5);
	    bflag = (bdiff < sqrt (bvar[j]) * 1.5
		     || bdiff < sqrt (bvar[hi_index]) * 1.5);
	    if ((rdiff < 1.0 && (gflag | bflag))
		|| (gdiff < 1.0 && (rflag | bflag))
		|| (bdiff < 1.0 && (gflag | bflag))) {
	      hi_flag = 1;
	    }
	  }
	  
	  /* process low overlap */

	  if (lo_flag && !hi_flag) {
	    lo_rmean+= rsum[j];
	    lo_gmean+= gsum[j];
	    lo_bmean+= bsum[j];
	    lo_rvar += rsum2[j];
	    lo_gvar += gsum2[j];
	    lo_bvar += bsum2[j];
	    lo_prob += prob_sum[j];
	    fprintf (stderr, "Mixing index %ld with lo mean\n", j);
	  }

	  /* process high overlap */

	  if (hi_flag && !lo_flag) {
	    hi_rmean+= rsum[j];
	    hi_gmean+= gsum[j];
	    hi_bmean+= bsum[j];
	    hi_rvar += rsum2[j];
	    hi_gvar += gsum2[j];
	    hi_bvar += bsum2[j];
	    hi_prob +=  prob_sum[j];
	    fprintf (stderr, "Mixing index %ld with hi mean\n", j);
	  }
	}
      }

      /* full black adjustment */

      {
      
	double lo_mean, hi_mean; 
	double a0_var, a1_var, a2_var;
	double red_offset[9], grn_offset[9], blu_offset[9];
	double red_scale[9], grn_scale[9], blu_scale[9];
	double max_range;

	lo_rmean/= lo_prob;
	lo_gmean/= lo_prob;
	lo_bmean/= lo_prob;
	hi_rmean/= hi_prob;
	hi_gmean/= hi_prob;
	hi_bmean/= hi_prob;

	lo_rvar = lo_rvar / lo_prob - lo_rmean * lo_rmean;
	lo_gvar = lo_gvar / lo_prob - lo_gmean * lo_gmean;
	lo_bvar = lo_bvar / lo_prob - lo_bmean * lo_bmean;
	hi_rvar = hi_rvar / hi_prob - hi_rmean * hi_rmean;
	hi_gvar = hi_gvar / hi_prob - hi_gmean * hi_gmean;
	hi_bvar = hi_bvar / hi_prob - hi_bmean * hi_bmean;

	if (lo_rvar < 0.1)
	  lo_rvar = 0.1;
	if (lo_gvar < 0.1)
	  lo_gvar = 0.1;
	if (lo_bvar < 0.1)
	  lo_bvar = 0.1;
	if (hi_rvar < 0.1)
	  hi_rvar = 0.1;
	if (hi_gvar < 0.1)
	  hi_gvar = 0.1;
	if (hi_bvar < 0.1)
	  hi_bvar = 0.1;

	/*
	lo_mean = lo_rmean * 0.3 + lo_gmean * 0.6 + lo_bmean * 0.1;
	hi_mean = hi_rmean * 0.3 + hi_gmean * 0.6 + hi_bmean * 0.1;
	*/
	lo_mean = (lo_rmean + lo_gmean + lo_bmean) / 3.0;
	hi_mean = (hi_rmean + hi_gmean + hi_bmean) / 3.0;

	/* use the low and high points for calculations */
	red_offset[0] = lo_rmean;
	grn_offset[0] = lo_gmean;
	blu_offset[0] = lo_bmean;
	red_scale[0]  = 1.0 / (hi_rmean - lo_rmean);
	grn_scale[0]  = 1.0 / (hi_gmean - lo_gmean);
	blu_scale[0]  = 1.0 / (hi_bmean - lo_bmean);

	i = 0;
	a0_var = 1.0;
	a1_var = 0.0;
	a2_var = 0.0;
	fprintf (stderr, "offset[%ld]=(%.1f,%.1f,%.1f), scale[%ld]=(%.5f,%.5f,%.5f), var=%f\n",
		 i, red_offset[i], grn_offset[i], blu_offset[i],
		 i, red_scale[i], grn_scale[i], blu_scale[i],
		 a0_var);

	min[0] = red_offset[0];
	min[1] = grn_offset[0];
	min[2] = blu_offset[0];
	sfact[0] = red_scale[0];
	sfact[1] = grn_scale[0];
	sfact[2] = blu_scale[0];
	/*
	  min[0]/= sfact[0];
	  min[1]/= sfact[1];
	  min[2]/= sfact[2];
	*/
	max[0] = min[0] + 1.0 / sfact[0];
	max[1] = min[1] + 1.0 / sfact[1];
	max[2] = min[2] + 1.0 / sfact[2];


	/* decide on final scaling factor */
	{
	  double lo_red, lo_grn, lo_blu;
	  double hi_red, hi_grn, hi_blu;

	  lo_red = lo_rmean - sqrt (lo_rvar) * 1.5;
	  lo_grn = lo_gmean - sqrt (lo_gvar) * 1.5;
	  lo_blu = lo_bmean - sqrt (lo_bvar) * 1.5;

	  hi_red = hi_rmean + sqrt (hi_rvar) * 1.5;
	  hi_grn = hi_gmean + sqrt (hi_gvar) * 1.5;
	  hi_blu = hi_bmean + sqrt (hi_bvar) * 1.5;

	  post_offset = (min[0] - lo_red + min[1] - lo_grn + min[2] - lo_blu) 
	    / 3.0;

	  /* to each of the lo_xmean's I need to add post_offset
	   * and subtract it, so the net result is that I do
	   * nothing */

	  max_range = ((hi_red + post_offset - min[0]) * sfact[0]
		       + (hi_grn + post_offset - min[1]) * sfact[1] 
		       + (hi_blu  + post_offset - min[2]) * sfact[2]) 
	    / 3.0;
	  post_scale = 1.0 / max_range;
	}
      }

      fprintf (stderr, 
	       "Final: r=[%.1f,%.1f]->[%.1f,%.1f], g=[%.1f,%.1f]->[%.1f,%.1f], b=[%.1f,%.1f]->[%.1f,%.1f]\n", 
	       min[0], max[0], post_offset, post_scale * 255.0 + post_offset,
	       min[1], max[1], post_offset, post_scale * 255.0 + post_offset,
	       min[2], max[2], post_offset, post_scale * 255.0 + post_offset);

      /* loop through the color channels */

      for (chan = 0; chan < chan_cnt; chan++) {

	min_val = min[chan];
	max_val = max[chan];

	/* check for zero range */
  
	if (max_val == min_val)
	  continue;

	/* Compute the transformation map */

	scale = (255.0 / (max_val - min_val)) * post_scale;
	offset = min_val - post_offset;
	for (i = 0; i < 256; i++) {
	  hist[i] = (int)(((double)i - offset) * scale + 0.5);
	  if (hist[i] > 255) hist[i] = 255;
	  if (hist[i] < 0)   hist[i] = 0;
	}

	/* Apply transformation to image */

	ptr = img->data + chan;
	for (i=0; i<size; i++, ptr+= chan_cnt) 
	  *ptr=hist[*ptr];
      }
    }

    /* quantize the image colors */
    else {

      double max_prob;
      long   max_index;

      ptr = img->data;
      for (i = 0; i < size; i++, ptr+= 3) {
	rval = ptr[0];
	gval = ptr[1];
	bval = ptr[2];

	max_prob  = -1.0;
	max_index = 0;
	for (j = 0; j < ADJUST6_CLUST_CNT; j++) {
	  rdiff = (rval - rmean[j]) * 0.3;
	  gdiff = (gval - gmean[j]) * 0.6;
	  bdiff = (bval - bmean[j]) * 0.1;
	  curr_prob = 1.0 / (rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);
	  if (curr_prob > max_prob) {
	    max_prob  = curr_prob;
	    max_index = j;
	  }
	}
	ptr[0] = (rmean[max_index] + 0.5);
	ptr[1] = (gmean[max_index] + 0.5);
	ptr[2] = (bmean[max_index] + 0.5);
      }
    }
  }
  return;
}


/*
 * Re-normalizes the image by scaling each channen independently.
 *
 * Afterwards all the sum of all three channels will have a mean
 * of new_mean.
 *
 * A single scale factor is applied to all three channels
 *
 * The original data is modified.
 */

void img_mean_scale (img_t *img, int new_mean)
{
  long        i, size;
  pix_val_t   xform[256];
  pix_val_t  *img_data;
  double      sum;
  int         new_val;

  size = img->width * img->height * img->chan_cnt;

  /* loop through the image pixels */

  sum = 0.0;
  img_data = img->data;
  for (i = 0; i < size; i++) {
    sum+= (double)*img_data++;
  }
  sum/= (double)size;

  /* build the xform table */

  for (i = 0; i < 256; i++) {
    new_val = (int)((double)i / (double)sum * (double)new_mean + 0.5);
    if (new_val < 0)
      new_val = 0;
    if (new_val > 255)
      new_val = 255;
    xform[i] = new_val;
  }
      
  /* xform the image */
  
  img_data = img->data;
  for (i = 0; i < size; i++, img_data++) {
    *img_data = xform[*img_data];
  }

  return;
}



/*
 * swap all the nybbles in an image
 *
 * The original data is modified.
 */

void img_nybble_swap (img_t *img)
{
  long        i, size;
  pix_val_t   xform[256];
  pix_val_t  *img_data;

  size = img->width * img->height * img->chan_cnt;

  /* compute the xform */
  for (i = 0; i < 256; i++) {
    xform[i] = (i & 0xf) * 16 + (i & 0xf0) / 16;

    /* xform[i] = (i * 2) & 0xff; */

    /*
    {
      int value, j;
      j = (i & 0xf) * 16 + (i & 0xf0) / 16;
      value = 0;
      if (j & 0x01) value+= 0x80;
      if (j & 0x02) value+= 0x40;
      if (j & 0x04) value+= 0x20;
      if (j & 0x08) value+= 0x10;
      if (j & 0x10) value+= 0x08;
      if (j & 0x20) value+= 0x04;
      if (j & 0x40) value+= 0x02;
      if (j & 0x80) value+= 0x01;
      xform[i] = value;
    }
    */
  }

  /* loop through the image pixels */

  img_data = img->data;
  for (i = 0; i < size; i++, img_data++) {
    *img_data = xform[*img_data];
  }

  return;
}



/*
 * Does simple sharpening of an on 3x3 pixel means
 *
 * A new image is created with the filtered data.
 *
 * New image values outside of the range of 0-255 are truncated.
 */

img_t *img_simple_sharpen (img_t *src_img, double weight)
{

  img_t         *new_img;
  img_dim_t      width, height, x, y, stride;
  pix_val_t     *new_data, *src_data;
  long           sum, pix_val, mx, px, my, py, src_val;
  long           chan, chan_cnt;

  /* create a new image */

  width  = src_img->width;
  height = src_img->height;
  stride = src_img->stride;
  chan_cnt = src_img->chan_cnt;
  new_img = img_create_basis (src_img, width, height, chan_cnt);
  if (new_img == NULL)
    return (NULL);

  /* sharpen the image */

  for (chan = 0; chan < chan_cnt; chan++) {
    src_data = src_img->data + chan;
    new_data = new_img->data + chan;
    for (y = 0; y < height; y++) {
      for (x = 0; x < width; x++) {
	if (x > 0) 
	  mx = -chan_cnt;
	else
	  mx = +chan_cnt;
	if (x < width-1)
	  px = +chan_cnt;
	else
	  px = -chan_cnt;
	if (y > 0) 
	  my = -stride;
	else
	  my = +stride;
	if (y < height-1)
	  py = +stride;
	else
	  py = -stride;

	src_val = *src_data;
	sum = 0
	  + (int)*(src_data+my+mx)
	  + (int)*(src_data+my)
	  + (int)*(src_data+my+px)
	  + (int)*(src_data+mx)
	  + (int)*(src_data+px)
	  + (int)*(src_data+py+mx)
	  + (int)*(src_data+py)
	  + (int)*(src_data+py+px);
      
	pix_val = src_val + (int)(((double)src_val - (double)sum / 8.0) * weight);
	if (pix_val < 0) 
	  pix_val = 0;
	else if (pix_val > 255)
	  pix_val = 255;
	*new_data = pix_val;
	new_data+= chan_cnt;
	src_data+= chan_cnt;
      }
    }
  }

  return (new_img);

}



/*
 * Does simple rotating of an image by 0, 90, 180, 270
 *
 * A rotation of 0 degrees or anything but 0, 90, 180, 270
 * results in a simple copy of the image
 * 
 * A new image is created with the new data.
 *
 */

img_t *img_simple_rotate (img_t *src_img, int angle)
{

  img_t         *new_img;
  img_dim_t      src_width, src_height, x, y, src_stride;
  img_dim_t      dest_width, dest_height, dest_stride;
  pix_val_t     *new_data, *src_data, *new_line_ptr, *new_pix_ptr;
  long           c, chan_cnt;

  /* fix up the angle */
  angle = angle % 360;
  if (angle < 0) angle = angle + 360;

  /* get the image info */

  src_width  = src_img->width;
  src_height = src_img->height;
  src_stride = src_img->stride;
  chan_cnt = src_img->chan_cnt;

  /* process the image */

  switch (angle) {
  case 90:
    dest_width  = src_height;
    dest_height = src_width;
    dest_stride = dest_width * chan_cnt;
    new_img = img_create_basis (src_img, dest_width, dest_height, chan_cnt);
    if (new_img == NULL)
      return (NULL);
    src_data = src_img->data;
    new_data = new_img->data;
    for (y = 0; y < src_height; y++) {
      new_line_ptr = new_data + (src_height - 1 - y) * chan_cnt;
      for (x = 0; x < src_width; x++) {
	new_pix_ptr = new_line_ptr + x * dest_stride;
	for (c = 0; c < chan_cnt; c++) {
	  *new_pix_ptr++ = *src_data++;
	}
      }
    }
    break;
  case 180:
    new_img = img_create_basis (src_img, src_width, src_height, chan_cnt);
    if (new_img == NULL)
      return (NULL);
    src_data = src_img->data;
    new_data = new_img->data;
    for (y = 0; y < src_height; y++) {
      new_line_ptr = new_data + (src_height - 1 - y) * src_stride;
      for (x = 0; x < src_width; x++) {
	new_pix_ptr = new_line_ptr + (src_width - 1 - x) * chan_cnt;
	for (c = 0; c < chan_cnt; c++) {
	  *new_pix_ptr++ = *src_data++;
	}
      }
    }
    break;
  case 270:
    dest_width  = src_height;
    dest_height = src_width;
    dest_stride = dest_width * chan_cnt;
    new_img = img_create_basis (src_img, dest_width, dest_height, chan_cnt);
    if (new_img == NULL)
      return (NULL);
    src_data = src_img->data;
    new_data = new_img->data;
    for (y = 0; y < src_height; y++) {
      new_line_ptr = new_data + y * chan_cnt;
      for (x = 0; x < src_width; x++) {
	new_pix_ptr = new_line_ptr + (src_width - 1 - x) * dest_stride;
	for (c = 0; c < chan_cnt; c++) {
	  *new_pix_ptr++ = *src_data++;
	}
      }
    }
    break;
  default:	/* the simple case */
    new_img = img_dup (src_img);
    break;
  }

  return (new_img);

}


/*
 * apply a gamma curve to an image
 *
 * The original data is modified.
 */

void img_apply_gamma (img_t *img, double gamma)
{
  long        i, size;
  pix_val_t  *img_data;

  static pix_val_t xform[256];
  static double last_gamma = -1.0;

  size = img->width * img->height * img->chan_cnt;

  /* compute the xform */
  if (gamma != last_gamma) {
    for (i = 0; i < 256; i++) {
      xform[i] = pow ((double)i / 255.0, gamma) * 255.0 + 0.5;
    }
    last_gamma = gamma;
  }

  /* loop through the image pixels */

  img_data = img->data;
  for (i = 0; i < size; i++, img_data++) {
    *img_data = xform[*img_data];
  }

  return;
}



/*
 * trim the borders in an image
 *
 * A new image is created with the cropped data.
 *
 * finds bounding box and crops, optionally leaving a border
 */

img_t *img_trim_borders (img_t *src_img, pix_val_t thresh, int above_flag,
			 int border_width, int cnt_thresh)
{

  img_dim_t      width, height, x, y, stride;
  img_dim_t      x_start, x_end, y_start, y_end;
  pix_val_t     *src_data;
  long           chan, chan_cnt, out_cnt;

  /* create a new image */

  width  = src_img->width;
  height = src_img->height;
  stride = src_img->stride;
  chan_cnt = src_img->chan_cnt;

  /* eliminate warnings */
  x_start = x_end = y_start = y_end = 0;

  /* find the bounding box */

  /* y_start */

  out_cnt = 0;
  src_data = src_img->data;
  for (y = 0; y < height; y++) {
    y_start = y;
    for (x = 0; x < stride; x++, src_data++) {	
      if (((above_flag && *src_data < thresh)
	   || (!above_flag && *src_data > thresh))
	  && ++out_cnt >= cnt_thresh)
	goto y_start_done;
    }
  }
 y_start_done:

  /* y_end */

  out_cnt = 0;
  for (y = 0; y < height; y++) {
    y_end = (height - 1) - y;
    src_data = src_img->data + y_end * stride;
    for (x = 0; x < stride; x++, src_data++) {	
      if (((above_flag && *src_data < thresh)
	   || (!above_flag && *src_data > thresh))
	   && ++out_cnt >= cnt_thresh)
	goto y_end_done;
    }
  }
 y_end_done:

  /* x_start */

  out_cnt = 0;
  for (x = 0; x < width; x++) {
    x_start = x;
    for (chan = 0; chan < chan_cnt; chan++) {	
      src_data = src_img->data + x * chan_cnt + chan;
      for (y = 0; y < height; y++, src_data+= stride) {
	if (((above_flag && *src_data < thresh)
	     || (!above_flag && *src_data > thresh))
	     && ++out_cnt >= cnt_thresh)
	  goto x_start_done;
      }
    }
  }
 x_start_done:

  /* x_end */

  out_cnt = 0;
  for (x = 0; x < width; x++) {
    x_end = (width - 1) - x;
    for (chan = 0; chan < chan_cnt; chan++) {	
      src_data = src_img->data + x_end * chan_cnt + chan;
      for (y = 0; y < height; y++, src_data+= stride) {
	if (((above_flag && *src_data < thresh)
	     || (!above_flag && *src_data > thresh))
	     && ++out_cnt >= cnt_thresh)
	  goto x_end_done;
      }
    }
  }
 x_end_done:


  /* add in borders */

  if (x_start > border_width)
    x_start-= border_width;
  else
    x_start = 0;

  if (y_start > border_width)
    y_start-= border_width;
  else
    y_start = 0;
  
  x_end+= border_width;
  if (x_end > width-1)
    x_end = width-1;

  y_end+= border_width;
  if (y_end > height-1)
    y_end = height-1;


  /* return the new cropped image */

  return (img_crop (src_img, x_start, y_start, x_end, y_end));

}



/*
 * Does simple sharpening of an on 3x3 pixel means
 *
 * A new image is created with the filtered data.
 *
 * New image values are bounded by the neighboring values.
 *
 * Each channel is bound independently 
 */

img_t *img_simple_sharpen_bound (img_t *src_img, double weight)
{

  img_t         *new_img;
  img_dim_t      width, height, x, y, stride;
  pix_val_t     *new_data, *src_data;
  int           sum, out_val, mx, px, my, py, src_val;
  int           chan, chan_cnt;
  int 	 	min_val, max_val;

  /* create a new image */

  width  = src_img->width;
  height = src_img->height;
  stride = src_img->stride;
  chan_cnt = src_img->chan_cnt;
  new_img = img_create_basis (src_img, width, height, chan_cnt);
  if (new_img == NULL)
    return (NULL);

  /* sharpen the image */

  for (chan = 0; chan < chan_cnt; chan++) {
    src_data = src_img->data + chan;
    new_data = new_img->data + chan;
    for (y = 0; y < height; y++) {
      for (x = 0; x < width; x++) {
	if (x > 0) 
	  mx = -chan_cnt;
	else
	  mx = +chan_cnt;
	if (x < width-1)
	  px = +chan_cnt;
	else
	  px = -chan_cnt;
	if (y > 0) 
	  my = -stride;
	else
	  my = +stride;
	if (y < height-1)
	  py = +stride;
	else
	  py = -stride;

	src_val = *src_data;
	sum = 0
	  + (int)*(src_data+my+mx)
	  + (int)*(src_data+my)
	  + (int)*(src_data+my+px)
	  + (int)*(src_data+mx)
	  + (int)*(src_data+px)
	  + (int)*(src_data+py+mx)
	  + (int)*(src_data+py)
	  + (int)*(src_data+py+px);
      
	out_val = src_val + (int)(((double)src_val - (double)sum / 8.0) * weight);

	/* bound the value */

	min_val = max_val = src_val;

	src_val = *(src_data+my+mx);
	if (src_val > max_val)
	  max_val = src_val;
	else if (src_val < min_val)
	  min_val = src_val;

	src_val = *(src_data+my);
	if (src_val > max_val)
	  max_val = src_val;
	else if (src_val < min_val)
	  min_val = src_val;

	src_val = *(src_data+my+px);
	if (src_val > max_val)
	  max_val = src_val;
	else if (src_val < min_val)
	  min_val = src_val;

	src_val = *(src_data+mx);
	if (src_val > max_val)
	  max_val = src_val;
	else if (src_val < min_val)
	  min_val = src_val;

	src_val = *(src_data+px);
	if (src_val > max_val)
	  max_val = src_val;
	else if (src_val < min_val)
	  min_val = src_val;

	src_val = *(src_data+py+mx);
	if (src_val > max_val)
	  max_val = src_val;
	else if (src_val < min_val)
	  min_val = src_val;

	src_val = *(src_data+py);
	if (src_val > max_val)
	  max_val = src_val;
	else if (src_val < min_val)
	  min_val = src_val;

	src_val = *(src_data+py+px);
	if (src_val > max_val)
	  max_val = src_val;
	else if (src_val < min_val)
	  min_val = src_val;

	/* set the value */

	if (out_val > max_val)
	  out_val = max_val;
	else if (out_val < min_val)
	  out_val = min_val;
	*new_data = out_val;
	new_data+= chan_cnt;
	src_data+= chan_cnt;
      }
    }
  }

  return (new_img);

}
