/*%****************************************************************
 *
 * Utility functions for handling JPEG images
 *
 * Charles Rosenberg, 10/28/99
 *
 * Modification History:
 *
 * 10/28/99, cjr: created from pnm.c
 *                basically a copy of "example.c" from libjpeg
 *
 ******************************************************************
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <jpeglib.h>
#include "util.h"
#include "image.h"

#define STR_SIZE 1024

/* NOTES
 *
 * Quality goes from 1-100.
 *
 * According to the libjpeg docs standard quality is about 75
 * which is quite good and there is no reason to ever go above 90
 *
 */


/* 
 *
 * function definitions 
 *
 *
 */


/* return values: image = OK, NULL = problem */

img_t *img_read_jpeg (char *fname)
{
  img_t     *new;
  FILE      *infile;
  img_dim_t  wd, ht, pix_cnt, row_stride;
  int        chan_cnt;
  pix_val_t *data;
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;

  new = img_new();
  if (strcmp (fname, "stdin") == 0)
    infile = stdin;
  else if ((infile = fopen(fname, "r")) == NULL) {
    fprintf(stderr,"IMGOPEN: Couldn't open '%s'\n", fname);
    return(NULL);
  }

  new->name = strdup (img_basename(fname));


  /* Step 1: allocate and initialize JPEG decompression object */

  /* We set up the normal JPEG error routines, then override error_exit. */
  cinfo.err = jpeg_std_error(&jerr);

  /* Now we can initialize the JPEG decompression object. */
  jpeg_create_decompress(&cinfo);


  /* Step 2: specify data source (eg, a file) */

  jpeg_stdio_src(&cinfo, infile);


  /* Step 3: read file parameters with jpeg_read_header() */

  (void) jpeg_read_header(&cinfo, TRUE);
  /* We can ignore the return value from jpeg_read_header since
   *   (a) suspension is not possible with the stdio data source, and
   *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
   * See libjpeg.doc for more info.
   */

  /* Step 4: set parameters for decompression */

  /* In this example, we don't need to change any of the defaults set by
   * jpeg_read_header(), so we do nothing here.
   */

  /* Step 5: Start decompressor */

  (void) jpeg_start_decompress(&cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  /* We may need to do some setup of our own at this point before reading
   * the data.  After jpeg_start_decompress() we have the correct scaled
   * output image dimensions available, as well as the output colormap
   * if we asked for color quantization.
   */ 

  /* cjr */
  ht = cinfo.output_height;
  wd = cinfo.output_width;
  chan_cnt = cinfo.output_components;
  row_stride = cinfo.output_width * cinfo.output_components;
  pix_cnt = ht * wd * chan_cnt;
  data = (pix_val_t *) EMALLOC (pix_cnt * sizeof(pix_val_t));
  new->data = data;
  new->height = ht;
  new->width = wd;
  new->chan_cnt = chan_cnt;
  new->stride = row_stride;
  new->ftype = IMG_FTYPE_JPEG;

  /* Step 6: while (scan lines remain to be read) */
  /*           jpeg_read_scanlines(...); */

  /* Here we use the library's state variable cinfo.output_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   */
  while (cinfo.output_scanline < cinfo.output_height) {
    /* jpeg_read_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could ask for
     * more than one scanline at a time if that's more convenient.
     */
    (void) jpeg_read_scanlines(&cinfo, (JSAMPARRAY)&data, 1);
    data+= row_stride;
  }

  /* Step 7: Finish decompression */

  (void) jpeg_finish_decompress(&cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  /* Step 8: Release JPEG decompression object */

  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_decompress(&cinfo);

  /* all done */

  fclose(infile);
  return (new);
}


/* output in jpeg (jfif) format */

/* return values: 0 = OK, 1 = problem */

int img_write_jpeg (img_t *img, char *filename, int quality)
{
  img_dim_t row_stride;
  FILE      *outfile;
  pix_val_t *data;
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;

  if (strcmp (filename, "stdout") == 0
      || strcmp (filename, "-") == 0) {
    outfile = stdout;
  }
  else {
    outfile = fopen(filename, "w");
    if (outfile == NULL)
      return (1);
  }


  /* Step 1: allocate and initialize JPEG compression object */

  /* We have to set up the error handler first, in case the initialization
   * step fails.  (Unlikely, but it could happen if you are out of memory.)
   * This routine fills in the contents of struct jerr, and returns jerr's
   * address which we place into the link field in cinfo.
   */
  cinfo.err = jpeg_std_error(&jerr);
  /* Now we can initialize the JPEG compression object. */
  jpeg_create_compress(&cinfo);


  /* Step 2: specify data destination (eg, a file) */
  /* Note: steps 2 and 3 can be done in either order. */

  /* Here we use the library-supplied code to send compressed data to a
   * stdio stream.  You can also write your own code to do something else.
   * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
   * requires it in order to write binary files.
   */
  jpeg_stdio_dest(&cinfo, outfile);


  /* Step 3: set parameters for compression */

  /* First we supply a description of the input image.
   * Four fields of the cinfo struct must be filled in:
   */
  cinfo.image_width = img->width; 	/* image width and height, in pixels */
  cinfo.image_height = img->height;
  cinfo.input_components = img->chan_cnt; /* # of components per pixel */
  if (img->chan_cnt == 1)
    cinfo.in_color_space = JCS_GRAYSCALE;   /* colorspace of input image */
  else
    cinfo.in_color_space = JCS_RGB;         /* colorspace of input image */
  /* Now use the library's routine to set default compression parameters.
   * (You must set at least cinfo.in_color_space before calling this,
   * since the defaults depend on the source color space.)
   */
  jpeg_set_defaults(&cinfo);
  /* Now you can set any non-default parameters you wish to.
   * Here we just illustrate the use of quality (quantization table) scaling:
   */
  jpeg_set_quality(&cinfo, quality, TRUE /* limit to baseline-JPEG values */);


  /* Step 4: Start compressor */

  /* TRUE ensures that we will write a complete interchange-JPEG file.
   * Pass TRUE unless you are very sure of what you're doing.
   */
  jpeg_start_compress(&cinfo, TRUE);


  /* Step 5: while (scan lines remain to be written) */
  /*           jpeg_write_scanlines(...); */

  /* Here we use the library's state variable cinfo.next_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   * To keep things simple, we pass one scanline per call; you can pass
   * more if you wish, though.
   */

  row_stride = img->width * img->chan_cnt;
  data = img->data;

  while (cinfo.next_scanline < cinfo.image_height) {
    /* jpeg_write_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could pass
     * more than one scanline at a time if that's more convenient.
     */
    (void) jpeg_write_scanlines(&cinfo, &data, 1);
    data+= row_stride;
  }


  /* Step 6: Finish compression */

  jpeg_finish_compress(&cinfo);


  /* Step 7: release JPEG compression object */

  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_compress(&cinfo);

  /* And we're done! */
  fclose(outfile);
  return (0);
}
