/* 
 * 01-27-98 CJR: Created
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "image.h"
#include "scale.h"


/* local function prototypes */

pix_val_t get_median (pix_val_t *pix_vals, img_dim_t size);
pix_val_t get_median_avg (pix_val_t *pix_vals, img_dim_t size);


/* function definitions */


/*
 * Down scales an image by subsampling.
 * Only integral scale factors are supported.
 *
 * A new image is created with the scaled down version.
 */

img_t *img_downscale_subsample (img_t *img, img_dim_t scale_factor)
{

  img_t         *new_img;
  img_dim_t      src_stride;
  img_dim_t      new_width, new_height, x, y;
  pix_val_t     *new_data, *src_data, *src_line_ptr;

  if (scale_factor < 1) 
    return (NULL);

  /* create a new image of the correct size */

  new_width = img->width / scale_factor;
  new_height = img->height / scale_factor;
  new_img = img_create (img->name, new_width, new_height, img->chan_cnt);
  if (new_img == NULL)
    return (NULL);
  new_data = new_img->data;

  /* scale down the image using subsampling */

  src_stride = img->width * scale_factor;
  src_line_ptr = img->data + scale_factor / 2 * img->width;
  for (y = 0; y < new_height; y++, src_line_ptr+= src_stride) {
    src_data = src_line_ptr + scale_factor / 2;
    for (x = 0; x < new_width; x++, src_data+=scale_factor) {
      *new_data++ = *src_data;
    }
  }
    
  return (new_img);

}



/*
 * Down scales an image by averaging
 * Only integral scale factors are supported.
 *
 * A new image is created with the scaled down version.
 *
 * supports color images
 */


img_t *img_downscale_average (img_t *img, img_dim_t scale_factor)
{

  img_t     *new_img;
  img_dim_t  src_width, src_height, src_stride;
  img_dim_t  new_width, new_height, x, y, xx, yy;
  pix_val_t *new_data, *src_data, *src_line_ptr;
  unsigned long *sum_buf, *sum_ptr;
  img_dim_t  div_factor;
  int chan_cnt, chan;

  if (scale_factor < 1) 
    return (NULL);

  src_width  = img->width;
  src_height = img->height;
  div_factor = scale_factor * scale_factor;
  chan_cnt = img->chan_cnt;
  src_stride = img->stride;

  /* create a new image of the correct size */

  new_width = src_width / scale_factor;
  new_height = src_height / scale_factor;
  new_img = img_create (img->name, new_width, new_height, chan_cnt);
  if (new_img == NULL)
    return (NULL);

  /* allocate a temporary buffer */

  sum_buf = (unsigned long *)malloc (new_width * sizeof (unsigned long));
  if (sum_buf == NULL) {
    img_free (new_img);
    return (NULL);
  }

  /* do it once for each channel */

  for (chan = 0; chan < chan_cnt; chan++) {

    /* scale down the image using averaging */

    src_line_ptr = img->data + chan;
    new_data = new_img->data + chan;
    for (y = 0; y < new_height; y++) {

      /* sum up enough pixels to build one output row */ 

      memset (sum_buf, 0, new_width * sizeof (unsigned long));
      for (yy = 0; yy < scale_factor; yy++, src_line_ptr+= src_stride) {
	sum_ptr = sum_buf;
	src_data = src_line_ptr;
	for (x = 0; x < new_width; x++, sum_ptr++) {
	  for (xx = 0; xx < scale_factor; xx++) {
	    *sum_ptr+= *src_data;
	    src_data+= chan_cnt;
	  }
	}
      }
      sum_ptr = sum_buf;
      for (x = 0; x < new_width; x++) {
	*new_data = *sum_ptr++ / div_factor;
	new_data+= chan_cnt;
      }
    }
  }

  free (sum_buf);
  return (new_img);

}


/*
 * Down scales an image by finding the median
 * Only integral scale factors are supported.
 *
 * A new image is created with the scaled down version.
 *
 * option = 0, use straight median
 * option = 1, use three middle values and average them
 */


img_t *img_downscale_median (img_t *img, img_dim_t scale_factor, int option)
{

  img_t     *new_img;
  img_dim_t  src_width, src_height;
  img_dim_t  new_width, new_height, x, y, xx, yy;
  pix_val_t *new_data, *src_data, *src_line_ptr;
  pix_val_t *area_buf, *area_ptr;
  img_dim_t  area_size;

  if (scale_factor < 1) 
    return (NULL);

  src_width  = img->width;
  src_height = img->height;
  area_size  = scale_factor * scale_factor;

  /* create a new image of the correct size */

  new_width = src_width / scale_factor;
  new_height = src_height / scale_factor;
  new_img = img_create (img->name, new_width, new_height, img->chan_cnt);
  if (new_img == NULL)
    return (NULL);
  new_data = new_img->data;

  /* allocate a temporary buffer */

  area_buf = (pix_val_t *)malloc (area_size * new_width * sizeof (pix_val_t));
  if (area_buf == NULL) {
    img_free (new_img);
    return (NULL);
  }

  /* scale down the image using averaging */

  src_line_ptr = img->data;
  for (y = 0; y < new_height; y++) {

    /* copy the pixel values to build one output row */ 

    for (yy = 0; yy < scale_factor; yy++, src_line_ptr+= src_width) {
      area_ptr = area_buf + yy * scale_factor;
      src_data = src_line_ptr;
      for (x = 0; x < new_width; x++, area_ptr+= area_size) {
	for (xx = 0; xx < scale_factor; xx++) {
	  area_ptr[xx] = *src_data++;
	}
      }
    }

    area_ptr = area_buf;
    for (x = 0; x < new_width; x++, area_ptr+= area_size) {
      if (option)
	*new_data++ = get_median_avg (area_ptr, area_size);
      else
	*new_data++ = get_median (area_ptr, area_size);
    }
  }
    
  free (area_buf);
  return (new_img);

}


/* inefficient version of get median, sorts (n/2)+1 of the
 * items and then returns the average of the middle of the array and +/- 1
 *
 * reorders the elements in place in the array
 *
 */

pix_val_t get_median_avg (pix_val_t *pix_vals, img_dim_t size)
{

  img_dim_t i, j, half_size, low_index, max_index;
  pix_val_t low_val;
  
  half_size = size / 2 + 1;

  /* do the sort loop half_size times */

  max_index = size;
  for (j = 0; j <= half_size; j++) {
    low_index = 0;
    low_val = pix_vals[low_index];
    for (i = 1; i < max_index; i++) {
      if (pix_vals[i] < low_val) {
	low_val = pix_vals[i];
	low_index = i;
      }
    }
    max_index--;
    pix_vals[low_index] = pix_vals[max_index];
    pix_vals[max_index] = low_val;
  }
  
  return (((unsigned short)pix_vals[half_size]
	   + (unsigned short)pix_vals[half_size-1]
	   + (unsigned short)pix_vals[half_size-2]) / 3);

}


/* inefficient version of get median, sorts (n/2)+1 of the
 * items and then returns the middle of the array
 *
 * reorders the elements in place in the array
 */

pix_val_t get_median (pix_val_t *pix_vals, img_dim_t size)
{

  img_dim_t i, j, half_size, low_index, max_index;
  pix_val_t low_val;
  
  half_size = size / 2;

  /* do the sort loop half_size times */

  max_index = size;
  for (j = 0; j <= half_size; j++) {
    low_index = 0;
    low_val = pix_vals[low_index];
    for (i = 1; i < max_index; i++) {
      if (pix_vals[i] < low_val) {
	low_val = pix_vals[i];
	low_index = i;
      }
    }
    max_index--;
    pix_vals[low_index] = pix_vals[max_index];
    pix_vals[max_index] = low_val;
  }
  
  return (pix_vals[half_size]);

}





/***********************************************************************/
/*
 * Up scales an image using pixel replication
 *
 * Only integral scale factors are supported.
 *
 * A new image is created with the scaled up version.
 *
 * Supports multi-channel images.
 *
 ***********************************************************************/

img_t *img_upscale_replicate (img_t *img, int scale_factor)
{

  img_t *new_img;
  img_dim_t src_width, src_height, xx, yy, src_stride;
  img_dim_t new_width, new_height, x, y;
  pix_val_t *new_data, *src_data, *src_line_ptr;
  int        chan_cnt;

  if (scale_factor < 1) 
    return (NULL);

  /* get the original image size */
  
  src_width  = img->width;
  src_height = img->height;
  src_stride = img->stride;
  chan_cnt   = img->chan_cnt;

  /* create a new image of the correct size */

  new_width  = img->width  * scale_factor;
  new_height = img->height * scale_factor;
  new_img = img_create (img->name, new_width, new_height, img->chan_cnt);
  if (new_img == NULL)
    return (NULL);
  new_data = new_img->data;

  /* scale up the image using replication */

  src_line_ptr = img->data;
  for (y = 0; y < src_height; y++, src_line_ptr+= src_stride) {
    for (yy = 0; yy < scale_factor; yy++) {
      src_data = src_line_ptr;
      for (x = 0; x < src_width; x++, src_data+=chan_cnt) {
	for (xx = 0; xx < scale_factor; xx++) {
	  if (chan_cnt == 1) {
	    *new_data++ = *src_data;
	  }
	  else {
	    *new_data++ = src_data[0];
	    *new_data++ = src_data[1];
	    *new_data++ = src_data[2];
	  }
	}
      }
    }
  }
    
  return (new_img);
}

