/*%********************************************************************
 *
 * File: util.c
 *
 * Author:  Charles Rosenberg 
 * Created: July 7, 1998
 * 
 * General purpose utilities
 *
 * Modification History
 *
 * 07/07/98, cjr: created from fnetutil.c
 *
 **********************************************************************
 */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <dirent.h>

#include "util.h"


/*%************************************************************************
 *
 * Enhanced version of fopen
 *
 * Return:  A FILE pointer
 * Errors:  Exits on errors
 *
 **************************************************************************
 */

FILE *efopen (char *path, char *mode, char *name)
{
  FILE *fptr;

  fptr = fopen (path, mode);
  if (fptr == NULL) {
    fprintf (stderr, "\nUnable to open file \"%s\" for: %s\n\n", path, name);
    exit (1);
  }

  return (fptr);

}

FILE *efopen2 (char *path, char *mode, 
	       char *src_name, int line, char *fun_name)
{
  FILE *fptr;

  fptr = fopen (path, mode);
  if (fptr == NULL) {
    fprintf (stderr, "\nUnable to open file \"%s\" for: %s:%d %s()\n\n", 
	     path, src_name, line, fun_name);
    exit (1);
  }

  return (fptr);

}


/*%************************************************************************
 *
 * Enhanced version of malloc that catches errors
 *
 * Return:  A pointer to the allocated memory block
 * Errors:  Exits on errors
 *
 **************************************************************************
 */

void *emalloc (size_t size, char *name)
{
  void *data;

  data = malloc (size);
  if (data == NULL) {
    fprintf (stderr, "\nFailure allocating memory for: %s\n\n", name);
    exit (1);
  }

  return (data);

}

void *emalloc2 (size_t size, char *src_name, int line, char *fun_name)
{
  void *data;

  data = malloc (size);
  if (data == NULL) {
    fprintf (stderr, "\nFailure allocating memory for: %s:%d %s()\n\n", 
	     src_name, line, fun_name);
    exit (1);
  }

  return (data);

}


/*%************************************************************************
 *
 * enhanced free, checks for nulls
 *
 **************************************************************************
 */

void efree (void *ptr)
{
  if (ptr != NULL)
    free (ptr);
  return;
}


/*%************************************************************************
 *
 * Enhanced version of realloc that catches errors
 *
 * Return:  A pointer to the allocated memory block
 * Errors:  Exits on errors
 *
 **************************************************************************
 */

void *erealloc (void *ptr, size_t size, char *name)
{

  ptr = realloc (ptr, size);
  if (ptr == NULL) {
    fprintf (stderr, "\nFailure allocating memory for: %s\n\n", name);
    exit (1);
  }

  return (ptr);

}

void *erealloc2 (void *ptr, size_t size, 
		 char *src_name, int line, char *fun_name)
{

  ptr = realloc (ptr, size);
  if (ptr == NULL) {
    fprintf (stderr, "\nFailure allocating memory for: %s:%d %s\n\n", 
	     src_name, line, fun_name);
    exit (1);
  }

  return (ptr);

}

/*%************************************************************************
 *
 * Enhanced version of strdup that catches errors
 *
 * Return:  A pointer to the allocated memory block
 * Errors:  Exits on errors
 *
 **************************************************************************
 */

char *estrdup (const char *s, char *name)
{

  if (s == NULL) {
    return (NULL);
  }

  else {
    char *str = strdup (s);
    if (str == NULL) {
      fprintf (stderr, "\nFailure allocating memory for: %s\n\n", name);
      exit (1);
    }
    
    return (str);
  }

}

char *estrdup2 (const char *s, char *src_name, int line, char *fun_name)
{

  if (s == NULL) {
    return (NULL);
  }

  else {
    char *str = strdup (s);
    if (str == NULL) {
      fprintf (stderr, "\nFailure allocating memory for: %s:%d %s()\n\n", 
	       src_name, line, fun_name);
      exit (1);
    }
    
    return (str);
  }

}



/*%************************************************************************
 *
 * returns a pointer to a string with the leading and trailing
 * whitespace characters removed
 *
 * Note: The input string is modified with the insertion of nulls
 *       to remove trailing white space.
 *       The pointer returned is a pointer offset into the string
 *       data to remove leading white space.
 *
 **************************************************************************
 */


char *fix_str (char *str) {

  long i, len;
  char *new_str;
  
  if (str == NULL) {
// debug
fprintf (stderr, "fix_str got a null\n");
    return (NULL);
  }
// debug
     if (strlen (str) > 4096) {
fprintf (stderr, "fix_str got a very long string\n");
     }

  len = strlen (str);
  new_str = str;

  /* remove the leading white space */

  for (i = 0; i < len; i++, new_str++)
    if (!isspace((int)*new_str))
      break;

  /* remove the trailing white space */

  for (i = len - 1; i >= 0; str[i] = '\0', i--)
    if (!isspace((int)str[i])
	&& str[i] != '\015')	/* control-M */
      break;

  return (new_str);
}


/*%************************************************************************
 *
 * useful for printing strings, changes a null pointer into
 * the empty string
 *
 **************************************************************************
 */

char *snf (char *str) 
{
  if (str == NULL)
    return ("");
  else
    return (str);
}


/*%************************************************************************
 *
 * helper function for the regular expression matcher
 * counts the number of non '*' characters in the regular expression
 *
 **************************************************************************
 */

int regexp_len (char *regexp) 
{

  int len = 0;
  char c;

  while ((c = *regexp++) != '\0') {
    if (c != '*')
      len++;
  }
  return (len);
}


/*%************************************************************************
 *
 * simple (but slow) regular expression matcher
 * handles '?' and '*'
 *
 * note that the comparison is case insensitive
 *
 * a speed improvement could be gained by keeping track of src_len
 * and rxp_len as we go, this would require another helper function
 * which the main entry point function would call and would be used
 * for recursive calls
 *
 **************************************************************************
 */

int simple_regexp_cmp (char *srcstr, char *regexp)
{

  int  src_len = strlen (srcstr);
  int  rxp_len = regexp_len (regexp);
  char src, rxp;

  /* handle NULL's */
  if (srcstr == NULL || regexp == NULL)
    return (0);

  /* we've exhausted both strings */
  if (src_len == 0 && rxp_len == 0)
    return (1);

  /* strings can't possibly match */
  if (src_len < rxp_len)
    return (0);

  /* character by character checking */
  while (1) {

    src = *srcstr;
    rxp = *regexp;

    /* we've exhausted the source string */
    if (src == '\0') {
      if (regexp_len (regexp) == 0)
	return (1);
      else
	return (0);
    }

    /* we've exhausted the regular expression string 
     * but there are still characters to match in the source string */
    else if (rxp == '\0') {
      return (0);	
    }

    /* do we have a '?' wild card? */
    else if (rxp == '?') {
      srcstr++;
      regexp++;
    }

    /* do we have a '*' wild card? */
    else if (rxp == '*') {

      /* check if this is the final character of the regexp */
      if (*(regexp+1) == '\0') {
	return (1);
      }
      /* try consuming no src characters */
      else if (simple_regexp_cmp (srcstr, regexp+1)) {
	return (1);
      }
      /* try consuming one source character */
      else {	
	srcstr++;
      }
    }

    /* simple character match */
    else if (tolower (src) == tolower (rxp)) {
      srcstr++;
      regexp++;
    }

    /* no match found */
    else {
      break;
    }
  }
  return (0);
}


/*%************************************************************************
 *
 * check is string 2 exists as a substring of string 1
 *
 * the staring character of the match is returned 
 * or NULL if no match
 *
 * match is case insensitive, but does not handle wildcards
 *
 **************************************************************************
 */

char *find_substr (char *src_str, char *find_str)
{
  int check_len;

  if (src_str == NULL || find_str == NULL)
    return (NULL);

  check_len = strlen (find_str);
  if (check_len > 0) {
    check_len--;
    while (*src_str != '\0') {
      if (tolower (*src_str) == tolower (*find_str)) {
	if (check_len == 0 
	    || strncasecmp (src_str+1, find_str+1, check_len) == 0) {
	  return (src_str);
	}
      }
      src_str++;
    }
  }
  return (NULL);
}


/*%************************************************************************
 *
 * Function to determine the length of a file
 *
 * Return:  The file name
 * Errors:  A length of -1 is returned if the file does not exist
 *          (or any other stat error occured)
 *
 **************************************************************************
 */

long get_flen (char *fname)
{
  struct stat stat_info;

  if (stat (fname, &stat_info)) 
    return (-1);
  else
    return (stat_info.st_size);
}


/*%************************************************************************
 *
 * Function to determine the existence of a directory
 *
 * Arguments: The directory name
 * Returns:   A value of zero if it does not exist
 *            Non-zero if it does
 *
 **************************************************************************
 */

int exist_dir (char *dname)
{
  DIR        *dptr;

  dptr = opendir (dname);
  if (dptr == NULL) {
    return (0);
  }
  else {
    closedir (dptr);
    return (1);
  }
}


/* generate a floating point random number between 0 and 1 */

double drandom()
{
  return ((double)(random() & RAND_MAX) / (double)RAND_MAX);
}

/* generate a long integer between 0 and n-1 */

long lrandom (long n)
{
  long val = (long)floor(drandom() * (double)n);
  if (val == n) {
    val--;
  }
  return (val);
}


/* generate a normally distributed random number 
 * with mean 0 and variance 1 
 *
 * From Numerical Recipes in C
 */

double norm_random () 
{
  static int    iset = 0;
  static double gset;
  double        fac, rsq, v1, v2;
  
  /* generate two new random numbers  */
  if (iset == 0) {
    do {
      v1 = 2.0 * drandom() - 1.0;
      v2 = 2.0 * drandom() - 1.0;
      rsq = v1 * v1 + v2 * v2;
    } while (rsq >= 1.0 || rsq == 0.0);
    fac = sqrt (-2.0 * log (rsq) / rsq);
    gset = v1 * fac;
    iset = 1;
    return (v2 * fac);
  }

  /* return the random number we already have */
  else {
    iset = 0;
    return (gset);
  }
}



/*%************************************************************************
 *
 * Utilities for reading text from files files
 *
 **************************************************************************
 */

/* these utilities skip blank spaces and lines
 * and treat lines that begin with # as comments
 */

/* this version passes back NULL when EOF is encountered */

char *fgets_clean_eof(char *s, FILE *stream) 
{
  char *ret_str;
  while (1) {
    if (fgets (s, FGETS_TEMP_STR_SIZE, stream) == NULL)
      return NULL;
    s[FGETS_TEMP_STR_SIZE-1] = '\0';
    if (s[0] == '#')
      continue;
    ret_str = fix_str(s);
    if (ret_str[0] != '\0')
      break;
  }
  return ret_str;
}

/* this version EXITS when eof is encountered */

char *fgets_clean(char *s, FILE *stream) 
{
  char *retval = fgets_clean_eof(s, stream);
  if (retval == NULL) {
    fprintf (stderr, "Error parsing input file.\n");
    exit (1);
  }
  return retval;
}


/* similar to strsep
 * but handles multiple spaces between tokens
 * and removes them and handles errors */

char *estrsep(char **str) 
{
  int   i, len;
  char *ret_str;
  
  if (*str == NULL)
    return NULL;

  len = strlen (*str);

  /* skip leading white space */
  for (i = 0; i < len; i++, (*str)++)
    if (!isspace((int)**str))
      break;
  ret_str = *str;
  
  /* terminate the string */
  for (; i < len; i++, (*str)++)
    if (isspace((int)**str)) {
      **str = '\0';
      (*str)++;
      break;
    }

  if (*ret_str == '\0') {
    fprintf (stderr, "Error parsing input file.\n");
    exit (1);
  }

  return ret_str;
}


/*%************************************************************************
 *
 * converts all the characters in a string to lowercase.
 *
 **************************************************************************
 */


void str_tolower (char *str) {
  while (*str != '\0') 
    *str++= tolower (*str);
  return;
}
